# Synced from ballistica-master-server.
# EFRO_SYNC_HASH=216690941326255100548625722057999443110
#
# Copyright (c) 2011-2024 Eric Froemling
"""Running builds in the cloud.

CloudBuild Fundamentals:
- Unlike the classic Make system, CloudBuild does not use mod-times; it
  instead uses hashes. All the input for a given build is hashed and
  generated output is associated with that hash.
- All builds with the same input hash share the same output. Thus, common
  builds will tend to return cached output and not require an actual run.
- One should expect build output to expire, be cleared, or be regenerated
  at any time, and newly generated output may not match previous output if
  there have been changes to the build environment or if the build is
  simply not deterministic. Therefore, all build output should be treated as
  volitile and immediately stored elsewhere when it is needed.
- CloudBuild only deals with direct inputs and outputs of single builds;
  it does not handle complex dependencies (builds using outputs of other
  builds as inputs/etc.). Such a system must be layered on top of
  CloudBuild.
- CloudBuilds support the concept of temporary and permanent failure. If
  a build fails in a temporary way, it is assumed that it may succeed if
  retried. Permanent failure, however, will return an error result for
  anyone attempting the same build for the duration of its lifespan.
"""

from __future__ import annotations

import os
from enum import Enum
from dataclasses import dataclass
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING, TypeVar, Generic, assert_never, Annotated

from typing_extensions import override
from efro.dataclassio import ioprepped, IOMultiType, IOAttrs
from efro.error import CleanError

from baserver.cloudfile import cloud_file_id_for_data, cloud_file_id_for_path

if TYPE_CHECKING:
    from typing import Sequence, Self, Any, Iterable, Callable


class RecipeComponentTypeID(Enum):
    """Registry of recipe component types."""

    TEST_INPUT = 'test_input'
    TEST_INPUT_2 = 'test_input_2'
    TEST_OUTPUT = 'test_output'
    TEST_OUTPUT_2 = 'test_output_2'
    WORKSPACE_COMPILE_INPUT = 'workspace_compile_input'
    WORKSPACE_COMPILE_OUTPUT = 'workspace_compile_output'


class RecipeComponent(IOMultiType[RecipeComponentTypeID]):
    """Top level class for inputs or outputs to recipes.

    Fundamentally a component is simply a collection of arbitrary data
    and CloudFiles.
    """

    @classmethod
    def set_override(
        cls,
        type_id: RecipeComponentTypeID,
        override_getter: Callable[[], type[RecipeComponent]],
    ) -> None:
        """Provide a specific type dynamically.

        Types only available in certain environments are set this way
        during environment setup.
        """
        # Defining this as a ClassVar is problematic due to dataclass
        # magic so doing direct attr wrangling here.
        overrides = getattr(cls, '_iomtoverrides', {})
        overrides[type_id] = override_getter
        setattr(cls, '_iomtoverrides', overrides)

    @override
    @classmethod
    def get_type_id(cls) -> RecipeComponentTypeID:
        # Require child classes to supply this themselves. If we
        # did a full type registry/lookup here it would require us
        # to import everything and would prevent lazy loading.
        raise NotImplementedError()

    @override
    @classmethod
    def get_type(cls, type_id: RecipeComponentTypeID) -> type[RecipeComponent]:
        # pylint: disable=cyclic-import

        # Types that require overrides to use can set None here.
        out: type[Any] | None

        t = RecipeComponentTypeID
        if type_id is t.TEST_INPUT:
            out = TestInput
        elif type_id is t.TEST_INPUT_2:
            out = TestInput2
        elif type_id is t.TEST_OUTPUT:
            out = TestOutput
        elif type_id is t.TEST_OUTPUT_2:
            out = TestOutput2
        elif type_id is t.WORKSPACE_COMPILE_INPUT:
            from baserver.workspace import WorkspaceCompileInput

            out = WorkspaceCompileInput
        elif type_id is t.WORKSPACE_COMPILE_OUTPUT:
            from baserver.workspace import WorkspaceCompileOutput

            out = WorkspaceCompileOutput
        else:
            # Make sure we handle all types.
            assert_never(type_id)

        # Swap in any override.
        overrides: dict[
            RecipeComponentTypeID, Callable[[], type[RecipeComponent]]
        ] = getattr(cls, '_iomtoverrides', {})
        override_getter = overrides.get(type_id)
        if override_getter is not None:
            override_cls = override_getter()
            assert issubclass(override_cls, RecipeComponent)
            assert override_cls.get_type_id() == type_id
            out = override_cls

        if out is None:
            raise RuntimeError('Recipe is not available in this environment.')

        return out

    def get_cloud_file_ids(self) -> set[str]:
        """Return all cloud-file-ids contained in this data."""
        raise NotImplementedError()


@dataclass
class RecipeBuildConstraints:
    """Describes how and where a build can be run."""

    allow_basn: bool = False
    allow_bamaster: bool = False

    # The build will be automatically errored if ensure_build is called
    # and it has been running for this amount of time.
    timeout_seconds: float | None = 60.0


class RecipeTypeID(Enum):
    """Registry of recipes we are able to build."""

    TEST = 'test'
    TEST2 = 'test2'
    WORKSPACE_COMPILE = 'workspace_compile'


InT = TypeVar('InT', bound=RecipeComponent)
OutT = TypeVar('OutT', bound=RecipeComponent)


class Recipe(Generic[InT, OutT]):
    """Defines the process of produing some output from some input.

    Can contain code to perform the build as well as build-specific
    context or settings for use by the CloudBuildContext.
    """

    # Types only available in certain environments can be set here
    # during environment bootstrapping.
    overrides: dict[RecipeTypeID, Callable[[], type[Recipe]]] = {}

    def __init__(self, env: CloudBuildContext) -> None:
        self.env = env

    @classmethod
    def get_type_id(cls) -> RecipeTypeID:
        """Return the type id for this recipe."""
        # We require child classes to supply this themselves. If we
        # did a full type registry/lookup here it would require us
        # to import everything and would prevent lazy loading.
        raise NotImplementedError()

    @classmethod
    def get_type(cls, type_id: RecipeTypeID) -> type[Recipe]:
        """Return the type for a given type id."""
        # pylint: disable=cyclic-import

        # Types that require overrides to use can set None here.
        out: type[Recipe] | None

        t = RecipeTypeID
        if type_id is t.TEST:
            out = TestRecipe
        elif type_id is t.TEST2:
            out = TestRecipe2
        elif type_id is t.WORKSPACE_COMPILE:
            from baserver.workspace import WorkspaceCompile

            out = WorkspaceCompile
        else:
            # Make sure we handle all types.
            assert_never(type_id)

        # Swap in any override.
        override_getter = cls.overrides.get(type_id)
        if override_getter is not None:
            override_cls = override_getter()
            assert override_cls.get_type_id() == type_id
            out = override_cls

        if out is None:
            raise RuntimeError('Recipe is not available in this environment.')

        return out

    @classmethod
    def get_input_class(cls) -> type[RecipeComponent]:
        """Return the type of input data."""
        if not TYPE_CHECKING:
            # A bit of magic to extract the generic args passed to Recipe.
            out = cls.__orig_bases__[-1].__args__[0]
            assert issubclass(out, RecipeComponent)
            return out
        raise RuntimeError('Should not get here.')

    @classmethod
    def build_constraints_for_input(
        cls, in_data: InT
    ) -> RecipeBuildConstraints:
        """Provide requirements needed to build given input data."""
        del in_data  # Unused.

        # By default we only support building on basn.
        return RecipeBuildConstraints(allow_basn=True)

    def build(self, in_data: InT) -> OutT:
        """Should be overridden by subclasses to build output given input.

        Files in in_data should be read using get_input_file_data() and
        files added to output data should be registered with
        set_output_file_data().
        """
        raise NotImplementedError()


class TemporaryBuildError(CleanError):
    """A build error that is temporary in nature.

    Builds should raise this in cases such as out-of-disk-space errors
    or network communication errors. This lets the build system know
    that the build may succeed if retried.
    """


class PermanentBuildError(CleanError):
    """A build error that is permanent in nature.

    Builds should raise this in cases such as invalid parameters,
    or input data or misconfigured environments. This lets the build
    system know that a repeat build attempt would be futile.
    """


@ioprepped
@dataclass
class BuildErrorInfo:
    """Describes an error that occurred while running a build."""

    class ErrorType(Enum):
        """Type of error that occurred."""

        TEMPORARY = 'temporary'
        PERMANENT = 'permanent'
        UNDEFINED = 'undefined'

    # Type of error that occurred. Undefined means the error was neither
    # a TemporaryBuildError nor a PermanentBuildError.
    error_type: Annotated[ErrorType, IOAttrs('t')]

    # Simple error message suitable for showing the user. Should be
    # considered public and contain no sensitive information. More
    # detailed info such as tracebacks should be made available in
    # the build log which is not public. Setting to None will show
    # an 'internal error' message.
    error_message: Annotated[str | None, IOAttrs('m')]


class CloudBuildContext(Generic[InT, OutT]):
    """Provides services such as file IO for a build.

    Can be defined for specific environments such as bamaster, basn, etc.
    """

    class State(Enum):
        """State the build context can be in."""

        NOT_STARTED = 'not_started'
        BUILDING = 'building'
        ERRORED = 'errored'
        COMPLETED = 'completed'

    def __init__(
        self,
        recipe_type: type[Recipe[InT, OutT]],
        output_dir: str | None = None,
    ) -> None:
        self._state = self.State.NOT_STARTED
        self._recipe_type = recipe_type

        self._input_file_data: dict[str, bytes] = {}
        self._input_file_paths: dict[str, str] = {}
        self._output_file_data: dict[str, bytes] = {}
        self._output_file_paths: dict[str, str] = {}

        self._output_dir = output_dir
        self._temp_dir: TemporaryDirectory | None = None
        self._next_new_file_num = 1

    def __del__(self) -> None:
        # Explicitly clean up any temp dir we made. This happens on its
        # own but gives us a resource-warning when it does. There's an
        # ignore_cleanup_errors option to TemporaryDirectory(), but that
        # doesn't seem to be quite what we want since we'd like to be
        # informed if there's something preventing the dir from actually
        # getting deleted.
        if self._temp_dir is not None:
            self._temp_dir.cleanup()

    def exec_build(self, in_data: InT) -> OutT:
        """Execute the build."""
        recipe = self._recipe_type(self)
        out: OutT | None = None
        try:
            self._begin_build(in_data)
            out = recipe.build(in_data)
            return out
        finally:
            self._end_build(out)

    def _begin_build(self, in_data: InT) -> None:
        """Call when the build is starting."""
        assert self._state is self.State.NOT_STARTED
        self._state = self.State.BUILDING

        # Ask the input-data which files it needs and make sure that
        # exact set was provided to us.
        input_file_ids: set[str] = in_data.get_cloud_file_ids()

        # Complain about needed files that weren't provided.
        for file_id in input_file_ids:
            if (
                file_id not in self._input_file_data
                and file_id not in self._input_file_paths
            ):
                raise RuntimeError(
                    f'Cloud file id {file_id} is needed as a build input'
                    f' but was not provided in the environment.'
                    f' {self._input_file_data=} {self._input_file_paths=}'
                )
        # Complain about provided files that aren't needed.
        fdict: Iterable[str]
        for fdict in (
            self._input_file_data.keys(),
            self._input_file_paths.keys(),
        ):
            for file_id in fdict:
                if file_id not in input_file_ids:
                    raise RuntimeError(
                        f'Cloud file id {file_id} was provided to the'
                        f' build environment as an input but is not needed'
                        f' by the build.'
                    )

    def _end_build(self, output: OutT | None) -> None:
        """Call when the build is done (completed or errored)."""
        assert self._state is self.State.BUILDING
        self._state = (
            self.State.ERRORED if output is None else self.State.COMPLETED
        )

        # Sanity test: check the full set of output datas and files and
        # make sure that lines up with what the build output gives us.
        if output is not None:
            all_output_files = set(self._output_file_data) | set(
                self._output_file_paths
            )
            assert output.get_cloud_file_ids() == all_output_files

    def set_data_for_input_file(
        self, data: bytes, cloud_file_id: str | None = None
    ) -> str:
        """Provide a build input file as inline data.

        Must be called before the build is run. Returns a cloud-file-id
        for the given data. If a cloud-file-id is already known for the
        data, it can be passed to avoid the calculation.
        """
        if cloud_file_id is None:
            cloud_file_id = cloud_file_id_for_data(data)
        self._input_file_data[cloud_file_id] = data
        return cloud_file_id

    def set_path_for_input_file(
        self, path: str, cloud_file_id: str | None = None
    ) -> str:
        """Provide a build input file as an existing file on disk.

        Must be called before the build is run. Returns a cloud-file-id
        for the given data. If a cloud-file-id is already know for the
        file, it can be passed to avoid the calculation.
        """
        if cloud_file_id is None:
            cloud_file_id = cloud_file_id_for_path(path)
        self._input_file_paths[cloud_file_id] = path
        return cloud_file_id

    def get_input_file_data(self, file_id: str) -> bytes:
        """Get an input file in its entirety.

        Must be called during a build.
        """
        if self._state is not self.State.BUILDING:
            raise RuntimeError('This can only be called while building.')

        # If we already have it as data, easy peasy.
        data = self._input_file_data.get(file_id)
        if data is not None:
            return data

        # If we have it as a file, convert to data.
        path = self._input_file_paths.get(file_id)
        if path is not None:
            with open(path, 'rb') as infile:
                data = infile.read()
            # Keep it around in case anyone else asks.
            self._input_file_data[file_id] = data
            return data

        raise ValueError(
            f'Input file id {file_id} is being requested'
            f' but was not provided to the environment.'
        )

    def get_input_file_path(self, file_id: str) -> str:
        """Get an input file as a path to a file on disk.

        Should be passed a fileid that is stored with input data.
        """
        if self._state is not self.State.BUILDING:
            raise RuntimeError('This can only be called while building.')

        # If we already have it as a path, easy peasy.
        path = self._input_file_paths.get(file_id)
        if path is not None:
            return path

        # If we have it as data, convert it to a file.
        data = self._input_file_data.get(file_id)
        if data is not None:
            path = self.new_file_path()
            with open(path, 'wb') as outfile:
                outfile.write(data)
            # Keep it around in case anyone else asks.
            self._input_file_paths[file_id] = path
            return path

        raise ValueError(
            f'Input file id {file_id} is being requested'
            f' but was not provided to the environment.'
        )

    def set_output_file_data(self, data: bytes) -> str:
        """Provide inline data as a build output file.

        Must be called during a build.
        """
        if self._state is not self.State.BUILDING:
            raise RuntimeError('This can only be called while building.')
        cloud_file_id = cloud_file_id_for_data(data)
        self._output_file_data[cloud_file_id] = data
        return cloud_file_id

    def set_output_file_path(self, path: str) -> str:
        """Provide a file on disk as a build output file.

        Returns a cloud-file-id that should be stored with output data.
        """
        if self._state is not self.State.BUILDING:
            raise RuntimeError('This can only be called while building.')
        cloud_file_id = cloud_file_id_for_path(path)
        self._output_file_paths[cloud_file_id] = path
        return cloud_file_id

    def new_file_path(self) -> str:
        """Provide a new file path in the environment.

        Can be used to write file output data. If an output-dir was
        provided when creating the context, files will go there.
        Otherwise they go to a temp dir that is cleaned up with the
        context.
        """
        if self._output_dir is not None:
            output_dir = self._output_dir
        else:
            if self._temp_dir is None:
                # pylint: disable=consider-using-with
                self._temp_dir = TemporaryDirectory()
            output_dir = self._temp_dir.name

        filenum = self._next_new_file_num
        self._next_new_file_num += 1
        return os.path.join(output_dir, str(filenum))

    def get_native_output_file_datas(self) -> dict[str, bytes]:
        """Returns build outputs that were provided as data."""
        if self._state is not self.State.COMPLETED:
            raise RuntimeError('This can only be called on completed builds.')
        return self._output_file_data

    def get_native_output_file_paths(self) -> dict[str, str]:
        """Returns build outputs that were provided as files.

        Be aware that these files only remain valid as long as the
        build context exists.
        """
        if self._state is not self.State.COMPLETED:
            raise RuntimeError('This can only be called on completed builds.')
        return self._output_file_paths

    def get_output_files_as_paths(self) -> dict[str, str]:
        """Return all build output files as file paths.

        This requires writing any files provided as data to disk.
        """
        native_file_datas = self.get_native_output_file_datas()
        native_file_paths = self.get_native_output_file_paths()
        for fileid, filedata in native_file_datas.items():
            # Its technically possible that a file was provided as both
            # a file and as data. Can ignore those.
            if fileid in native_file_paths:
                continue
            path = self.new_file_path()
            with open(path, 'wb') as outfile:
                outfile.write(filedata)
            self._output_file_paths[fileid] = path

        # At this point _output_file_paths should hold everything.
        return self._output_file_paths


@ioprepped
@dataclass
class TestInput(RecipeComponent):
    """Test input data."""

    in_value: int
    in_file_id: str
    in_file_2_id: str

    @override
    @classmethod
    def get_type_id(cls) -> RecipeComponentTypeID:
        return RecipeComponentTypeID.TEST_INPUT

    @override
    def get_cloud_file_ids(self) -> set[str]:
        return {self.in_file_id, self.in_file_2_id}


@ioprepped
@dataclass
class TestOutput(RecipeComponent):
    """Test output data 2."""

    out_value: int
    out_file_id: str
    out_file_2_id: str

    @override
    @classmethod
    def get_type_id(cls) -> RecipeComponentTypeID:
        return RecipeComponentTypeID.TEST_OUTPUT

    @override
    def get_cloud_file_ids(self) -> set[str]:
        return {self.out_file_id, self.out_file_2_id}


class TestRecipe(Recipe[TestInput, TestOutput]):
    """A recipe to build a lovely TestOutput from a TestInput."""

    @override
    @classmethod
    def get_type_id(cls) -> RecipeTypeID:
        return RecipeTypeID.TEST

    @override
    @classmethod
    def build_constraints_for_input(
        cls,
        in_data: TestInput,
    ) -> RecipeBuildConstraints:
        return RecipeBuildConstraints(
            allow_basn=True, allow_bamaster=True, timeout_seconds=7.0
        )

    @override
    def build(self, in_data: TestInput) -> TestOutput:

        print('Hello from TestRecipe build!')

        env = self.env

        # We can access input/output files simply as bytes:
        in_file_data = env.get_input_file_data(in_data.in_file_id)
        out_file_id = env.set_output_file_data(in_file_data + b'-output')

        # Or as files on disk (to handle larger data or support command
        # line calls):
        in_file_2_path = env.get_input_file_path(in_data.in_file_2_id)
        with open(in_file_2_path, 'rb') as infile:
            in_file_2_data = infile.read()
        out_file_2_path = env.new_file_path()
        with open(out_file_2_path, 'wb') as outfile:
            outfile.write(in_file_2_data + b'-output2')
        out_file_2_id = env.set_output_file_path(out_file_2_path)

        return TestOutput(
            out_value=in_data.in_value + 1,
            out_file_id=out_file_id,
            out_file_2_id=out_file_2_id,
        )


@ioprepped
@dataclass
class TestInput2(RecipeComponent):
    """Test input data."""

    foof2: str

    @override
    @classmethod
    def get_type_id(cls) -> RecipeComponentTypeID:
        return RecipeComponentTypeID.TEST_INPUT_2

    @override
    def get_cloud_file_ids(self) -> set[str]:
        return set()


@ioprepped
@dataclass
class TestOutput2(RecipeComponent):
    """Test output data 2."""

    barf2: str

    @override
    @classmethod
    def get_type_id(cls) -> RecipeComponentTypeID:
        return RecipeComponentTypeID.TEST_OUTPUT_2

    @override
    def get_cloud_file_ids(self) -> set[str]:
        return set()


class TestRecipe2(Recipe[TestInput2, TestOutput2]):
    """Testing2."""

    @override
    @classmethod
    def get_type_id(cls) -> RecipeTypeID:
        return RecipeTypeID.TEST2

    @override
    def build(self, in_data: TestInput2) -> TestOutput2:
        return TestOutput2(barf2=in_data.foof2 + '_out')
