# Synced from ballistica-master-server.
# EFRO_SYNC_HASH=123402920171454696187098419772689954678
#
"""Cryptography related functionality."""

from __future__ import annotations

import json
import zlib
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any

# NOTE: 'wire' format here means base64 strings or json dicts of simple
# alphanumeric keys pointing at base64 strings. This data should be safe
# to transmit raw over http, etc.


def str_to_wire_key_aes(s: str, key: bytes) -> str:
    """Compress/encrypt a string to a wire-ready string."""
    return encrypt_aes(zlib.compress(s.encode()), key)


def str_from_wire_key_aes(data: str, key: bytes) -> str:
    """Decrypt/decompress a wire-ready string to its original string form."""
    zipdata = decrypt_aes(data, key)
    return zlib.decompress(zipdata).decode()


def encrypt_aes(data: bytes, key: bytes) -> str:
    """Fancy-pants authenticated encryption."""
    # noinspection PyProtectedMember
    from Cryptodome.Cipher import AES, _mode_ccm
    from base64 import b64encode

    assert isinstance(data, bytes)
    assert isinstance(key, bytes)
    # (header can be used to send data unencrypted but
    # still authenticated)
    header = b''
    cipher = AES.new(key, AES.MODE_CCM)
    assert isinstance(cipher, _mode_ccm.CcmMode)
    cipher.update(header)
    ciphertext, tag = cipher.encrypt_and_digest(data)
    json_k = [
        'n',  # nonce
        'h',  # header
        'c',  # ciphertext
        't',  # tag
    ]
    json_v = [
        b64encode(x).decode() for x in (cipher.nonce, header, ciphertext, tag)
    ]
    return json.dumps(dict(zip(json_k, json_v)), separators=(',', ':'))


def decrypt_aes(data: str, key: bytes) -> bytes:
    """Fancy-pants authenticated decryption."""
    # noinspection PyProtectedMember
    from Cryptodome.Cipher import AES, _mode_ccm
    from base64 import b64decode

    try:
        b64 = json.loads(data)
        json_k = [
            'n',  # nonce
            'h',  # header
            'c',  # ciphertext
            't',  # tag
        ]
        jvals = {k: b64decode(b64[k]) for k in json_k}
        cipher = AES.new(key, AES.MODE_CCM, nonce=jvals['n'])
        assert isinstance(cipher, _mode_ccm.CcmMode)
        cipher.update(jvals['h'])
        output = cipher.decrypt_and_verify(jvals['c'], jvals['t'])
        assert isinstance(output, bytes)
        return output
    except (ValueError, KeyError) as exc:
        raise ValueError('Incorrect decryption') from exc
