# Synced from ballistica-master-server.
# EFRO_SYNC_HASH=225150693661615286224568497766245138850
#
# Copyright (c) 2011-2024 Eric Froemling
"""Cryptography related functionality."""
from __future__ import annotations

from base64 import urlsafe_b64encode, b64encode, urlsafe_b64decode, b64decode
import zlib
from typing import TYPE_CHECKING

from efro.error import IntegrityError

if TYPE_CHECKING:
    pass

# NOTE: 'wire' format here means base64 strings or json dicts of simple
# alphanumeric keys pointing at base64 strings. This data should be safe
# to transmit raw over http, etc.


def str_to_wire(s: str) -> str:
    """Compress/encrypt a string to a wire-ready string."""
    return b64encode(zlib.compress(s.encode())).decode()


def str_from_wire(data: str) -> str:
    """Decrypt/decompress a wire-ready string to its original string form."""
    return zlib.decompress(b64decode(data)).decode()


def str_to_wire_key_plaintext(s: str, key: bytes) -> str:
    """Compress/encrypt a string to a wire-ready string.

    This version is for use with SSL connections where we don't actually
    need encryption; it simply bundles the key alongside the data.
    """
    return b64encode(encrypt_plaintext(zlib.compress(s.encode()), key)).decode()


def str_from_wire_key_plaintext(data: str, key: bytes) -> str:
    """Decrypt/decompress a wire-ready string to its original string form.

    This version is for use with SSL connections where we don't actually
    need encryption; it simply bundles the key alongside the data.
    """
    zipdata = decrypt_plaintext(b64decode(data.encode()), key)
    return zlib.decompress(zipdata).decode()


def str_to_wire_key_xor(
    s: str,
    key: bytes,
    compress: bool = True,
    checksum: bool = False,
    urlsafe: bool = False,
    offset: int = 0,
) -> str:
    """Compress/encrypt a string to wire-ready form with xor encryption.

    If compress is True, zlib compression will be applied to the data.
    If checksum is True, a checksum will be appended with the data.
    Note that the exact same settings must be used when pulling data
    from the wire.
    """
    sbytes = s.encode()

    # If desired, add a 32 bit checksum at the end.
    if checksum:
        sbytes += zlib.crc32(sbytes).to_bytes(4, 'big')

    sbytes = xor_bytes(
        zlib.compress(sbytes) if compress else sbytes, key, offset=offset
    )

    return (
        urlsafe_b64encode(sbytes) if urlsafe else b64encode(sbytes)
    ).decode()


def str_from_wire_key_xor(
    data: str,
    key: bytes,
    compress: bool = True,
    checksum: bool = False,
    urlsafe: bool = False,
    offset: int = 0,
) -> str:
    """Decrypt/decompress a wire-ready string to its original string form.

    If the data is checksummed, raises an efro.IntegrityError on mismatches.
    """
    unbased = (
        urlsafe_b64decode(data.encode())
        if urlsafe
        else b64decode(data.encode())
    )
    decoded = xor_bytes(unbased, key, offset=offset)
    unzipped = zlib.decompress(decoded) if compress else decoded

    if checksum:
        if len(unzipped) < 4:
            raise ValueError('Invalid checksum')
        csval = int.from_bytes(unzipped[-4:], 'big')
        unzipped = unzipped[:-4]
        if zlib.crc32(unzipped) != csval:
            raise IntegrityError()

    return unzipped.decode()


def str_to_wire_key_legacy(s: str, key: bytes) -> str:
    """Convert a string into a encrypted/compressed base64 string.

    This uses simple encryption which is not terribly secure and
    should only be used for barebones encryption purposes.
    This is used in communication with master-server v1,
    among other things.
    """
    return b64encode(legacy_xor_bytes(zlib.compress(s.encode()), key)).decode()


def str_from_wire_key_legacy(data: str, key: bytes) -> str:
    """Convert an encrypted/compressed base64 string to a raw string.

    This uses simple encryption which is not terribly secure and
    should only be used for barebones encryption purposes.
    This is used in communication with master-server v1,
    among other things.
    """
    unbased = b64decode(data.encode())
    decoded = legacy_xor_bytes(unbased, key)
    unzipped = zlib.decompress(decoded)
    return unzipped.decode()


def encrypt_plaintext(data: bytes, key: bytes) -> bytes:
    """Combine a key with data but don't actually encrypt.

    Obviously this only makes sense if paired with TLS for security.
    """
    assert len(key) < 255

    # First byte is key length; then key; then data.
    return bytes([len(key)]) + key + data


def decrypt_plaintext(data: bytes, key: bytes) -> bytes:
    """Not-actually-encrypted decryption (for use alongside ssl/etc).

    Simply checks for the presence of the key in the data.
    """
    try:
        keylen = int(data[0])
        if (1 + keylen) > len(data):
            raise RuntimeError(
                f'Data (len {len(data)}) not long enough'
                f' to hold key ({keylen}).'
            )
        embedded_key = data[1 : 1 + keylen]
        if key != embedded_key:
            raise RuntimeError('Data key is incorrect.')
        return data[1 + keylen :]
    except Exception as exc:
        raise ValueError('Invalid data.') from exc


# Note: this is bad. It xors everything by each byte of the key
# individually so it would only take 256 * len(key) attempts to guess,
# versus 256^len(key) if we just did a single pass with the key. But we
# need it for some legacy purposes.
def legacy_xor_bytes(data: bytes, key: bytes) -> bytes:
    """Apply simple xor encryption/decryption to bytes."""
    key2 = list(key)
    keylen = len(key2)
    output = b''
    for i, val in enumerate(data):
        output += bytes([val ^ key2[i % keylen]])
    return output


def xor_bytes(data: bytes, key: bytes, offset: int = 0) -> bytes:
    """Simply xor data by some key.

    Note: xor encryption can be considered perfectly secure
    ONLY IF we use a one-time-pad where:
    - the key is as long as the data
    - the key is fully random (os.urandom())
    - the key is never reused
    """
    assert key
    assert offset >= 0
    keylen = len(key)
    return bytes(data[i] ^ key[(i + offset) % keylen] for i in range(len(data)))
