# Synced from ballistica-master-server.
# EFRO_SYNC_HASH=315550106750083109901652057148281005946
#
"""Misc types shared between server components."""

from __future__ import annotations

from typing import TYPE_CHECKING
from enum import Enum, unique

if TYPE_CHECKING:
    pass


@unique
class OSType(Enum):
    """OS we run on a ServerNode."""

    UBUNTU_20 = 'ubuntu20'
    UBUNTU_22 = 'ubuntu22'


@unique
class EC2InstanceType(Enum):
    """EC2 instance type for servernodes."""

    T3A_NANO = 't3a.nano'
    T3A_MICRO = 't3a.micro'
    T3A_SMALL = 't3a.small'
    T3A_MEDIUM = 't3a.medium'
    T3A_LARGE = 't3a.large'
    T3A_XLARGE = 't3a.xlarge'
    T3A_X2LARGE = 't3a.2xlarge'
    T3_NANO = 't3.nano'
    T3_MICRO = 't3.micro'
    T3_SMALL = 't3.small'
    T3_MEDIUM = 't3.medium'
    T3_LARGE = 't3.large'
    T3_XLARGE = 't3.xlarge'
    T3_X2LARGE = 't3.2xlarge'
    T4G_NANO = 't4g.nano'
    T4G_MICRO = 't4g.micro'
    T4G_SMALL = 't4g.small'
    T4G_MEDIUM = 't4g.medium'
    T4G_LARGE = 't4g.large'
    T4G_XLARGE = 't4g.xlarge'
    T4G_X2LARGE = 't4g.2xlarge'


@unique
class ServerNodeRunMode(Enum):
    """Overall mode of operation for a server node."""

    # Fully opt with min log spewage; for production.
    RELEASE = 'release'

    # Medium spewage; for general dev and some production testing.
    TEST = 'test'

    # Lots of log spewage and tracemalloc enabled; for debugging
    # problems.
    DEBUG = 'debug'

    @property
    def level(self) -> int:
        """Numeric level (higher being more debuggy)."""
        cls = type(self)
        match self:
            case cls.RELEASE:
                return 0
            case cls.TEST:
                return 1
            case cls.DEBUG:
                return 2

    @property
    def env_vars(self) -> dict[str, str]:
        """Env vars we set for this run mode."""
        cls = type(self)
        match self:
            case cls.RELEASE:
                return {'PYTHONOPTIMIZE': '1'}
            case cls.TEST:
                return {'PYTHONDEVMODE': '1'}
            case cls.DEBUG:
                return {'PYTHONDEVMODE': '1', 'PYTHONTRACEMALLOC': '1'}

    def get_env_var_assign_code(self) -> str:
        """Shell code to set all env vars for this run mode.

        This is intended to be placed before the Python binary in
        a shell command.
        """
        out = ''
        # Build list of flags for our run-mode.
        dev_flags: list[str] = []
        for key, val in sorted(self.env_vars.items()):
            dev_flags.append(f'{key}={val}')
        out = ' '.join(dev_flags)
        if out:
            out += ' '
        return out


@unique
class ServerNodeTaskState(Enum):
    """Current state of a task."""

    UNKNOWN = 'unknown'
    LAUNCHING = 'launching'
    RUNNING = 'running'
    CLEANING_UP = 'cleaningup'
    COMPLETED = 'completed'
    ERRORED = 'errored'
    KILLED = 'killed'


@unique
class TaskTestType(Enum):
    """Types of test-tasks that can be run on a basn node."""

    PARTIAL_LINES = 'partiallines'
    STDOUT_STDERR = 'stdoutstderr'
    LOG_DEBUG = 'logdebug'
    LOG_INFO = 'loginfo'
    LOG_WARNING = 'logwarning'
    LOG_ERROR = 'logerror'
    LOG_CRITICAL = 'logcritical'
    RETURN_ERROR = 'returnerror'
    MINUTE = 'minute'
    TEN_MINUTES = 'tenminutes'
    HOUR = 'hour'
    DAY = 'day'
    INLINE = 'inline'
    INLINE_ERROR = 'inlineerror'
    SANDBOX = 'sandbox'

    @property
    def description(self) -> str:
        """Return description for the enum."""
        # pylint: disable=too-many-return-statements
        cls = type(self)
        match self:
            case cls.PARTIAL_LINES:
                return 'Test Partial Lines'
            case cls.STDOUT_STDERR:
                return 'Test Stdout Stderr'
            case cls.LOG_DEBUG:
                return 'Test Log Debug'
            case cls.LOG_INFO:
                return 'Test Log Info'
            case cls.LOG_WARNING:
                return 'Test Log Warning'
            case cls.LOG_ERROR:
                return 'Test Log Error'
            case cls.LOG_CRITICAL:
                return 'Test Log Critical'
            case cls.RETURN_ERROR:
                return 'Test Return Error'
            case cls.MINUTE:
                return 'Test Minute'
            case cls.TEN_MINUTES:
                return 'Ten Minutes'
            case cls.HOUR:
                return 'Test Hour'
            case cls.DAY:
                return 'Test Day'
            case cls.INLINE:
                return 'Test Inline'
            case cls.INLINE_ERROR:
                return 'Test Inline Error'
            case cls.SANDBOX:
                return 'Test Sandbox'


class V1Store(Enum):
    """V1 Store IDs."""

    APPLE1 = 'apple1'
    APPLE2 = 'apple2'
    AMAZON = 'amazon'
    GOOGLE = 'google'
