# Synced from ballistica-master-server.
# EFRO_SYNC_HASH=32463321148405830986585237568542397041
#
"""Provides HostEnv."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from baserver.request import ProvidesRequestAuthentication


class HostEnv:
    """Provides various hosting config and functionality.

    A custom implementation should be set for each hosting app.
    """

    _env: HostEnv | None = None

    @classmethod
    def set(cls, env: HostEnv) -> None:
        """Set the singleton."""
        assert cls._env is None
        cls._env = env

    @classmethod
    def get(cls) -> HostEnv:
        """Get the singleton."""
        if cls._env is None:
            raise RuntimeError('A HostEnv has not been set.')
        return cls._env

    def get_auth(self) -> ProvidesRequestAuthentication:
        """Return authentication for the current rendering page."""
        raise NotImplementedError()

    @classmethod
    def get_support_email(cls, html: bool = False) -> str:
        """Return support email to show to users."""
        addr = cls.get().do_get_support_email()
        return f'<a href="mailto:{addr}">{addr}</a>' if html else addr

    def do_get_support_email(self) -> str:
        """Return support email to show to users.

        This should be overrridden by subclasses. Callers should
        use get_support_email().
        """
        raise NotImplementedError()

    def is_prod_fleet(self) -> bool:
        """Return whether the current instance is in the production fleet."""
        raise NotImplementedError()

    def get_fleet_name(self) -> str:
        """Return the name of the current fleet."""
        raise NotImplementedError()

    @classmethod
    def get_default_log_extra(cls) -> dict:
        """Return a default logextra dict."""
        if cls._env is None:
            return {'baSrc': 'unknown'}
        return cls._env.do_get_default_log_extra()

    def do_get_default_log_extra(self) -> dict:
        """Return a default logextra dict.

        Callers should use the get_default_log_extra classmethod which
        gracefully handles the case where HostEnv has not yet been inited.
        """
        raise NotImplementedError()

    def get_root_page_address(self) -> str:
        """Get the root page address for the current fleet."""
        raise NotImplementedError()
