# Synced from ballistica-master-server.
# EFRO_SYNC_HASH=194800195061300098379379529949138081903
#
"""Utils for creating and parsing html forms."""

from __future__ import annotations

import html
import json
import copy
from enum import Enum
from typing import TYPE_CHECKING, TypeVar, Generic, get_type_hints

from typing_extensions import override
from efro.error import CleanError

if TYPE_CHECKING:
    from typing import Callable, Any, Literal

    from baserver.page import HandlerBase

T = TypeVar('T')
ET = TypeVar('ET', bound=Enum)

FORM_ID_NAME = 'bamfid'
INPUT_BASE_ID = 'bamfval'


class Form:
    """Class to create html forms and apply their posted output to objects."""

    class Style(Enum):
        """How a form can be rendered."""

        HORIZONTAL = 'horizontal'
        VERTICAL = 'vertical'

    def __init__(
        self,
        form_id: str,
        inputs: list[FormInput] | None = None,
        label: str | None = None,
        on_apply: Callable[[], None] | None = None,
    ):
        self.form_id = form_id
        self.inputs = inputs if inputs is not None else []
        self.label = label if label is not None else form_id
        self.on_apply = on_apply
        self.input_map: dict[str, FormInput] = {}

    def emit_html(
        self,
        style: Style = Style.HORIZONTAL,
        extras: dict[str, str] | None = None,
        method: Literal['get', 'post'] = 'post',
    ) -> str:
        """Spit out html for the form."""
        if extras is None:
            extras = {}
        self._init_inputs()
        # joiner = '<br>' if style is self.Style.VERTICAL else ''
        joiner = ''
        flexflow = (
            'row wrap' if style is self.Style.HORIZONTAL else 'column wrap'
        )
        inputs = joiner.join(
            f'<div style="display:flex; align-items: center">'
            f'{i.emit_html(style)}</div>'
            for i in self.inputs
        )
        return (
            f'<form method="{method}">'
            f'<input type="hidden" name="{FORM_ID_NAME}"'
            f' value="{self.form_id}">'
            + ''.join(
                f'<input type="hidden" name="{extraname}"'
                f' value="{extraval}">'
                for extraname, extraval in extras.items()
            )
            + (
                '<div style="display:flex;align-items:center;'
                f'justify-content:center;flex-flow:{flexflow};'
                # 'column-gap:8px'
                '">'
            )
            + f'{inputs}'
            f'{joiner}'
            # f'<input type="submit"'
            f'<input type="submit" style="margin:5px"'
            f' value="{self.label}">'
            '</div>'
            f'</form>'
        )

    def handle_output(self, handler: HandlerBase) -> None:
        """Do the thing."""
        self._init_inputs()

        for inp in self.inputs:
            inp.handle_output(handler)

        # Call our apply callback after everything else has run.
        if self.on_apply is not None:
            self.on_apply()

    def _init_inputs(self) -> None:
        """Set up input-ids for our inputs."""
        for i, inp in enumerate(self.inputs):
            input_id = f'{INPUT_BASE_ID}{i}'
            inp.input_id = input_id
            self.input_map[input_id] = inp

    @classmethod
    def can_dispatch(cls, forms: list[Form], handler: HandlerBase) -> bool:
        """Return whether one of the provided forms will handle the handler."""
        form_id = handler.param_str(FORM_ID_NAME, doraise=False)
        if form_id is not None:
            forms_by_id = {f.form_id for f in forms}
            return form_id in forms_by_id
        return False

    @classmethod
    def dispatch(cls, forms: list[Form], handler: HandlerBase) -> bool:
        """Dispatch handler data to the form it originated from, if applicable.

        Returns whether the handler was handled by a form.
        """

        form_id = handler.param_str(FORM_ID_NAME, doraise=False)
        if form_id is not None:
            forms_by_id = {f.form_id: f for f in forms}

            # Make sure there's no duplicate IDs.
            if len(forms) != len(forms_by_id):
                raise RuntimeError('Form list contains duplicate ids.')

            if form_id in forms_by_id:
                forms_by_id[form_id].handle_output(handler)
                return True

        return False


class FormInput:
    """Base class for form inputs."""

    def __init__(self) -> None:
        self.input_id: str | None = None

    def emit_html(self, style: Form.Style) -> str:
        """Spit out html for the input."""
        del style  # Unused.
        return ''

    def handle_output(self, handler: HandlerBase) -> None:
        """Fetch and apply resulting data for the input."""


class BoolInput(FormInput):
    """Wraps a bool value."""

    def __init__(
        self,
        getter: Callable[[], bool],
        setter: Callable[[bool], None] | None = None,
        label: str | None = None,
    ):
        super().__init__()
        self.getter = getter
        self.setter = setter
        self.label = label

    @override
    def emit_html(self, style: Form.Style) -> str:
        out = ''
        assert self.input_id is not None
        checkedval = ' checked' if self.getter() else ''
        out += (
            f'<input type="checkbox" style="margin:5px 5px 5px 10px;"'
            f' id="{self.input_id}"'
            f' name="{self.input_id}" value="1"{checkedval}>'
        )
        if self.label is not None:
            label_escaped = html.escape(self.label)
            out += (
                f'<label for="{self.input_id}"'
                f' style="'
                # f'margin-left:5px;'
                f'">'
                f'{label_escaped}'
                f'</label>'
            )
        return out

    @override
    def handle_output(self, handler: HandlerBase) -> None:
        assert self.input_id is not None
        value = handler.param_bool(self.input_id, default=False)
        if self.setter is not None:
            self.setter(value)


class _BaseTextInput(FormInput, Generic[T]):
    """Base class for inputs based on text inputs."""

    def __init__(
        self,
        getter: Callable[[], T],
        setter: Callable[[T], None] | None,
        label: str | None,
    ):
        super().__init__()
        self.getter = getter
        self.setter = setter
        self.label = label

    @override
    def emit_html(self, style: Form.Style) -> str:
        out = ''
        assert self.input_id is not None
        if self.label is not None:
            label_escaped = html.escape(self.label)
            out += (
                f'<label for="{self.input_id}"'
                f' style="margin-left:5px">'
                f'{label_escaped}'
                f'</label>'
            )
        value_escaped = html.escape(self.val_to_str(self.getter()))
        out += (
            f'<input type="text"'
            # f' style="margin:10px 3px"'
            f' id="{self.input_id}"'
            f' name="{self.input_id}" value="{value_escaped}">'
        )
        return out

    def val_to_str(self, value: T) -> str:
        """Convert our value type to str form."""
        raise RuntimeError('This should be overridden.')

    def val_from_str(self, stringvalue: str) -> T:
        """Convert our value type from str form."""
        raise RuntimeError('This should be overridden.')

    @override
    def handle_output(self, handler: HandlerBase) -> None:
        assert self.input_id is not None
        value = handler.param_str(self.input_id)
        if self.setter is not None:
            self.setter(self.val_from_str(value))


class TextInput(_BaseTextInput):
    """Wraps a str value."""

    def __init__(
        self,
        getter: Callable[[], str],
        setter: Callable[[str], None] | None = None,
        label: str | None = None,
    ):
        super().__init__(getter, setter, label)

    @override
    def val_to_str(self, value: str) -> str:
        """Convert our value type to str form."""
        return value

    @override
    def val_from_str(self, stringvalue: str) -> str:
        """Convert our value type from str form."""
        return stringvalue


class FloatInput(_BaseTextInput):
    """Wraps a float value."""

    def __init__(
        self,
        getter: Callable[[], float],
        setter: Callable[[float], None] | None = None,
        label: str | None = None,
    ):
        super().__init__(getter, setter, label)

    @override
    def val_to_str(self, value: float) -> str:
        """Convert our value type to str form."""
        return str(value)

    @override
    def val_from_str(self, stringvalue: str) -> float:
        """Convert our value type from str form."""
        try:
            return float(stringvalue)
        except Exception as exc:
            raise CleanError(
                f'{stringvalue} is not a valid float value.'
            ) from exc


class EnumInput(FormInput, Generic[ET]):
    """Wraps an enum value."""

    def __init__(
        self,
        enum_type: type[ET],
        getter: Callable[[], ET],
        setter: Callable[[ET], None] | None = None,
        label: str | None = None,
        getlabel: Callable[[ET], str] | None = None,
    ):
        super().__init__()
        self.enum_type = enum_type
        self.getter = getter
        self.setter = setter
        self.label = label
        self.getlabel = getlabel

    @override
    def emit_html(self, style: Form.Style) -> str:
        out = ''
        assert self.input_id is not None
        if self.label is not None:
            label_escaped = html.escape(self.label)
            out += (
                f'<label for="{self.input_id}"'
                f' style="margin-left:5px">'
                f'{label_escaped}'
                f'</label>'
            )
        out += (
            f'<select name="{self.input_id}" style="margin:5px"'
            f'id="{self.input_id}">\n'
        )

        # Enum options.
        curvalue = self.getter()
        for val in self.enum_type:
            selstr = ' selected' if val is curvalue else ''
            value = html.escape(json.dumps(val.value))
            label = (
                self.getlabel(val)
                if self.getlabel is not None
                else str(val.value)
            )
            out += f'<option value="{value}"{selstr}>{label}</option>'
        out += '</select>'
        return out

    @override
    def handle_output(self, handler: HandlerBase) -> None:
        assert self.input_id is not None
        rawval = json.loads(handler.param_str(self.input_id))
        value: ET = self.enum_type(rawval)
        if self.setter is not None:
            self.setter(value)


class OptionalEnumInput(FormInput, Generic[ET]):
    """Wraps an optional enum value."""

    def __init__(
        self,
        enum_type: type[ET],
        getter: Callable[[], ET | None],
        setter: Callable[[ET | None], None] | None = None,
        label: str | None = None,
        getlabel: Callable[[ET | None], str] | None = None,
    ):
        super().__init__()
        self.enum_type = enum_type
        self.getter = getter
        self.setter = setter
        self.label = label
        self.getlabel = getlabel

    @override
    def emit_html(self, style: Form.Style) -> str:
        out = ''
        assert self.input_id is not None
        if self.label is not None:
            label_escaped = html.escape(self.label)
            out += (
                f'<label for="{self.input_id}"'
                f' style="margin-left:5px">'
                f'{label_escaped}'
                f'</label>'
            )
        out += (
            f'<select name="{self.input_id}" style="margin:5px"'
            f'id="{self.input_id}">\n'
        )

        curvalue = self.getter()
        # None option.
        selstr = ' selected' if curvalue is None else ''
        value = html.escape(json.dumps(None))
        label = self.getlabel(None) if self.getlabel is not None else 'None'
        out += f'<option value="{value}"{selstr}>{label}</option>'

        # Enum options.
        for val in self.enum_type:
            selstr = ' selected' if val is curvalue else ''
            value = html.escape(json.dumps(val.value))
            label = (
                self.getlabel(val)
                if self.getlabel is not None
                else str(val.value)
            )
            out += f'<option value="{value}"{selstr}>{label}</option>'
        out += '</select>'
        return out

    @override
    def handle_output(self, handler: HandlerBase) -> None:
        assert self.input_id is not None
        rawval = json.loads(handler.param_str(self.input_id))
        value: ET | None = None if rawval is None else self.enum_type(rawval)
        if self.setter is not None:
            self.setter(value)


class Label:
    """Generates display names for a type."""

    def __init__(self, cls: type[T], call: Callable[[T], str]):
        self.cls = cls
        self.call = call


class DataclassForm(Form, Generic[T]):
    """Form wrapping all fields of a single dataclass instance."""

    def __init__(
        self,
        form_id: str,
        getter: Callable[[], T],
        setter: Callable[[T], None],
        labels: list[Label] | None = None,
    ):
        import dataclasses
        from functools import partial

        # We need to introspect the dataclass type we'll be using,
        # so we need to grab a current value immediately. Ideally
        # it would seem we should only call getter() in response to
        # the getters on our sub-inputs, but then we'd need to take
        # the class as an explicit argument or something.

        # Note: Need to make a copy otherwise we'll modify what getter
        # brings in even if we have no setter.
        self._value = copy.deepcopy(getter())
        if not dataclasses.is_dataclass(self._value) or isinstance(
            self._value, type
        ):
            raise TypeError(
                f'Expected a dataclass instance;' f' got a {type(self._value)}'
            )

        cls = type(self._value)

        # NOTE: Now passing the class' __dict__ (vars()) as locals
        # which allows us to pick up nested classes, etc.
        typehints = get_type_hints(cls, localns=vars(cls))  # type: ignore

        self._labelcalls: dict[type, Callable[[Any], str]] = (
            {t.cls: t.call for t in labels} if labels is not None else {}
        )
        self._setter = setter
        inputs: list[FormInput] = []
        for field in dataclasses.fields(self._value):
            fieldtype = typehints[field.name]
            getfieldval = partial(getattr, self._value, field.name)
            setfieldval = partial(setattr, self._value, field.name)
            if fieldtype is float:
                inputs.append(
                    FloatInput(getfieldval, setfieldval, label=field.name)
                )
            elif fieldtype is bool:
                inputs.append(
                    BoolInput(getfieldval, setfieldval, label=field.name)
                )
            elif fieldtype is str:
                inputs.append(
                    TextInput(getfieldval, setfieldval, label=field.name)
                )
            elif issubclass(fieldtype, Enum):
                inputs.append(
                    EnumInput(
                        fieldtype,
                        getfieldval,
                        setfieldval,
                        label=field.name,
                        getlabel=self._labelcalls.get(fieldtype, None),
                    )
                )
            else:
                raise RuntimeError(
                    f'Field {field.name} has unsupported' f' type: {fieldtype}.'
                )

        super().__init__(form_id, inputs, on_apply=self._apply_dataclass)

    def _apply_dataclass(self) -> None:
        self._setter(self._value)
