# Synced from ballistica-master-server.
# EFRO_SYNC_HASH=238341747753035160164867830186809135031
#
"""Create html tables."""

from __future__ import annotations

import html
from typing import TYPE_CHECKING, Generic, TypeVar

T = TypeVar('T')

if TYPE_CHECKING:
    from typing import Callable, Any, Sequence, Iterable


class Table(Generic[T]):
    """Wrapper around an html table."""

    # Special values that callbacks can return
    ROWNUM = '_row_'

    def __init__(
        self,
        rows: Iterable[T],
        columns: list[tuple[str, Callable[[T], Any]]] | None = None,
        footer: Sequence[Any] | None = None,
        escape: bool = True,
    ) -> None:
        if columns is None:
            columns = []
        self.rows = rows
        self.columns = columns
        if footer is not None:
            if len(footer) != len(columns):
                raise ValueError(
                    f'Footer column count ({len(footer)})'
                    f' does not match table column count ({len(columns)})'
                )
        self.footer = footer
        self.escape = escape

    def emit(self) -> str:
        """Emit html code for this table."""
        bits: list[str] = ['<table style="width:100%"><tr class="faded small">']
        for cname, _ccall in self.columns:
            cname2 = html.escape(cname) if self.escape else cname
            bits.append(f'<td>{cname2}</td>')
        bits.append('</tr>')
        for i, row in enumerate(self.rows):
            bits.append('<tr class="controlrow">')
            for _cname, ccall in self.columns:
                rawval = ccall(row)
                sval: str
                if rawval is self.ROWNUM:
                    sval = str(i + 1)
                else:
                    sval = (
                        html.escape(str(rawval)) if self.escape else str(rawval)
                    )
                bits.append(f'<td>{sval}</td>')
            bits.append('</tr>')
        if self.footer:
            bits.append('<tr class="controlrow">')
            for footer_column in self.footer:
                sval = html.escape(str(footer_column))
                bits.append(f'<td>{sval}</td>')
            bits.append('</tr>')
        bits.append('</table>\n')
        return ''.join(bits)
