# Synced from ballistica-master-server.
# EFRO_SYNC_HASH=80193281007798418002839962942567249185
#
"""General utilities for emitting html."""

from __future__ import annotations

import sys
import html
import logging
from enum import Enum
from typing import TYPE_CHECKING

from baserver.hostenv import HostEnv

if TYPE_CHECKING:
    from typing import Any, Sequence

    from flask.wrappers import Response


def get_exception_html(is_admin: bool, home_link: bool = True) -> str:
    """Return html code for the active exception (traceback, etc)."""
    import traceback

    from efro.error import CleanError

    s = '<div style="text-align:center"><h3 class="redtxt">Error</h3></div>\n'

    # For CleanErrors, we simply print the message; otherwise we do a full
    # stack trace (admins) or an opaque message (non-admins).
    exc = sys.exc_info()[1]
    if isinstance(exc, CleanError):
        # Hmm; <pre> looks to code-ish here. But we want to preserve newlines
        # and whatnot..
        excstr = html.escape(str(exc)).replace('\n', '<br>')
        s += f'<div style="text-align:center">{excstr}</div>\n'
    else:
        if is_admin:
            s += '<pre>' + html.escape(traceback.format_exc()) + '</pre>\n'
        else:
            # Log the actual error so we see it.
            # If we ever get multiple admins perhaps we should log in
            # that case too, but currently assuming the admin is me so
            # I'll see it immediately.
            logging.exception(
                'An error occurred resulting in an'
                " 'Internal Error' message to html client.",
                extra=HostEnv.get_default_log_extra(),
            )
            s += (
                f'<div style="text-align:center">An internal error has'
                f' occurred. If this continues, please report it'
                f' to {HostEnv.get_support_email(html=True)}</div>'
            )

    # Give them an escape path if desired. We use the full path so that
    # http errors will get redirected to a https path.
    if home_link:
        root_addr = HostEnv.get().get_root_page_address()
        s += (
            '<br>'
            '<div style="text-align:center">'
            f'<a href="{root_addr}">Home</a>'
            '</div>'
        )

    return s


def _value_for_action(action: Enum | str) -> str:
    if isinstance(action, str):
        return action
    if isinstance(action, Enum):
        assert isinstance(action.value, str)
        return action.value
    # if dataclasses.is_dataclass(action[1]):
    #    return json.dumps(dataclass_to_dict(action[1]), separators=(',', ':'))
    raise ValueError(
        f'Expected an Enum or dataclass for action {action};'
        f' got {type(action[1])}.'
    )


def redirect_timer(url: str, wait_milliseconds: int = 500) -> str:
    """Build javascript code to redirect to another url."""
    return (
        '<script>\n'
        '    var timer = setTimeout(function() {\n'
        f"        window.location='{url}'\n"
        '    }, ' + str(wait_milliseconds) + ');\n'
        '</script>\n'
    )


def form_embed_value(name: str, value: str) -> str:
    """Get html code to embed a value in a form via a hidden input."""
    return f'<input type="hidden" name="{name}" value="{value}">'


def copy_button(text: str) -> str:
    """Copy button for provided static text."""
    from uuid import uuid4

    assert '</script>' not in text
    text = text.replace('"', '\\"')
    btn_id = f'cpbtn-{str(uuid4())}'.replace('-', '_')
    return (
        '<script>\n'
        f'function do_{btn_id}() {{\n'
        f'  let btn = document.getElementById("{btn_id}");\n'
        '  btn.innerText = "Copied!!";\n'
        f'  navigator.clipboard.writeText("{text}");\n'
        '}\n'
        '</script>'
        f'<button id="{btn_id}" type="button" onclick="do_{btn_id}()">'
        'Copy'
        '</button>'
    )


def show_value_pairs(pairs: list[tuple[str, Any]]) -> str:
    """Given key:value pairs, displays them in a lovely table."""
    out = '<table class="log">'
    for pair in pairs:
        name = html.escape(pair[0])
        value = html.escape(str(pair[1]))
        out += (
            '<tr>'
            '<td style="vertical-align: top; width:90px">'
            f'<b>{name}:</b></td><td>'
            f'<pre class="log">{value}</pre>'
            '</td></tr>'
        )
    out += '</table>'
    return out


def inline_javascript_file(
    path: str, replacements: list[tuple[str, str]] | None = None
) -> str:
    """Embed a javascript file in the html, with optional replacements."""
    with open(path, encoding='utf-8') as infile:
        jscode = infile.read()

    # As far as escaping this, I think we just need to watch
    # for </script> showing up somewhere.
    assert '</script>' not in jscode.lower()

    if replacements is not None:
        for src, dst in replacements:
            assert src in jscode, f'Replacement "{src}" not found in {path}'
            jscode = jscode.replace(src, dst)

    return f'<script>\n{jscode}\n</script>'


def format_log(logtxt: str, title: str | None, truncated_lines: int) -> str:
    """Given lines from a structured log file, returns pretty html."""
    # pylint: disable=too-many-locals
    from efro.dataclassio import dataclass_from_json
    from efro.util import utc_now, timedelta_str
    from efro.log import LogEntry, LogLevel

    out = ''
    loglevelcodes = {
        LogLevel.DEBUG: ' class="bluetxt"',
        LogLevel.WARNING: ' class="orangetxt"',
        LogLevel.ERROR: ' class="redtxt"',
        LogLevel.CRITICAL: ' class="darkredtxt"',
    }
    now = utc_now()
    if title is not None:
        out += (
            '<table class="toolbar" style="width:100%">'
            '<tr class="toolbar"><td>\n'
        )
        out += f'<h2>{title}</h2>'
        out += '</td></tr></table>\n'
    lines = logtxt.splitlines()
    if truncated_lines:
        total_lines = len(lines) + truncated_lines
        out += (
            f'<div style="text-align:center" class="small faded">'
            f'({truncated_lines}'
            f' of {total_lines} lines truncated)</div>'
        )
    out += '<table class="log">\n'
    for i, line in enumerate(lines):
        try:
            entry = dataclass_from_json(LogEntry, line)
            timestr = timedelta_str(now - entry.time, maxparts=3, decimals=3)

            out += (
                f'<tr class="log"><td class="logtime"'
                f' style="text-align: center;'
                f' vertical-align: center; width:80px">'
                f'<span class="faded small">{timestr}</span></td>'
            )
            llcode = loglevelcodes.get(entry.level, '')
            message = html.escape(entry.message)
            out += (
                '<td>'
                '<table><tr><td>'
                f'<span class="small faded">{entry.name}</span></td>'
                f'<td><pre class="log"><span{llcode}>{message}'
                '</span></pre></td>'
                '</tr></table>'
                '</td>\n'
            )
        except Exception:
            logging.error(
                'Error parsing log line %s: %s',
                i,
                line,
                extra=HostEnv.get_default_log_extra(),
            )
            llcode = loglevelcodes[LogLevel.ERROR]
            out += (
                f'<td><span{llcode}>ERROR PARSING LINE'
                f' - SEE LOG.</span></td>\n'
            )
        out += '</tr>'
    out += '</table>\n'
    return out
