# Synced from ballistica-master-server.
# EFRO_SYNC_HASH=140448152394913879929145797493518888811
#
"""Page rendering stuff."""

from __future__ import annotations

import time
import datetime
import weakref
from enum import Enum
from dataclasses import dataclass
from typing import TYPE_CHECKING, overload, TypeVar, final

from typing_extensions import override
from flask import request, make_response, redirect
from efro.error import CleanError
from baserver.hostenv import HostEnv


if TYPE_CHECKING:
    from typing import Any, Literal

    from flask.wrappers import Response
    from baserver.request import ProvidesRequestAuthentication

ET = TypeVar('ET', bound=Enum)


def _is_phone() -> bool:

    # Turns out the user_agents stuff is very slow (like half a second
    # on an app-engine instance to do a parse). Ew. Will just use simple
    # checking for now and should migrate to elegant client-side
    # responsiveness over time.

    #     uastring = request.headers.get('User-Agent', '')
    #     assert uastring is not None
    #     try:
    #         from user_agents import parse

    #         user_agent = parse(uastring)
    #         out = user_agent.is_mobile
    #         assert isinstance(out, bool)
    #         return out
    #     except Exception:
    #         logging.exception('Error parsing with user-agents module.')

    #         # Currently just looking for apple gear; should make this more
    #         # robust.
    #         return 'iPhone;' in uastring or 'iPad;' in uastring

    user_agent = request.headers.get('User-Agent', '')
    user_agent = user_agent.lower()
    phones = ['android', 'iphone']

    if any(x in user_agent for x in phones):
        return True
    return False


class HandlerBase:
    """Common functionality shared by GetHandler and PostHandlers."""

    def param_raw(self, name: str) -> Any:
        """Return request args or form args depending on what we are."""
        raise RuntimeError('This should be overridden.')

    # When passed doraise=True we always return a str.
    @overload
    def param_str(
        self,
        name: str,
        default: str | None = None,
        doraise: Literal[True] = True,
    ) -> str: ...

    # When passed doraise=False we might return None.
    @overload
    def param_str(
        self,
        name: str,
        default: str | None = None,
        doraise: Literal[False] = False,
    ) -> str | None: ...

    # Actual Implementation.
    def param_str(
        self, name: str, default: str | None = None, doraise: bool = True
    ) -> str | None:
        """Get a request arg as a string.

        If no default is provided, will error on missing params
        or return None if doraise is False.
        """
        val = self.param_raw(name)
        assert isinstance(val, (str, type(None)))
        if val is None:
            if default is None:
                if doraise:
                    raise CleanError(
                        f"Required parameter '{name}' not provided."
                    )
                return None
            return default
        return val

    # When passed doraise=True we always return a value.
    @overload
    def param_enum(
        self,
        name: str,
        enum_type: type[ET],
        default: ET | None = None,
        doraise: Literal[True] = True,
    ) -> ET: ...

    # When passed doraise=False we might return None.
    @overload
    def param_enum(
        self,
        name: str,
        enum_type: type[ET],
        default: ET | None = None,
        doraise: Literal[False] = False,
    ) -> ET | None: ...

    # Actual Implementation.
    def param_enum(
        self,
        name: str,
        enum_type: type[ET],
        default: ET | None = None,
        doraise: bool = True,
    ) -> ET | None:
        """Get a request arg as an Enum.

        If no default is provided, will error on missing params
        or return None if doraise is False.
        """
        val = self.param_raw(name)
        assert isinstance(val, (str, type(None)))
        if val is None:
            if default is None:
                if doraise:
                    raise CleanError(
                        f"Required parameter '{name}' not provided."
                    )
                return None
            return default
        try:
            if all(isinstance(val.value, int) for val in enum_type):
                return enum_type(int(val))
            if all(isinstance(val.value, str) for val in enum_type):
                return enum_type(val)
            raise RuntimeError('Found a mixture of enum values.')
        except Exception as exc:
            raise CleanError(
                f'{val} is not a valid {enum_type.__name__}'
            ) from exc

    def param_float(self, name: str, default: float | None) -> float:
        """Get a request arg as a float.

        If no default is provided, will error on missing params.
        """
        arg = self.param_raw(name)
        assert isinstance(arg, (str, type(None)))
        if arg is None:
            if default is None:
                raise CleanError(f"Required parameter '{name}' not provided.")
            return default
        try:
            val = float(arg)
        except ValueError as exc:
            raise CleanError(f"Invalid float param value: '{arg}'") from exc
        return val

    def param_int(self, name: str, default: int | None) -> int:
        """Get a request arg as an int.

        If no default is provided, will error on missing params.
        """
        arg = self.param_raw(name)
        assert isinstance(arg, (str, type(None)))
        if arg is None:
            if default is None:
                raise CleanError(f"Required parameter '{name}' not provided.")
            return default
        try:
            val = int(arg)
        except ValueError as exc:
            raise CleanError(f"Invalid int param value: '{arg}'") from exc
        return val

    def param_bool(self, name: str, default: bool | None) -> bool:
        """Get a request param as a bool.

        If no default is provided, will error on missing params."""
        arg = self.param_raw(name)
        assert isinstance(arg, (str, type(None)))
        if arg is None:
            if default is None:
                raise CleanError(f"Required parameter '{name}' not provided.")
            return default

        # Let's support for a preset list of true/false values.
        if arg in {'True', 'true', '1', 'on'}:
            return True
        if arg in {'False', 'false', '0', 'off'}:
            return False
        raise CleanError(f"Invalid bool param value: '{arg}'")


@dataclass
class PostResponse:
    """Defines the response to a post."""

    center: bool = True
    auto_return: bool = True
    auto_return_wait: int = 500
    hide_ok: bool = False
    output: str = ''
    return_url: str = ''
    wait_for_servernode: str | None = None


class PostHandler(HandlerBase):
    """Handles POST requests.

    Posts generally display a confirmation or some other bit of info
    for a short moment and then automatically return to the GET handler.
    """

    def __init__(self, page: Page) -> None:
        # Are we serving to a phone?
        self.is_phone = _is_phone()

        # Max width of our top level div. Pages showing lots of
        # information can increase this or set it to None.
        self.max_width: float | None = 600

        self._page = weakref.ref(page)
        self.response = PostResponse()

    @property
    def page(self) -> Page:
        """The page we are associated with."""
        out = self._page()
        if out is None:
            raise RuntimeError('No page currently set.')
        return out

    @override
    def param_raw(self, name: str) -> Any:
        return request.form.get(name)

    def post(self) -> None:
        """Do the actual handling (should be overridden by child classes)."""

    @final
    def call_post(self) -> str:
        """Set up and call our post func."""
        from baserver.html.utils import redirect_timer

        try:
            self.post()
        except Exception:
            from baserver.html.utils import get_exception_html

            self.response.center = False
            self.response.auto_return = False
            self.response.output = get_exception_html(
                is_admin=self.page.auth.is_admin, home_link=False
            )
            # Allow stack traces/etc to show up in their full glory.
            self.max_width = None

        if self.response.auto_return:
            # Can go with either html meta refresh here or javascript.
            # Update: it seems that the meta route only supports whole
            # int seconds, so sticking with javascript for now.
            if bool(False):
                self.page.refresh_url = self.response.return_url
                self.page.refresh_seconds = int(
                    self.response.auto_return_wait / 1000.0
                )
            else:
                self.response.output += redirect_timer(
                    self.response.return_url, self.response.auto_return_wait
                )
        else:
            if self.response.hide_ok:
                pass
            else:
                self.response.output += (
                    f'<div style="text-align:center">'
                    f'<br><a href="{self.response.return_url}">okiedokie</a>'
                    f'</div>'
                )

        aln = 'center' if self.response.center else 'left'
        pad = 30 if self.response.center else 10
        self.response.output = (
            f'<div style="text-align:{aln};'
            f' padding:{pad}px;">{self.response.output}</div>'
        )

        return self.response.output


class GetHandler(HandlerBase):
    """Handles GET requests."""

    def __init__(self, page: Page) -> None:

        self._page = weakref.ref(page)
        self.output: str = ''

        # Get-handler can override title specified in Page.
        self.title: str | None = None

        # Are we serving to a phone?
        self.is_phone = _is_phone()

        # Max width of our top level div. Pages showing lots of
        # information can increase this or set it to None.
        self.max_width: float | None = 600

        # Draw debug stuff at bottom.
        self.show_footer = True

        # Messages can be subtly printed at screen bottom.
        # (timing info, etc)
        self.footer_notes: list[str] = []

        self.get_start_time: float | None = None

    @property
    def page(self) -> Page:
        """The page we are associated with."""
        out = self._page()
        if out is None:
            raise RuntimeError('No page currently set.')
        return out

    @override
    def param_raw(self, name: str) -> Any:
        return request.args.get(name)

    def get(self) -> None:
        """Do the actual handling. Should be overridden by child classes."""

    @final
    def call_get(self) -> str:
        """Set up and call our get func."""
        import html

        self.get_start_time = time.monotonic()
        try:
            self.get()
        except Exception:
            from baserver.html.utils import get_exception_html

            self.output = get_exception_html(is_admin=self.page.auth.is_admin)

        duration = time.monotonic() - self.get_start_time
        kbytes = len(self.output.encode()) // 1024
        msgs_str = html.escape(
            ', '.join(
                self.footer_notes + [f'total {kbytes}k in {duration:.2f}s']
            )
        )
        if self.show_footer:
            self.output += (
                f'<hr>'
                f'<div class="faded" style="text-align:center">'
                f'{msgs_str}</div>'
            )
        return self.output

    class PageFooterTimer:
        """For timing parts of page loading."""

        def __init__(self, desc: str, handler: GetHandler) -> None:
            self.desc = desc
            self.handler = weakref.ref(handler)
            self.starttime = time.monotonic()
            self.ended = False

        # Support the with statement.
        def __enter__(self) -> Any:
            return self

        def __exit__(self, exc_type: Any, exc_value: Any, exc_tb: Any) -> None:
            self._end()

        def _end(self) -> None:
            if self.ended:
                return
            self.ended = True
            now = time.monotonic()
            handler = self.handler()
            if handler is not None:
                duration_s = f'{now - self.starttime:.2f}'
                assert handler.get_start_time is not None
                endtime_s = f'{now - handler.get_start_time:.2f}'
                # Include end-time if its different from duration
                # (so we can judge delays in when things started).
                handler.footer_notes.append(
                    f'{self.desc}: {duration_s}s'
                    if duration_s == endtime_s
                    else f'{self.desc}: {duration_s}s@{endtime_s}'
                )

        def __del__(self) -> None:
            self._end()

    def footer_timer(self, desc: str) -> PageFooterTimer:
        """Create a footer timer. Shows object lifespan in page footer."""
        timer = self.PageFooterTimer(desc, self)
        return timer


@dataclass
class _CookieSet:
    """Defines a cookie set operation."""

    value: str
    duration: datetime.timedelta | None
    domain: str | None


class Page:
    """Combines handlers into a page."""

    def __init__(
        self,
        gethandler: type[GetHandler] | None = None,
        posthandler: type[PostHandler] | None = None,
        title: str = 'Page',
    ):
        assert posthandler is not None or gethandler is not None
        self._gethandler = gethandler
        self._posthandler = posthandler
        self._title = title
        self._auth: ProvidesRequestAuthentication | None = None
        self._cookie_sets: dict[str, _CookieSet] = {}

        self.redirect_url: str | None = None
        self.explicit_response: Response | None = None
        self.refresh_seconds: int | None = None
        self.refresh_url: str | None = None
        self.head_extras: str = ''

    def set_cookie(
        self,
        name: str,
        value: str,
        duration: datetime.timedelta | None = None,
        domain: str | None = None,
    ) -> None:
        """Set a cookie in the response."""
        self._cookie_sets[name] = _CookieSet(
            value=value, duration=duration, domain=domain
        )

    def delete_cookie(self, name: str, domain: str | None = None) -> None:
        """Delete a cookie in the response."""

        # Seems the official way to do this is just set a dummy
        # value with max age 0.
        self._cookie_sets[name] = _CookieSet(
            value='dummy', duration=datetime.timedelta(seconds=0), domain=domain
        )

    @property
    def auth(self) -> ProvidesRequestAuthentication:
        """Return page authentication manager."""
        if self._auth is None:
            raise RuntimeError(
                'auth is not currently available (only valid during requests)'
            )
        return self._auth

    def access_check(self) -> bool:
        """A blanket access check for the page.

        This can be used to easily restrict overall access to admins or
        specific users. The default implementation restricts to admins.
        """
        return self.auth.is_admin

    def handle_request(self) -> Response:
        """Do the thing."""

        from baserver.html.utils import get_exception_html

        self._auth = HostEnv.get().get_auth()

        viewport_width: float | None

        try:

            # Even though our underlying server may technically allow http
            # traffic, for the pages system we want to make sure its https.
            # We set our cookies as secure so http wouldn't work anyway.
            # Note that in our dev setup this currently requires us to
            # forward X-Forwarded-Proto to our http back-end by adding
            # the following in our apache2.conf:
            # RequestHeader set "X-Forwarded-Proto" expr=%{REQUEST_SCHEME}
            # RequestHeader set "X-Forwarded-SSL" expr=%{HTTPS}
            if not request.is_secure:
                raise CleanError(
                    'This site requires an https connection; not http.'
                )

            if not self.access_check():
                # Note that they're not signed in when that's the case.
                # Often access-denied errors are simply due to accounts
                # signing out automatically; let's not make them scarier
                # than they have to be.
                nli = ' (not signed in)' if not self.auth.signed_in else ''
                raise CleanError(f'Access denied.{nli}')
            if request.method == 'GET':
                assert self._gethandler is not None
                gethandler = self._gethandler(self)
                raw_content = gethandler.call_get()
                viewport_width = None
                max_width = gethandler.max_width
                title_override: str | None = gethandler.title
            elif request.method == 'POST':
                assert self._posthandler is not None
                posthandler = self._posthandler(self)
                raw_content = posthandler.call_post()
                viewport_width = None
                max_width = posthandler.max_width
                title_override = None
            else:
                # We get other stuff occasionally. Just present
                # a nice error message and don't make a fuss.
                raise CleanError(f'Unhandled request method {request.method}.')
        except Exception:
            # Present a bare-bones error page if anything went wrong.
            # It's preferable that the handlers handle errors since they
            # can take more appropriate actions (an 'ok' button for post
            # handlers, etc.).
            raw_content = get_exception_html(is_admin=self._auth.is_admin)
            viewport_width = None
            title_override = None
            max_width = None

        # If the user provided an explicit raw response, use it.
        if self.explicit_response is not None:
            return self.explicit_response

        # If a redirect was requested:
        if self.redirect_url is not None:
            return redirect(self.redirect_url)  # type: ignore

        response = make_response(
            self._filter_content(
                raw_content,
                viewport_width=viewport_width,
                max_width=max_width,
                title_override=title_override,
            )
        )

        for cookiename, cookieset in self._cookie_sets.items():
            response.set_cookie(
                cookiename,
                value=cookieset.value,
                max_age=cookieset.duration,
                domain=cookieset.domain,
                secure=True,
            )

        return response

    def _filter_content(
        self,
        content: str,
        viewport_width: float | None,
        max_width: float | None,
        title_override: str | None = None,
    ) -> str:
        # from bamaster.env import FLEET, Fleet
        from flask import render_template

        # Note to self; apparently user-scalable is no longer honored
        # as of iOS 10 or so, but it does still seem to prevent the site
        # from zooming in when a text field is tapped, which I like.

        # UPDATE: no longer setting content width here; this seems to
        # interfere with the functionality of the view scaling buttons
        # on iOS safari, which are quite handy for pages such as code editors.
        if bool(True):
            viewport_width = None
        viewportmeta = (
            (
                '<meta name="viewport"'
                f' content="width={viewport_width:.0f}, user-scalable=no">'
            )
            if viewport_width is not None
            else (
                '<meta name="viewport"'
                ' content="width=device-width, user-scalable=no">'
            )
        )

        # viewportmeta = ('<meta name="viewport" content="width=device-width,'
        #                 ' user-scalable=no">')

        if max_width is not None:
            content = (
                f'<div style="max-width:{max_width:.0f}px;margin:auto;">'
                f'{content}</div>'
            )

        # Pure html way to refresh with optional address change.
        # The downside to this is it is limited to whole number seconds.
        if self.refresh_seconds is not None:
            if self.refresh_url is not None:
                refreshmeta = (
                    f'<meta http-equiv="refresh"'
                    f' content="{self.refresh_seconds};'
                    f' url={self.refresh_url}">'
                )
            else:
                refreshmeta = (
                    f'<meta http-equiv="refresh"'
                    f' content="{self.refresh_seconds}">'
                )
        else:
            refreshmeta = ''

        title = title_override if title_override is not None else self._title
        if not HostEnv.get().is_prod_fleet():
            fleetname = HostEnv.get().get_fleet_name()
            title += f' (via {fleetname} fleet)'
        output: str = render_template(
            'ba_web_tools.html',
            content=content,
            viewportmeta=viewportmeta,
            refreshmeta=refreshmeta,
            headextras=self.head_extras,
            title=title,
        )
        assert isinstance(output, str)
        return output
