# Synced from ballistica-master-server.
# EFRO_SYNC_HASH=279640937769717160130599069124419438175
#
"""Common components related to cloud hosted games."""

from __future__ import annotations

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    import datetime
    from typing import Callable, Any

    from redis import Redis
    from redis.client import Pipeline

_g_get_redis: Callable[[], Redis] | None = None


def set_redis_provider(call: Callable[[], Redis]) -> None:
    """Set a function to call to get a redis instance.

    This must be called before any functionality here can be used.
    """
    global _g_get_redis  # pylint: disable=global-statement
    if _g_get_redis is not None:
        raise RuntimeError('A redis provider is already set.')
    _g_get_redis = call


def _get_redis() -> Redis:
    if _g_get_redis is None:
        raise RuntimeError('A redis provider call has not been set.')
    return _g_get_redis()


def should_do_timed_action(key: str, interval_seconds: int) -> bool:
    """Atomically test/increment a time value for a periodic action."""

    assert isinstance(interval_seconds, int)

    redis = _get_redis()

    def _update_time_in_transaction(pipe: Pipeline) -> None:
        # These calls happen in immediate mode so we can grab values
        # (the WATCH calls that this wrapper added took care of that).
        existing_data: Any = pipe.get(key)
        assert isinstance(existing_data, (bytes, type(None)))

        # We rely purely on redis expiration; we only run
        # if a value is not there.
        if existing_data is not None:
            return

        # Flip the pipeline into batch-write mode and then write our
        # new value.
        pipe.multi()
        pipe.set(key, b'1', ex=interval_seconds)

    # This will WATCH the given key(s) and then run the given call
    # and tell us if the call went through (which only happens if no
    # watched keys changed in the meantime).
    tresult = redis.transaction(
        _update_time_in_transaction, key
    )  # type: ignore

    # This should give us either an empty list (if we didn't write)
    # or a single result (if we did).
    assert isinstance(tresult, list)
    assert len(tresult) == 0 or (
        len(tresult) == 1 and isinstance(tresult[0], bool)
    )

    # For us to do the action, we need to have set the key
    # AND the transaction needs to have gone through (the bool).
    out = len(tresult) > 0 and tresult[0] is True
    assert isinstance(out, bool)
    return out


# def sharedflag(key: str, exp: datetime.timedelta) -> bool:
#     """Set a flag based on the hash of a provided string.

#     Returns whether the flag was already set. If there is an error
#     talking to redis, returns False.
#     """
#     import hashlib
#     import base64

#     try:
#         _redis = _get_redis()
#         md5 = hashlib.md5()
#         md5.update(key.encode())

#         # Technically we could use the binary hash as a key but let's go
#         # with something small but human readable (base64 with padding
#         # stripped).
#         hashstr = base64.b64encode(md5.digest()).decode().replace('=', '')
#         key = f'sharedflag:{hashstr}'
#         print('FINAL KEY', key)

#         print('WOULD SET WITH EXP', exp)
#         return True

#     except Exception:
#         # Be resilient to anything going wrong here but make sure it is
#         # noticed.
#         logging.warning('Error getting redis for sharedflag.', exc_info=True)
#         return False


class DistributedCounter:
    """A distributed counter for things like counting in-flight tasks.

    Individual ids can be added and removed from the list and will also
    expire after some amount of time in case they are not removed. Then
    the total number of ids in the list can be efficiently queried.
    """

    def __init__(
        self,
        expire: datetime.timedelta,
        update_period: datetime.timedelta,
    ) -> None:
        self.expire = expire
        self.update_period = update_period

    def add(self, itemid: str) -> None:
        """Add an id to the counter."""

    def remove(self, itemid: str) -> None:
        """Remove an item from the counter."""

    def get_count(self) -> int:
        """Get the current item count."""
        return 0


class DistributedRateCounter:
    """A distributed counter for tracking a continuous rate."""

    def __init__(
        self,
        period: datetime.timedelta,
        update_period: datetime.timedelta,
    ) -> None:
        self.period = period
        self.update_period = update_period

    def register_event(self) -> None:
        """Register an event."""

    def get_rate(self) -> float:
        """Get the recent average number of events per period."""
        return 0.0
