# Synced from ballistica-master-server.
# EFRO_SYNC_HASH=117593856308477255092392573372225892459
#
"""Defines workspace behavior."""

from __future__ import annotations

import datetime
from enum import Enum
from typing import TYPE_CHECKING, assert_never, Annotated
from dataclasses import dataclass, field

from typing_extensions import override
from efro.dataclassio import (
    IOMultiType,
    ioprep,
    ioprepped,
    will_ioprep,
    IOAttrs,
)
from baserver.cloudbuild import (
    RecipeComponent,
    RecipeComponentTypeID,
    Recipe,
    RecipeTypeID,
)

if TYPE_CHECKING:
    pass

# Incrementing this will cause all workspace-compiles to result in new
# unique builds. This will effectively resolve permanent failures
# without waiting for the errored build to expire.
COMPILE_RECIPE_VERSION = 3


class WorkspaceTypeID(Enum):
    """Types a workspace can be."""

    UNTYPED = 'untyped'
    ASSETS_V1 = 'assets_v1'


class WorkspaceType(IOMultiType[WorkspaceTypeID]):
    """Top level class for our workspace types."""

    @override
    @classmethod
    def get_type_id(cls) -> WorkspaceTypeID:
        # Require child classes to supply this themselves. If we
        # did a full type registry/lookup here it would require us
        # to import everything and would prevent lazy loading.
        raise NotImplementedError()

    @override
    @classmethod
    def get_type(cls, type_id: WorkspaceTypeID) -> type[WorkspaceType]:
        """Return the subclass for each of our type-ids."""
        # pylint: disable=cyclic-import
        out: type[WorkspaceType]

        t = WorkspaceTypeID
        if type_id is t.UNTYPED:
            from baserver.workspace.untyped import UntypedWorkspace

            out = UntypedWorkspace
        elif type_id is t.ASSETS_V1:
            from baserver.workspace.assetsv1 import AssetsV1Workspace

            out = AssetsV1Workspace
        else:
            # Important to make sure we provide all types.
            assert_never(type_id)
        return out


@will_ioprep
@dataclass
class WorkspaceDirEntry:
    """Something in a workspace directory."""

    class Type(Enum):
        """Type for a WorkspaceDirEntry."""

        FILE = 'f'
        DIRECTORY = 'd'

    type: Annotated[Type, IOAttrs('t')]

    # Only valid for DIRECTORY type.
    children: Annotated[
        dict[str, WorkspaceDirEntry], IOAttrs('c', store_default=False)
    ] = field(default_factory=dict)

    # Only valid for FILE type.
    cloud_file_id: Annotated[str | None, IOAttrs('f', store_default=False)] = (
        None
    )

    # Mod times from the server perspective (currently FILE only).
    modtime: Annotated[
        datetime.datetime | None, IOAttrs('m', store_default=False)
    ] = None

    def get_file_size(self) -> int:
        """Size of the file we point to."""
        from baserver.cloudfile import size_from_cloud_file_id

        if self.cloud_file_id is None:
            raise RuntimeError('Not a file.')
        return size_from_cloud_file_id(self.cloud_file_id)


# Have to explicitly prep this fella *after* he is fully declared since
# he refers to his own type.
ioprep(WorkspaceDirEntry)


@ioprepped
@dataclass
class WorkspaceCompileInput(RecipeComponent):
    """Standard inputs to a workspace compile."""

    root: Annotated[WorkspaceDirEntry, IOAttrs('root')]

    compile_recipe_version: Annotated[int, IOAttrs('rv')] = (
        COMPILE_RECIPE_VERSION
    )

    @override
    @classmethod
    def get_type_id(cls) -> RecipeComponentTypeID:
        return RecipeComponentTypeID.WORKSPACE_COMPILE_INPUT

    @override
    def get_cloud_file_ids(self) -> set[str]:

        # Currently the only file we need is workspace.toml (if present).
        print('FIXME LOOK FOR WORKSPACE.TOML')
        return set()


@ioprepped
@dataclass
class WorkspaceCompileOutput(RecipeComponent):
    """The results of successfully compiling a workspace."""

    workspace_type: Annotated[WorkspaceType, IOAttrs('wtp')]

    @override
    @classmethod
    def get_type_id(cls) -> RecipeComponentTypeID:
        return RecipeComponentTypeID.WORKSPACE_COMPILE_OUTPUT

    @override
    def get_cloud_file_ids(self) -> set[str]:
        # We only spit out data; no files.
        return set()


class WorkspaceCompile(Recipe[WorkspaceCompileInput, WorkspaceCompileOutput]):
    """Recipe to compile a workspace."""

    @override
    @classmethod
    def get_type_id(cls) -> RecipeTypeID:
        return RecipeTypeID.WORKSPACE_COMPILE

    @override
    def build(self, in_data: WorkspaceCompileInput) -> WorkspaceCompileOutput:
        from baserver.workspace.untyped import UntypedWorkspace

        print(f'COMPILING WORKSPACE VERSION {in_data.compile_recipe_version}.')
        return WorkspaceCompileOutput(workspace_type=UntypedWorkspace())
