"""ServerNode base App class functionality."""

# pylint: disable=too-many-lines

from __future__ import annotations

import os
import sys
import ssl
import time
import copy
import random
import weakref
import asyncio
import datetime
import tracemalloc
from pathlib import Path
from base64 import urlsafe_b64decode, urlsafe_b64encode
from enum import Enum, unique
from typing import TYPE_CHECKING

import aiohttp.web
import psutil

from efro.dataclassio import dataclass_from_json, dataclass_to_json
from efro.message import BoolResponse
from efro.error import (
    is_asyncio_streams_communication_error,
    CommunicationError,
    CleanError,
)
from efro.util import DirtyBit, utc_now
from efro.call import tpartial
from efro.log import LogLevel
from baserver.data import MESSAGE_SIDECAR_TIMESTAMP_KEY
from baserver.cryptosimple import (
    str_to_wire_key_plaintext,
    str_from_wire_key_plaintext,
)
from baserver.data import BASNStats, BASNStatsSparse
from baserver.enums import ServerNodeTaskState
from baserver.msg import batocloud
from baserver.msg import bamastertobasn
from baserver.msg import basntobamaster
from baserver.msg import basntobamasterboot
from baserver.util import run, generate_self_signed_certificate

from basn.mgen.bamastersender import BAMasterMessageSender
from basn.mgen.bamastersenderboot import BAMasterMessageSenderBoot
from basn.mgen.bamasterreceiver import BAMasterMessageReceiver
from basn.taskset import TaskSet, TaskNotFoundError
from basn.logging import (
    logextra,
    loghandler,
    transportlog,
    balog,
    hostedgamelog,
)

if TYPE_CHECKING:
    from typing import Any

    from efro.message import Message, Response
    from efro.log import LogEntry

    from basn.taskwrapper.transportagent import TransportAgentTaskWrapper


@unique
class State(Enum):
    """Overall states the app can be in."""

    STARTING = 0
    RUNNING = 1
    SHUTTING_DOWN = 2
    SHUTDOWN_COMPLETE = 3


class App:
    """Main class for app functionality."""

    # pylint: disable=too-many-public-methods

    msg_out_boot = BAMasterMessageSenderBoot()
    msg_out = BAMasterMessageSender()
    msg_in = BAMasterMessageReceiver()

    def __init__(self, config_path: str) -> None:
        # pylint: disable=too-many-statements
        from basn.env import BASN_VERSION, config, DEFAULT_BALLISTICAKIT_VERSION
        import urllib.request

        # Importing this here when spinning up avoids a hitch later
        # which can lead to asyncio timing warnings.
        import basn.taskwrapper.transportagent as _unused

        self.httpclient: aiohttp.ClientSession | None = None
        self._loop: asyncio.AbstractEventLoop | None = None

        self._tasks: list[asyncio.Task] = []

        self._early_logs: list[LogEntry] = []
        self._app_start_time = time.monotonic()

        self._config_path = config_path
        assert os.path.exists(self._config_path)

        # For querying info about ourself.
        self._process = psutil.Process()
        self._last_swap_io_total: int = 0

        # Keep track of whether we're dangerously close to exhausting
        # system memory; we'll avoid launching more work/etc. in that
        # case.
        self._mem_danger = False
        self._mem_danger_level: float | None = None
        self._sockets_danger = False

        self._stats_push_fail_count = 0
        self._stats_sparse_push_fail_count = 0
        self._ship_logs_fail_count = 0
        self._ship_log_index = 0

        self._fuzzy_comm_error_start_time = 0.0
        self._fuzzy_comm_error_count = 0

        self._next_minute_upkeep_time = time.time()

        # Our current stats, the last ones we pushed to master, and the
        # flag for whether we should consider pushing an update.
        self.stats = BASNStats()
        self._stats_master = BASNStats()
        self._stats_to_master = DirtyBit(
            dirty=False, use_lock=True, min_update_interval=0.5
        )

        loghandler.add_callback(self.log_callback)

        balog.info(
            (
                'basn v%s starting up in %s mode...\n'
                'Python %s\n'
                'Python optimize is %s.\n'
                'Python development-mode is %s.\n'
                'tracemalloc is %s.'
            ),
            BASN_VERSION,
            config.run_mode.name,
            sys.version,
            sys.flags.optimize,
            'ENABLED' if sys.flags.dev_mode else 'DISABLED',
            'ENABLED' if tracemalloc.is_tracing() else 'DISABLED',
            extra=logextra(),
        )

        self._state = State.STARTING
        self.tasks = TaskSet()

        # Transport tasks indexed by token.
        self.transport_tasks_by_uuid: dict[
            str, weakref.ref[TransportAgentTaskWrapper]
        ] = {}

        # Keep track of transport-task tokens that never receive
        # connections (purely for debugging).
        self.unclaimed_transport_task_uuids: dict[str, datetime.datetime] = {}

        # EC2 magic to get info about ourself.

        # Note: we can pull either the public hostname for this guy or
        # just grab its ip directly. According to the docs, they should
        # be the same for access via the internet. So let's just go with
        # the addr so we can save ourself dns lookups.
        # https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/
        # using-instance-addressing.html#concepts-public-addresses
        if bool(True):
            addrkey = 'public-ipv4'
        else:
            addrkey = 'public-hostname'

        with urllib.request.urlopen(
            urllib.request.Request(
                f'http://169.254.169.254/latest/meta-data/{addrkey}'
            )
        ) as url:
            self.our_address = url.read().decode()
        assert self.our_address
        balog.info(
            "Our address is '%s'.",
            self.our_address,
            extra=logextra(),
        )

        # Where we log to if running an update/restart.
        self.update_log_path = Path(
            Path(config.log_path).parent, 'basn_update.log'
        )

        # Dump some system log stuff so we can tune bootstrapping.
        self._log_bootstrapping()

        # Make sure the OS won't assign ports we're planning to use.
        self._check_port_range()

        # Set up our variable data dir (clearing out any existing).
        # Note: we need sudo on the remove just in case we're restarting
        # and there's leftover task dirs owned by the baXXX users.
        self.var_root: str = config.var_root
        assert isinstance(self.var_root, str)
        run(f'sudo rm -rf {self.var_root} && mkdir -p {self.var_root}')

        # Set up our tasks root within it.
        self.tasks_root: str = os.path.join(config.var_root, 'tasks')
        os.mkdir(self.tasks_root)

        # Set up our cert dir and create a self-signed certificate there.
        self.cert_root: str = os.path.join(config.var_root, 'cert')
        self.cert_key_path = os.path.join(self.cert_root, 'key.pem')
        self.cert_path = os.path.join(self.cert_root, 'cert.pem')
        os.mkdir(self.cert_root)
        if not os.path.exists(self.cert_key_path) or not os.path.exists(
            self.cert_path
        ):
            generate_self_signed_certificate(
                key_path=self.cert_key_path,
                cert_path=self.cert_path,
                ipaddress=self.our_address,
            )

        # Our SSL Context we use for any hosting we do. This uses our
        # custom cert stuff.
        cstarttm = time.monotonic()
        self._host_ssl_context = ssl.create_default_context(
            ssl.Purpose.CLIENT_AUTH
        )
        self._host_ssl_context.load_cert_chain(
            certfile=self.cert_path, keyfile=self.cert_key_path
        )
        balog.info(
            'Loaded ssl certs in %.2fs.',
            time.monotonic() - cstarttm,
            extra=logextra(),
        )

        # Also set up a regular SSL context we can use for any client
        # connections we make (just for efficienty.
        self._default_ssl_context = ssl.create_default_context()

        # Set up our resources dir and get some standard resources loading.
        self.resources_root: str = os.path.join(config.var_root, 'resources')
        run(
            f'rm -rf {self.resources_root} '
            f'&& mkdir -p {self.resources_root}/.locks'
        )

        # Generate a random 128 bit communication key
        # (which we store in url-safe base64 form).
        self._comm_token = urlsafe_b64encode(os.urandom(16)).decode()

        # Whether we've communicated it to the other end.
        self._comm_setup = DirtyBit(dirty=True, use_lock=True)

        self.aioapp = aiohttp.web.Application()
        self.aioapp.add_routes(
            [
                aiohttp.web.get('/', self.handle_get),
                aiohttp.web.post('/', self.handle_post),
            ]
        )

        # Issue warnings if we're missing a few optional modules (wait
        # until here so our logging machinery picks this up cleanly)
        self._optional_module_checks()

        # Let master know a few things about us immediately.
        self.stats.oldest_supported_ballistica_build = (
            DEFAULT_BALLISTICAKIT_VERSION
        )
        self.stats.run_mode = config.run_mode
        self._do_pstat_update()
        self.mark_stats_dirty()

    def postinit(self) -> None:
        """Run some init that should not happen until are are set as 'app'."""
        self._populate_default_resources()

    def run(self) -> None:
        """Run our main loop."""
        from basn.env import config

        self.aioapp.on_startup.append(self.startup)
        self.aioapp.on_shutdown.append(self.shutdown)

        aiohttp.web.run_app(
            self.aioapp, port=config.port, ssl_context=self._host_ssl_context
        )

    @property
    def default_ssl_context(self) -> ssl.SSLContext:
        """Our defaut ssl context for outgoing client calls."""
        assert self._default_ssl_context is not None
        return self._default_ssl_context

    @property
    def comm_token(self) -> str:
        """Our lovely comm token!"""
        assert self._comm_token is not None
        return self._comm_token

    def _log_bootstrapping(self) -> None:
        try:
            import re

            # If it seems that we've run an update/reboot cycle, show
            # the last log from that. Otherwise go with the initial
            # system setup log.
            logpath = (
                str(self.update_log_path)
                if self.update_log_path.exists()
                else '/var/log/cloud-init-output.log'
            )
            with open(logpath, 'r', encoding='utf-8') as infile:
                for line in infile.read().splitlines():
                    if 'BASN_BOOTSTRAP:' in line or 'BASN.INSTALL:' in line:
                        # Get rid of color control chars.
                        # noinspection RegExpSimplifiable
                        line = re.sub(
                            r'\x1b\[([0-9,A-Z]{1,2}(;[0-9]{1,2})?(;'
                            r'[0-9]{3})?)?[m|K]?',
                            '',
                            line,
                        )
                        balog.info(
                            line,
                            extra=logextra(),
                        )
        except Exception:
            balog.exception(
                'Error logging bootstrapping.',
                extra=logextra(),
            )

    def _check_port_range(self) -> None:
        try:
            from basn.env import config

            # Ask the OS which ports it will assign to sockets randomly.
            with open(
                '/proc/sys/net/ipv4/ip_local_port_range', 'r', encoding='utf-8'
            ) as infile:
                data = infile.read()
            minport, maxport = (int(x) for x in data.split())

            # Make sure this range is safely above any that we might use.
            if minport <= max(
                config.udp_port_max, config.port, config.transport_port
            ):
                balog.error(
                    'OS random-use port range %d-%d is not above our'
                    ' used ranges (%d-%d, %d, %d). Please fix this to'
                    ' avoid conflicts.',
                    minport,
                    maxport,
                    config.udp_port_min,
                    config.udp_port_max,
                    config.port,
                    config.transport_port,
                    extra=logextra(),
                )

        except Exception:
            balog.exception(
                'Error checking port range.',
                extra=logextra(),
            )

    def _optional_module_checks(self) -> None:
        """Look for some modules that are recommended by aiohttp.

        See https://docs.aiohttp.org/en/stable/
        """
        # pylint: disable=unused-import

        try:
            import aiodns
        except Exception:
            balog.warning(
                'Unable to import aiodns;'
                ' docs.aiohttp.org recommends installing this.',
                extra=logextra(),
            )

    def _populate_default_resources(self) -> None:
        from threading import Thread

        from basn.resource.standardtasklibs import StandardTaskLibs
        from basn.resource.serverpackage import ServerPackage
        from basn.env import DEFAULT_BALLISTICAKIT_VERSION

        # Create our set of standard modules that all tasks use.
        Thread(target=lambda: StandardTaskLibs().prepare()).start()

        # Grab the default version of the server that we'll probably
        # use.
        Thread(
            target=lambda: ServerPackage(
                flavor=ServerPackage.Flavor.VANILLA,
                build_number=DEFAULT_BALLISTICAKIT_VERSION,
            ).prepare()
        ).start()

    @property
    def mem_danger(self) -> bool:
        """Are we using dangerously too much memory (the entire machine)?"""
        return self._mem_danger

    @property
    def sockets_danger(self) -> bool:
        """Are we using dangerously too many sockets?"""
        return self._sockets_danger

    @property
    def state(self) -> State:
        """Read-only property to access app state."""
        return self._state

    @staticmethod
    def message_send_timestamp(message: Message) -> float:
        """Given an incoming message, return when it was sent.

        Note that this assumes sender's time is in sync with ours.
        """
        val = getattr(message, MESSAGE_SIDECAR_TIMESTAMP_KEY)
        assert isinstance(val, float)
        return val

    def mark_stats_dirty(self) -> None:
        """Note that stats has been touched and may need to be shipped off."""
        self._stats_to_master.dirty = True

    def on_comm_key_set(self) -> None:
        """Called after we successfully deliver our comm key to bamaster.

        This is a good 'calm before the storm' moment before messages
        start coming in (though technically they can start doing so
        before this gets called).
        """

        self._get_initial_memory_usage()
        self._start_shipping_logs_to_master()

    def _start_shipping_logs_to_master(self) -> None:
        assert self._loop is not None
        self._tasks.append(
            self._loop.create_task(self._ship_logs(), name='basn ship logs')
        )

    def _should_ship_log_entry(self, entry: LogEntry) -> bool:
        # For now let's ignore asyncio 'took-too-long' warnings that we
        # get randomly on nodes in test mode. They're handy to see on
        # the basn node itself but they kinda pollute google cloud
        # logging. If I find that it's possible to address these in some
        # way can revisit.
        if (
            entry.level is LogLevel.WARNING
            and entry.name == 'asyncio'
            and entry.message.startswith('Executing')
            and entry.message.endswith('seconds')
        ):
            return False
        return True

    async def _ship_logs(self) -> None:
        assert self._loop is not None
        while True:
            try:
                # Just in case constructing this archive takes a bit of time,
                # run it in an executor.
                archive = await self._loop.run_in_executor(
                    None,
                    lambda: loghandler.get_cached(
                        start_index=self._ship_log_index
                    ),
                )

                # If we get results starting later than we asked for, it
                # means some got expelled from the cache before we could
                # ship them. Make some noise in that case.
                if archive.start_index != self._ship_log_index:
                    balog.warning(
                        '_ship_logs: asked for logs starting at index %d'
                        ' but results start at index %d;'
                        ' some logs will not show up on bamaster.',
                        self._ship_log_index,
                        archive.start_index,
                        extra=logextra(),
                    )

                if archive.entries:
                    await self.msg_out.send_async(
                        basntobamaster.LogMessage(
                            log_entries=[
                                e
                                for e in archive.entries
                                if self._should_ship_log_entry(e)
                            ]
                        )
                    )
                self._ship_log_index = archive.start_index + len(
                    archive.entries
                )
            except CommunicationError:
                # Occasional comm-errors are expected; only escalate to
                # warnings if it keeps occurring.
                self._ship_logs_fail_count += 1
                logcall = (
                    balog.warning
                    if self._ship_logs_fail_count > 4
                    else balog.info
                )
                logcall(
                    'Pushing logs to master comm-errored (%d consecutive).',
                    self._ship_logs_fail_count,
                    extra=logextra(),
                )
            except Exception:
                # Make more fuss for anything else.
                balog.exception(
                    'Error shipping logs to master.',
                    extra=logextra(),
                )

            # Jitter this a bit to spread out load spikes on bamaster.
            await asyncio.sleep(random.uniform(8.0, 12.0))

    def _get_initial_memory_usage(self) -> None:
        try:
            bytes_per_meg = 1024 * 1024

            vmem = psutil.virtual_memory()
            mtotalraw: int = vmem.total
            mavailraw: int = vmem.available

            system_memory_total = mtotalraw / bytes_per_meg
            system_memory_used = (mtotalraw - mavailraw) / bytes_per_meg

            balog.info(
                'Initial process memory usage is %.1f mb.\n'
                'Initial system memory usage is %.1f mb of %.1f mb.',
                self._process.memory_info().rss / bytes_per_meg,
                system_memory_used,
                system_memory_total,
                extra=logextra(),
            )

            # Ship this home too.
            self.stats.memory_idle = system_memory_used
            self.mark_stats_dirty()

        except Exception:
            balog.exception(
                'Error checking initial memory usage.', extra=logextra()
            )

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        """Our main asyncio event loop."""
        if self._loop is None:
            raise RuntimeError('loop not yet created.')
        return self._loop

    async def startup(self, aioapp: aiohttp.web.Application) -> None:
        """Run startup stuff before we start serving."""
        del aioapp  # Unused.

        # Spin up a client so we can send out to bamaster. We want short
        # timeouts for this; would rather fail fast than hang.
        timeout = aiohttp.ClientTimeout(total=30)
        self.httpclient = aiohttp.ClientSession(timeout=timeout)

        assert self._loop is None
        self._loop = loop = asyncio.get_running_loop()

        assert loop.get_exception_handler() is None
        loop.set_exception_handler(self._handle_asyncio_exception)

        # Get any early logs processing.
        for lentry in self._early_logs:
            self.log_callback(lentry)
        self._early_logs.clear()

        # Start a task to handle periodic updates.
        self._tasks.append(
            loop.create_task(
                self.periodic_update(), name='basn periodic update'
            )
        )

        # Start a task to wrangle transport connections.
        self._tasks.append(
            loop.create_task(
                self.handle_transport_connections(),
                name='basn transport connection listener',
            )
        )

        # Start a task to ship sparse-stats to bamaster occasionally.
        self._tasks.append(
            loop.create_task(
                self.handle_stats_sparse(), name='basn stats sparse'
            )
        )

        balog.info(
            'basn startup complete.',
            extra=logextra(),
        )
        self._state = State.RUNNING

        # Stuff that wants to know when this happens..
        self.tasks.app_did_start_running()

    async def shutdown(self, aioapp: aiohttp.web.Application) -> None:
        """Run cleanup stuff as we shut down."""
        del aioapp  # Unused.
        balog.info(
            'basn shutting down...',
            extra=logextra(),
        )
        if self.httpclient is not None:
            await self.httpclient.close()
        else:
            balog.warning(
                'No httpclient on shutdown.',
                extra=logextra(),
            )
        balog.info(
            'basn shutdown complete.',
            extra=logextra(),
        )

    async def periodic_update(self) -> None:
        """Async task handling general periodic stuff."""
        try:
            while True:
                # If we haven't established communication yet, just do that..
                if self._comm_setup.dirty:
                    if self._comm_setup.should_update:
                        async with self._comm_setup.lock:
                            await self.send_comm_token()
                    continue

                # Do various stuff once per minute.
                if time.time() > self._next_minute_upkeep_time:
                    await self._pstat_update()
                    await self.tasks.minute_upkeep()
                    self._next_minute_upkeep_time = time.time() + 60.0

                # Push updated stats to master.
                if self._stats_to_master.should_update:
                    async with self._stats_to_master.lock:
                        await self._push_stats_to_master()

                # NOTE: disabling std rerouting stuff for now.
                #
                # Ping our custom stdout and stderr handlers
                # periodically so they can flush pending log entries
                # after a while. FIXME: we should build this into the
                # log handler. noinspection PyArgumentList
                # sys.stdout.flush(only_old=True) # type: ignore
                # noinspection PyArgumentList
                # sys.stderr.flush(only_old=True) # type: ignore

                await asyncio.sleep(1.0)

        # except asyncio.CancelledError:
        # This should happen on graceful shutdowns.
        # pass
        except Exception:
            # Hitting this is bad; it means our update loop is dead so
            # we need to just die loudly. Don't let exceptions rise to
            # this level.
            balog.critical(
                'Error in periodic_update.',
                exc_info=True,
                extra=logextra(),
            )
            self.shutdown_app()

    async def handle_stats_sparse(self) -> None:
        """Occasionally send sparse stats to master."""
        while True:
            await asyncio.sleep(BASNStatsSparse.SHIP_INTERVAL_SECONDS)

            starttime = time.monotonic()
            # Unlike with dense stats where we keep a live copy and only
            # push on changes, with sparse stats we just build a
            # complete copy each time.
            stats = BASNStatsSparse()
            self.tasks.get_sparse_stats(stats)

            # Checking to see if this lines up with asyncio warning.
            duration = time.monotonic() - starttime
            if duration > 0.05:
                balog.debug(
                    'stats-sparse took %.3fs', duration, extra=logextra()
                )
            await self._push_stats_sparse_to_master(stats)

    def _handle_asyncio_exception(
        self, loop: asyncio.AbstractEventLoop, context: dict
    ) -> None:
        """Handle asyncio exceptions."""
        import traceback

        assert loop is self._loop
        exc: Exception | None = context.get('exception')
        msg: str = context['message']

        # (mostly) ignore ssl connections that don't get fully
        # established. (should look into whether this is
        # harmless/expected).
        if 'Error on transport creation for incoming connection' in msg:
            balog.info(
                'Ignoring ssl setup error: "%s".',
                msg,
                extra=logextra(),
            )
            return

        # Make noise by default.
        if exc is not None:
            excstr = ''.join(traceback.format_exception(exc))
            msg += f'\n{excstr}'

        balog.error(
            'Error in main asyncio loop:\n%s',
            msg,
            extra=logextra(),
        )

    async def handle_transport_connections(self) -> None:
        """Async task handling transport connections."""
        from basn.env import config

        server = await asyncio.start_server(
            self.handle_transport_connection,
            '0.0.0.0',
            config.transport_port,
            ssl=self._host_ssl_context,
        )

        addrs = ', '.join(str(sock.getsockname()) for sock in server.sockets)
        print(f'Serving on {addrs}')

        async with server:
            await server.serve_forever()

    async def _do_handshake(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ) -> tuple[
        TransportAgentTaskWrapper, batocloud.TransportAgentHandshakeMessage
    ]:
        """Establish communication for a transport task."""
        # pylint: disable=too-many-locals
        from basn.env import BASN_VERSION

        # Read their connection-token; this lets us know if they are
        # one of the clients we are expecting to hear from.
        token = (await reader.readexactly(36)).decode()

        transport_task_ref = self.transport_tasks_by_uuid.get(token)
        transport_task = (
            None if transport_task_ref is None else transport_task_ref()
        )

        uuidsplits = token.split('-')
        looks_like_uuid = len(uuidsplits) == 5 and all(
            s.isalnum() for s in uuidsplits
        )

        # If this token is on our registered-but-not-yet-connected list,
        # take note when it was put there and remove it.
        registered_time: datetime.datetime | None = None
        if token in self.unclaimed_transport_task_uuids:
            registered_time = self.unclaimed_transport_task_uuids[token]
            del self.unclaimed_transport_task_uuids[token]

        # Sanity test: we expect one handoff to occur per task-uuid. If
        # this uuid is legit but has already been claimed, make some
        # noise.
        elif transport_task_ref is not None:
            transportlog.error(
                'Transport task uuid %s not found in unclaimed list'
                ' on handshake.',
                token,
                extra=logextra(),
            )

        # We expect some randos to connect here occasionally and send
        # who-knows-what. Don't make a big fuss if it appears we're
        # getting sent garbage data.
        if not looks_like_uuid:
            raise _IgnoreThisConnectionError()

        if transport_task is None:
            # If this *was* a uuid we created a task for but the task
            # has since been cleaned up (if client is somehow very
            # delayed in its connection attempt) then we can safely
            # ignore too.
            if registered_time is not None and transport_task is None:
                mins_since_registered = (
                    utc_now() - registered_time
                ) / datetime.timedelta(minutes=1)
                transportlog.info(
                    'Ignoring connection attempt for expired'
                    ' transport-task-uuid registered %.1f minutes'
                    ' ago for token %s.',
                    mins_since_registered,
                    token,
                    extra=logextra(),
                )
                raise _IgnoreThisConnectionError()
            raise RuntimeError(f'Transport-task not found for token {token}.')

        # Next they should send us info about themselves
        hslen = int.from_bytes(await reader.readexactly(2), 'big')

        hsdata = await reader.readexactly(hslen)

        handshake = dataclass_from_json(
            batocloud.TransportAgentHandshakeMessage, hsdata.decode()
        )

        # Send our response with info of our own.
        response = batocloud.TransportAgentHandshakeResponse(
            basn_build=BASN_VERSION
        )
        rdata = dataclass_to_json(response).encode()

        writer.write(len(rdata).to_bytes(2, 'big') + rdata)

        if transport_task is None:
            raise RuntimeError(
                f'Found dead-ref for transport-task token {token}.'
            )

        return transport_task, handshake

    async def handle_transport_connection(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ) -> None:
        """Handle a single transport connection."""
        from efro.rpc import (
            ssl_stream_writer_underlying_transport_info,
            ssl_stream_writer_force_close_check,
        )

        try:
            # Accept the token and exchange handoff info.
            #
            # NOTE: in June 2022 I was running into a situation where
            # this would block indefinitely reading the client handshake
            # but the socket connection would not drop. The client would
            # then ask for a new transport connection a few seconds
            # later which would also block, etc. Eventually this would
            # lead to basn meltdown due to hitting max-open-sockets. So
            # now we're being defensive and dropping the connection if
            # handoff hasn't completed in some reasonable amount of
            # time.
            transport_task, handshake = await asyncio.wait_for(
                asyncio.create_task(
                    self._do_handshake(reader, writer),
                    name='transport connection handshake',
                ),
                timeout=30.0,
            )

            # Hand the streams over to the already-running task. It now
            # owns them and is in charge of cleaning them up.
            await transport_task.client_connection_handoff(
                reader, writer, handshake
            )

        except Exception as exc:
            # Silently close if its a normal communication error; otherwise
            # make a bigger fuss.
            if is_asyncio_streams_communication_error(exc):
                transportlog.info(
                    'Closing transport connection pre-handoff due to %s.',
                    exc,
                    extra=logextra(),
                )
            elif isinstance(exc, asyncio.TimeoutError):
                transportlog.info(
                    'Closing transport connection'
                    ' pre-handoff due to handshake timeout (transport=%s).',
                    ssl_stream_writer_underlying_transport_info(writer),
                    extra=logextra(),
                )
            elif not isinstance(exc, _IgnoreThisConnectionError):
                transportlog.exception(
                    'Error in transport connection pre-handoff.',
                    extra=logextra(),
                )

            tinfo = ssl_stream_writer_underlying_transport_info(writer)
            try:
                # TEMP - add a timer to force-kill this socket if it doesn't
                # go down cleanly on its own.
                ssl_stream_writer_force_close_check(writer)

                writer.close()
                await writer.wait_closed()
            except Exception as exc2:
                if is_asyncio_streams_communication_error(exc2):
                    transportlog.info(
                        'Comm error closing transport'
                        ' connection writer pre-handoff (transport=%s).',
                        tinfo,
                        exc_info=True,
                        extra=logextra(),
                    )
                else:
                    transportlog.exception(
                        'Error closing transport connection'
                        ' writer pre-handoff.',
                        extra=logextra(),
                    )

        # We don't expect asyncio.CancelledErrors here (or other
        # base-exceptions); make a slight fuss if one comes through.
        except BaseException:
            transportlog.exception(
                'Got BaseException in handle_transport_connection; unexpected.',
                extra=logextra(),
            )
            raise

    @msg_in.decode_filter_method
    def _msg_in_decode_filter(self, indata: dict, message: Message) -> None:
        timestamp = indata.get(MESSAGE_SIDECAR_TIMESTAMP_KEY)
        assert isinstance(timestamp, float)
        setattr(message, MESSAGE_SIDECAR_TIMESTAMP_KEY, timestamp)

    @msg_in.handler
    async def _log_query(
        self, msg: bamastertobasn.LogQueryMessage
    ) -> bamastertobasn.LogQueryResponse:
        from basn.env import config
        from basn.util import read_lines_filtered

        balog.debug(
            'Handling a LogQueryMessage...',
            extra=logextra(),
        )

        max_size = None if msg.max_size is None else max(0, msg.max_size)

        # If they want a particular task's log, forward to the task.
        if msg.task_id is not None:
            try:
                task = self.tasks.task_by_id(msg.task_id)
            except TaskNotFoundError:
                return bamastertobasn.LogQueryResponse(error='Task not found.')
            return task.handle_log_query_message(msg)

        # Otherwise give them *our* log.
        def _get_log() -> bamastertobasn.LogQueryResponse:
            with open(config.log_path, 'r', encoding='utf-8') as infile:
                if max_size is not None or msg.contains is not None:
                    logtxt, trunclines2 = read_lines_filtered(
                        infile,
                        max_size=max_size,
                        min_log_level=msg.min_log_level,
                        contains=msg.contains,
                    )
                    return bamastertobasn.LogQueryResponse(
                        log=logtxt, truncated_lines=trunclines2
                    )
                return bamastertobasn.LogQueryResponse(
                    log=infile.read(), truncated_lines=0
                )

        # Reading our log can take a bit; do it in a bg thread to avoid
        # clogging asyncio.
        return await asyncio.get_running_loop().run_in_executor(None, _get_log)

    @msg_in.handler
    async def _shutdown(
        self, msg: bamastertobasn.ShutdownMessage
    ) -> bamastertobasn.ShutdownResponse:
        if self._state == State.RUNNING:
            balog.info(
                'Beginning basn shutdown.',
                extra=logextra(),
            )
            self._state = State.SHUTTING_DOWN

        # If they want to complete the shut-down, see if we have any
        # tasks still up. If not, we're shut-down; we just chill in a
        # nice zen state.
        if msg.request_complete:
            if self._state == State.SHUTTING_DOWN:
                if (
                    self.stats.running_task_count_inline == 0
                    and self.stats.running_task_count_standalone == 0
                ):
                    self._state = State.SHUTDOWN_COMPLETE
                    balog.info(
                        'basn shutdown complete.',
                        extra=logextra(),
                    )

        # Let them know in our response whether we're likely able to
        # complete the shutdown if requested, and also if we have
        # completed it.
        return bamastertobasn.ShutdownResponse(
            can_complete=(
                self.stats.running_task_count_inline == 0
                and self.stats.running_task_count_standalone == 0
            ),
            is_complete=self._state == State.SHUTDOWN_COMPLETE,
        )

    @msg_in.handler
    async def _test_err(self, msg: bamastertobasn.TestErrorMessage) -> None:
        del msg  # Unused.
        raise RuntimeError('Testing BASN Exception; this is expected.')

    @msg_in.handler
    async def _test_aio_err(
        self, msg: bamastertobasn.TestAIOErrorMessage
    ) -> None:
        del msg  # Unused.

        async def _do_aio_err_test() -> None:
            raise RuntimeError('Testing AIO Error')

        _t = asyncio.create_task(_do_aio_err_test(), name='aio err test')

        # Tasks can disappear if we don't hold refs to them; make sure
        # to keep _t around long enough to error.
        await asyncio.sleep(1.0)

    @msg_in.handler
    async def _server_node_settings(
        self, msg: bamastertobasn.ServerNodeSettingsMessage
    ) -> None:
        # We should never be getting None values here.
        assert msg.hard_capacity is not None

        balog.info(
            'Got updated settings: hard_capacity %.2f'
            ' and cpu_per_load_unit %.2f and mem_per_load_unit %.2f',
            msg.hard_capacity,
            msg.cpu_per_load_unit,
            msg.mem_per_load_unit,
            extra=logextra(),
        )

        self._mem_danger_level = msg.mem_danger_level

        # We just store these directly in stats.
        # self.stats.speed_factor_old = msg.speed_factor_old
        # self.stats.hard_capacity_old = msg.hard_capacity_old

        self.stats.hard_capacity = msg.hard_capacity
        self.stats.cpu_per_load_unit = msg.cpu_per_load_unit
        self.stats.mem_per_load_unit = msg.mem_per_load_unit
        self.mark_stats_dirty()

    @msg_in.handler
    async def _test_log(self, msg: bamastertobasn.TestLogMessage) -> None:
        log_call = {
            LogLevel.CRITICAL: balog.critical,
            LogLevel.ERROR: balog.error,
            LogLevel.WARNING: balog.warning,
            LogLevel.INFO: balog.info,
            LogLevel.DEBUG: balog.debug,
        }[msg.level]
        log_call('Testing log level %s.', msg.level, extra=logextra())

    @msg_in.handler
    async def _test_print(self, msg: bamastertobasn.TestPrintMessage) -> None:
        del msg  # Unused.
        print('Testing printing to stdout.')

    @msg_in.handler
    async def _update_and_restart(
        self, msg: bamastertobasn.UpdateAndRestartMessage
    ) -> None:
        self.update_and_restart(full=msg.full)

    @msg_in.handler
    async def _test_comm(self, msg: bamastertobasn.TestCommMessage) -> None:
        del msg  # Unused.
        balog.info('Got test-comm message; sending response.', extra=logextra())
        await self.msg_out.send_async(basntobamaster.TestCommResponseMessage())

    @msg_in.handler
    async def _task_test(self, msg: bamastertobasn.TaskTestMessage) -> None:
        from basn.taskwrapper.test import (
            TestTaskWrapperStandalone,
            TestTaskWrapperInline,
            TestTaskWrapperSandbox,
        )
        from baserver.enums import TaskTestType

        assert msg.test_type is not None
        cls = (
            TestTaskWrapperSandbox
            if msg.test_type is TaskTestType.SANDBOX
            else (
                TestTaskWrapperInline
                if msg.test_type
                in [TaskTestType.INLINE, TaskTestType.INLINE_ERROR]
                else TestTaskWrapperStandalone
            )
        )
        self.tasks.launch_task(cls(msg.test_type))

    @msg_in.handler
    async def _invalidate_token(
        self, msg: bamastertobasn.InvalidateTokenMessage
    ) -> None:
        self.tasks.invalidate_token(msg.token)

    @msg_in.handler
    async def _exec_cloud_build(
        self, msg: bamastertobasn.ExecCloudBuildMessage
    ) -> bamastertobasn.ExecCloudBuildResponse:
        # pylint: disable=cyclic-import
        from basn.taskwrapper.cloudbuild import CloudBuildTaskWrapper

        task_id: int | None = None
        try:
            taskwrapper = CloudBuildTaskWrapper(
                recipe_type_id=msg.recipe_type_id,
                input_data=msg.input_data,
                master_server_build_id=msg.master_server_build_id,
                master_server_run_id=msg.master_server_run_id,
                load=msg.load,
                extra=msg.extra,
            )
            task_id = self.tasks.launch_task(taskwrapper)
        except CleanError as exc:
            # Expected fail cases should give us CleanErrors.
            hostedgamelog.info(
                'Unable to launch task: %s', exc, extra=logextra()
            )
        except Exception:
            # Unexpected failures; make more noise.
            hostedgamelog.exception(
                'Error launching cloud-build task.', extra=logextra()
            )

        return bamastertobasn.ExecCloudBuildResponse(task_id=task_id)

    @msg_in.handler
    async def _host_game(
        self, msg: bamastertobasn.HostGameMessage
    ) -> bamastertobasn.HostGameResponse:
        from basn.taskwrapper.hostedgame import HostedGameTaskWrapper

        task_id: int | None = None
        udp_port: int | None = None
        try:
            taskwrapper = HostedGameTaskWrapper(msg.gamedef)
            task_id = self.tasks.launch_task(taskwrapper)
            udp_port = taskwrapper.udp_port
        except CleanError as exc:
            # Expected fail cases.
            hostedgamelog.info(
                'Unable to launch task: %s', exc, extra=logextra()
            )
        except Exception:
            # Unexpected failures; make more noise.
            hostedgamelog.exception(
                'Error launching cloud-build task.', extra=logextra()
            )

        return bamastertobasn.HostGameResponse(
            task_id=task_id, udp_port=udp_port
        )

    @msg_in.handler
    async def _clear_errors(
        self, msg: bamastertobasn.ClearErrorsMessage
    ) -> None:
        del msg  # Unused.
        balog.debug(
            'Handling ClearErrorsMessage...',
            extra=logextra(),
        )
        self.stats.max_log_level = LogLevel.INFO
        self.tasks.reset_errors()
        self.mark_stats_dirty()

    @msg_in.handler
    async def _test_scratch(
        self, msg: bamastertobasn.TestScratchMessage
    ) -> None:
        del msg  # Unused.
        from baserver.util import async_run
        from basn.env import config

        print('HELLO FROM NEW SCRATCH HANDLER', config.transport_port)
        try:
            await async_run('echo poo foofoo && false')
        except Exception as exc:
            print('GOT EXC', type(exc), exc)
        balog.info(
            'Got test-scratch message.',
            extra=logextra(),
        )

    @msg_in.handler
    async def _ping(self, msg: bamastertobasn.PingMessage) -> None:
        del msg  # Unused.
        balog.debug(
            'Handling PingMessage...',
            extra=logextra(),
        )

    @msg_in.handler
    async def _tasks_query(
        self, msg: bamastertobasn.TasksQueryMessage
    ) -> bamastertobasn.TasksQueryResponse:
        return self.tasks.get_tasks_info_message(msg.min_log_level)

    @msg_in.handler
    async def _task_kill(self, msg: bamastertobasn.TaskKillMessage) -> None:
        try:
            self.tasks.task_by_id(msg.task_id).force_kill()
        except TaskNotFoundError:
            # Got asked to kill an already-dead-and-pruned task. All good.
            pass

    @msg_in.handler
    async def _task_running(
        self, msg: bamastertobasn.TaskRunningMessage
    ) -> BoolResponse:
        try:
            running = (
                self.tasks.task_by_id(msg.task_id).state
                is ServerNodeTaskState.RUNNING
            )
        except TaskNotFoundError:
            running = False
        return BoolResponse(value=running)

    @msg_in.handler
    async def _task_query(
        self, msg: bamastertobasn.TaskQueryMessage
    ) -> bamastertobasn.TaskQueryResponse | None:
        # Give them the full response.
        try:
            return self.tasks.task_by_id(msg.task_id).get_task_info_message()
        except TaskNotFoundError:
            return None

    @msg_in.handler
    async def _set_run_mode(
        self, msg: bamastertobasn.SetRunModeMessage
    ) -> None:
        from baserver.data import BASNConfig

        with open(self._config_path, encoding='utf-8') as infile:
            newcfg = dataclass_from_json(BASNConfig, infile.read())
        newcfg.run_mode = msg.run_mode
        with open(self._config_path, 'w', encoding='utf-8') as outfile:
            outfile.write(dataclass_to_json(newcfg, pretty=True))
        balog.debug(
            'Wrote run_mode %s to config at %s.',
            msg.run_mode,
            self._config_path,
            extra=logextra(),
        )

    @msg_in.handler
    async def _wait_state(self, msg: bamastertobasn.WaitStateMessage) -> None:
        del msg  # Unused.

        # Master sends this after it's sent us something that it thinks
        # will result in us pushing it some new state.
        # So we block here for up to a few seconds waiting for our
        # latest state to be pushed.
        starttime = time.time()
        success = False
        while time.time() - starttime < 5.0:
            if self.stats == self._stats_master:
                success = True
                break
            await asyncio.sleep(0.25)
        if success:
            wait_time = time.time() - starttime
            if wait_time > 0.001:
                balog.debug(
                    'Waited %.2fs for node state update.',
                    time.time() - starttime,
                    extra=logextra(),
                )
        else:
            balog.warning(
                'Timed out waiting for node state update.', extra=logextra()
            )

    @msg_in.handler
    async def _task_action(
        self, msg: bamastertobasn.TaskActionMessage
    ) -> bamastertobasn.TaskActionResponse:
        return self.tasks.task_by_id(msg.task_id).action(msg.action)

    def _console_exec(self, code: str) -> None:
        import __main__

        try:
            # First try it as eval.
            try:
                evalcode = compile(code, '<console>', 'eval')
            except SyntaxError:
                evalcode = None
            except Exception:
                # hmm; when we can't compile it as eval will we always get
                # syntax error?
                balog.exception(
                    'Unexpected error compiling code for cloud-console eval.',
                    extra=logextra(),
                )
                evalcode = None
            if evalcode is not None:
                # pylint: disable=eval-used
                value = eval(evalcode, vars(__main__), vars(__main__))
                # For eval-able statements, print the resulting value if
                # it is not None (just like standard Python interpreter).
                if value is not None:
                    print(repr(value), file=sys.stderr)

            # Fall back to exec if we couldn't compile it as eval.
            else:
                execcode = compile(code, '<console>', 'exec')
                # pylint: disable=exec-used
                exec(execcode, vars(__main__), vars(__main__))

        except Exception:
            import traceback

            apptime = time.monotonic() - self._app_start_time
            print(f'Exec error at time {apptime:.2f}.', file=sys.stderr)
            traceback.print_exc()

            # This helps the logging system ship stderr back to the
            # cloud promptly.
            sys.stderr.flush()

    @msg_in.handler
    async def _cloud_console_update(
        self, msg: bamastertobasn.CloudConsoleUpdateMessage
    ) -> bamastertobasn.CloudConsoleUpdateResponse:
        submit_time = time.monotonic() - self._app_start_time

        if msg.code is not None:
            assert self._loop is not None
            self._loop.call_soon_threadsafe(
                tpartial(self._console_exec, msg.code)
            )

        # If we're running code, wait just a moment to allow resulting
        # logs/prints/errors to make their way across threads to the
        # logging cache. We'd pick these log messages up next time
        # anyway but it feels nice and crisp to get them immediately.
        if msg.code is not None:
            await asyncio.sleep(0.05)

        # This will never change for us so it doesn't matter.
        our_app_uuid = '-'

        # If they've been dealing with a different app instance,
        # start over at log index 0.
        if our_app_uuid != msg.target_uuid:
            log_index = 0
        else:
            # Ok, seems they may have some of our log already.
            log_index = msg.log_index

        # Now ship them our UUID and whatever log messages they don't
        # have.
        return bamastertobasn.CloudConsoleUpdateResponse(
            target_uuid=our_app_uuid,
            app_real_time=submit_time,
            log_archive=loghandler.get_cached(start_index=log_index),
        )

    @msg_in.handler
    async def _host_ping_target(
        self, msg: bamastertobasn.HostPingTargetMessage
    ) -> bamastertobasn.HostPingTargetResponse:
        from basn.taskwrapper.pingtarget import PingTargetTaskWrapper

        task_id: int | None = None
        port: int | None = None
        try:
            task_id = self.tasks.launch_task(
                PingTargetTaskWrapper(
                    duration_seconds=msg.duration_seconds, load=msg.load
                )
            )
            port = self.tasks.task_by_id(task_id).udp_port
            assert port is not None
            return bamastertobasn.HostPingTargetResponse(
                task_id=task_id, port=port
            )
        except Exception as exc:
            balog.info(
                'Unable to launch ping-target task: %s.', exc, extra=logextra()
            )
        return bamastertobasn.HostPingTargetResponse(task_id=None, port=None)

    @msg_in.handler
    async def _host_transport_agent(
        self, msg: bamastertobasn.HostTransportAgentMessage
    ) -> bamastertobasn.HostTransportAgentResponse:
        from basn.taskwrapper.transportagent import TransportAgentTaskWrapper
        from basn.env import config

        starttime = time.time()
        try:
            task = TransportAgentTaskWrapper(
                connection_wait_seconds=msg.connection_wait_seconds,
                max_duration_seconds=msg.max_duration_seconds,
                load=msg.load,
                token=msg.token,
                user_agent_string=msg.user_agent_string,
                remote_addr=msg.remote_addr,
                location=msg.location,
            )
            task_id = self.tasks.launch_task(task)

            # Go ahead and prune dead tasks from the list before we add
            # ours.
            self.transport_tasks_by_uuid = {
                k: v
                for k, v in self.transport_tasks_by_uuid.items()
                if v() is not None
            }
            # So we can get to this task later by its uuid (token)
            assert msg.token not in self.transport_tasks_by_uuid
            assert msg.token not in self.unclaimed_transport_task_uuids
            self.transport_tasks_by_uuid[msg.token] = weakref.ref(task)
            self.unclaimed_transport_task_uuids[msg.token] = utc_now()

            now = time.time()
            duration = now - starttime
            fullduration = now - self.message_send_timestamp(msg)
            transportlog.info(
                'Hosting transport-agent task %s:%s: token=%s'
                ' remote_addr=%s location=%s user_agent_string=%s'
                ' handletime=%.2f fulltime=%.2f',
                config.node_id,
                task_id,
                msg.token,
                msg.remote_addr,
                msg.location,
                msg.user_agent_string,
                duration,
                fullduration,
                extra=logextra(task=task),
            )

            return bamastertobasn.HostTransportAgentResponse(
                task_id=task_id, port=config.transport_port
            )
        except Exception as exc:
            transportlog.info(
                'Unable to launch transport-agent task: %s.',
                exc,
                extra=logextra(),
            )
            return bamastertobasn.HostTransportAgentResponse(
                task_id=None, port=None
            )

    @msg_in.handler
    async def _forward_msg_to_client(
        self, msg: bamastertobasn.ForwardedToClientMessage
    ) -> bamastertobasn.ForwardedToClientResponse:
        from basn.taskwrapper.transportagent import TransportAgentTaskWrapper

        response: str

        # Look for a transport-agent task with this id and have it
        # forward the message along to its client on our behalf.
        try:
            task = self.tasks.task_by_id(msg.client_task_id)
        except TaskNotFoundError:
            etp = bamastertobasn.ForwardedToClientResponse.ErrorType
            return bamastertobasn.ForwardedToClientResponse(
                error=etp.COMMUNICATION
            )

        if not isinstance(task, TransportAgentTaskWrapper):
            raise RuntimeError('Provided task is not an agent')

        try:
            # Call our raw sender call directly since the message is
            # already in wire form. We'll get a raw string back which we
            # send directly back to the caller.
            response = await task.send_raw_message_async(msg.message)
        except CommunicationError:
            # We expect this to happen sometimes when the client goes
            # offline while talking to them/etc.
            etp = bamastertobasn.ForwardedToClientResponse.ErrorType
            return bamastertobasn.ForwardedToClientResponse(
                error=etp.COMMUNICATION
            )
        except Exception:
            # In other unexpected cases let's make a bit of noise for now.
            balog.warning(
                'Unknown error forwarding message to client.',
                exc_info=True,
                extra=logextra(),
            )
            return bamastertobasn.ForwardedToClientResponse(
                error=bamastertobasn.ForwardedToClientResponse.ErrorType.UNKNOWN
            )

        return bamastertobasn.ForwardedToClientResponse(response=response)

    @msg_in.handler
    async def _get_debug_info(
        self, msg: bamastertobasn.GetDebugInfoMessage
    ) -> bamastertobasn.GetDebugInfoResponse:
        from basn.util import get_debug_info

        del msg  # Unused.
        return bamastertobasn.GetDebugInfoResponse(output=get_debug_info())

    async def _push_stats_to_master(self) -> None:
        """Attempt to send our latest stat values to master."""
        from efro.dataclassio.extras import DataclassDiff

        # If stats hasn't changed from the last one we sent, we're good.
        if self.stats == self._stats_master:
            self._stats_to_master.dirty = False
            return

        statsmsg = basntobamaster.ServerNodeStatsMessage()
        # Copy this so subsequent local stats changes won't affect it.
        statsmsg.stats = copy.deepcopy(self.stats)
        balog.debug(
            'Pushing stats changes to bamaster:\n%s',
            DataclassDiff(self._stats_master, self.stats),
            extra=logextra(),
        )
        try:
            await self.msg_out.send_async(statsmsg)

            # Set known server state to stats we just sent, NOT
            # self.stats (which may have changed at this point). Mark
            # ours clean ONLY if they haven't changed.
            self._stats_master = statsmsg.stats
            if self.stats == self._stats_master:
                self._stats_to_master.dirty = False
            self._stats_push_fail_count = 0
        except CommunicationError:
            # Occasional comm-errors are expected; only escalate to warnings
            # if it keeps occurring.
            self._stats_push_fail_count += 1
            logcall = (
                balog.warning if self._stats_push_fail_count > 4 else balog.info
            )
            logcall(
                'Stats push to master comm-errored (%d consecutive).',
                self._stats_push_fail_count,
                extra=logextra(),
            )
        except Exception:
            # Make more fuss for anything else.
            balog.exception(
                'Error pushing stats to master.',
                extra=logextra(),
            )

    async def _push_stats_sparse_to_master(
        self, stats: BASNStatsSparse
    ) -> None:
        """Attempt to send our latest stat values to master."""
        try:
            await self.msg_out.send_async(
                basntobamaster.ServerNodeStatsSparseMessage(stats)
            )
            self._stats_sparse_push_fail_count = 0
        except Exception:
            # Only escalate this up to a warning if it keeps happening.
            self._stats_sparse_push_fail_count += 1
            logcall = (
                balog.warning
                if self._stats_sparse_push_fail_count > 4
                else balog.info
            )
            logcall(
                'Stats push to master failed (%d consecutive).',
                self._stats_sparse_push_fail_count,
                extra=logextra(),
            )

    def _check_sockets_danger(self) -> None:
        if self.stats.open_sockets_count is None:
            return
        old_sockets_danger = self._sockets_danger
        danger_level = 900
        self._sockets_danger = self.stats.open_sockets_count >= danger_level

        # Raise a fuss when this initially happens.
        if not old_sockets_danger and self._sockets_danger:
            balog.critical(
                'Sockets usage %d reached danger-level %d.',
                self.stats.open_sockets_count,
                danger_level,
                extra=logextra(),
            )

    def _check_mem_danger(self, mtotalraw: int, mavailraw: int) -> None:
        if self._mem_danger_level is None:
            return

        old_mem_danger = self._mem_danger
        mem_used_amt = 1.0 - (mavailraw / mtotalraw)
        self._mem_danger = mem_used_amt > self._mem_danger_level

        # Raise a fuss when this initially happens.
        if not old_mem_danger and self._mem_danger:
            # Running manual mode sucks up more memory so let's be less
            # drastic there since it could be related.
            if os.environ.get('BASN_MANUAL_RUN') == '1':
                balog.error(
                    'Memory usage %.2f reached danger-level %.2f'
                    ' (but you are running in manual mode'
                    ' which may be partly to blame).',
                    mem_used_amt,
                    self._mem_danger_level,
                    extra=logextra(),
                )
            else:
                balog.critical(
                    'Memory usage %.2f reached danger-level %.2f.',
                    mem_used_amt,
                    self._mem_danger_level,
                    extra=logextra(),
                )

    async def _pstat_update(self) -> None:
        if self._loop is not None:
            await self._loop.run_in_executor(
                executor=None, func=self._do_pstat_update
            )
        self.mark_stats_dirty()

    def _do_pstat_update(self) -> None:
        starttime = time.monotonic()

        bytes_per_meg = 1024 * 1024

        proc = psutil.Process()
        self.stats.open_sockets_count = len(proc.connections(kind='inet'))
        self._check_sockets_danger()
        self.stats.open_files_count = len(proc.open_files())

        swap = psutil.swap_memory()
        self.stats.system_swap_used = swap.used // bytes_per_meg

        # Calc the megs of data we read/wrote to swap since the last
        # update.
        swap_io_total = swap.sin + swap.sout
        self.stats.system_swap_io = (
            swap_io_total - self._last_swap_io_total
        ) // bytes_per_meg
        self._last_swap_io_total = swap_io_total

        duse = psutil.disk_usage('/')
        self.stats.disk_total = duse.total // bytes_per_meg
        self.stats.disk_used = duse.used // bytes_per_meg
        self.stats.disk_free = duse.free // bytes_per_meg

        self.stats.basn_memory_used = round(
            self._process.memory_info().rss / bytes_per_meg, 1
        )

        vmem = psutil.virtual_memory()
        mtotalraw: int = vmem.total
        mavailraw: int = vmem.available
        self._check_mem_danger(mtotalraw=mtotalraw, mavailraw=mavailraw)

        # Round sys mem values to the nearest 5 megs to reduce updates
        # we send home. (mult by 2, round to nearest 10, divide by 2)
        val = mtotalraw / bytes_per_meg
        self.stats.system_memory_total = round(val * 2, -1) // 2
        val = (mtotalraw - mavailraw) / bytes_per_meg
        self.stats.system_memory_used = round(val * 2, -1) // 2

        duration = time.monotonic() - starttime

        # Complain if this starts taking crazy amounts of time.
        if duration > 10.0:
            balog.warning(
                '_pstat_update_bg took %.1f seconds.',
                duration,
                extra=logextra(),
            )

    async def send_comm_token(self) -> None:
        """Send our comm-key to our controller"""
        from basn.env import BASN_VERSION, config

        # We also send them our self-signed cert so they can confidently
        # establish communication back to us.
        with open(self.cert_path, encoding='utf-8') as infile:
            cadata = infile.read()

        msg = basntobamasterboot.CommKeyMessage(
            key=self._comm_token,
            address=self.our_address,
            port=config.port,
            version=BASN_VERSION,
            cadata=cadata,
        )
        balog.info(
            'Sending comm-key...',
            extra=logextra(),
        )
        try:
            response = await self.msg_out_boot.send_async(msg)
        except Exception:
            response = None

        # Its possible that we get *no* response or we get a valid
        # response but something went wrong on the other end. In either
        # case we'll give up for this pass but will keep retrying since
        # we're dead in the water without a comm-key set.
        if response is None or not response.success:
            balog.exception(
                'Error sending comm-key.',
                extra=logextra(),
            )
            return

        balog.info(
            'Comm-key set.',
            extra=logextra(),
        )
        self.on_comm_key_set()

        # Take note that we can stop sending them comm-key messages.
        self._comm_setup.dirty = False

    # So remote errors/etc. are clear about where they happened.
    @msg_out.peer_desc_method
    def _get_peer_desc(self) -> str:
        return 'bamaster'

    @msg_out.send_async_ex_method
    async def _send_message_to_master(self, data: str, message: Message) -> str:
        return await self._do_send_message_to_master(
            data=data, boot=False, message=message
        )

    @msg_out_boot.send_async_method
    async def _send_message_to_master_boot(self, data: str) -> str:
        return await self._do_send_message_to_master(
            data=data, boot=True, message=None
        )

    async def _do_send_message_to_master(
        self, data: str, boot: bool, message: Message | None
    ) -> str:
        from basn.env import config

        data_encrypted = await self.encrypt_str(data, use_boot_token=boot)
        args: dict[str, str] = {'eid': config.node_id, 'd': data_encrypted}

        target = 'basninit' if boot else 'basncomm'

        # We embed the message type into the url so we get more fine grained
        # stats on bamaster.
        if message is not None:
            target = f'{target}_{type(message).__name__}'

            # If the message has a 'forwarded-internal-type' stapled
            # to it, include that as well.
            internal_type = getattr(message, '_fwd_int_tp', None)
            if isinstance(internal_type, int):
                target = f'{target}_{internal_type}'

        # HMM: should we support this?...
        # IMPORTANT: if we DO want to support this, make sure to
        # not pass None to timeout for the post() call when we're
        # not providing this. Otherwise we override our default.
        # timeout_seconds: int | None = None

        assert self.httpclient is not None
        url = f'{config.master_server_addr}/{target}'
        try:
            async with self.httpclient.post(
                url,
                data=args,
                ssl=self.default_ssl_context,
            ) as rsp:
                rsp.raise_for_status()
                return await self.decrypt_str(
                    await rsp.text(), use_boot_token=boot
                )
        except aiohttp.client_exceptions.ServerDisconnectedError as exc:
            # Seeing these very rarely; assuming this fits the definition
            # of comm-error.
            raise CommunicationError() from exc
        except aiohttp.client_exceptions.ClientConnectorError as exc:
            # We get these when the host closes the connection unexpectedly.
            # Should be rare.
            self._log_fuzzy_comm_error(
                f'Treating ClientConnectorError to {url} as comm-error.'
            )
            raise CommunicationError() from exc
        except aiohttp.client_exceptions.ClientOSError as exc:
            excstr = str(exc)
            # Look for some standard stuff that fits our definition of
            # CommunicationError.
            if 'Connection reset by peer' in excstr:
                raise CommunicationError() from exc
            raise

        except aiohttp.client_exceptions.ClientResponseError as exc:
            # I'm a bit hesitant to lump 502 'Bad Gateway' in with
            # communication errors since it can be a sign of a serious
            # issue and not simply a momentary problem, but with
            # bamaster we do seem to get them periodically at random. I
            # suppose we can just try and be sure to always print
            # warnings for even CommunicationErrors since they should be
            # rare between us and bamaster so that no legit issues slip
            # through the cracks.

            # 503 'Service Unavailable' errors are technically supposed
            # to represent transient problems (and that seems to be the
            # case with app-engine) so that seems like a good mapping to
            # comm-error.
            if exc.status in {502, 503}:
                raise CommunicationError() from exc

            # 404 'Not Found' is usually a permanent issue such as an
            # incorrect URL, but I'm seeing app-engine occasionally
            # return these for valid urls; seemingly right at launch but
            # I'm not positive. So for now let's count it as a
            # comm-error but emit a warning since it's unusual.
            # Hopefully can remove this later.
            if exc.status == 404:
                self._log_fuzzy_comm_error(
                    f'Treating 404 not-found error for {url} as comm-error.'
                )
                raise CommunicationError() from exc

            raise
        except asyncio.TimeoutError as exc:
            # Getting occasional asyncio.TimeoutErrors. Should we just
            # silently treat these as comm-errors or is this indicative
            # of a deeper problem? For now making a bit of noise to be
            # safe.
            self._log_fuzzy_comm_error(
                f'Treating asyncio.TimeoutError for {url} as comm-error.',
            )
            raise CommunicationError() from exc

    def _log_fuzzy_comm_error(self, msg: str) -> None:
        # Called for errors which technically might not be communication
        # errors but which we're seeing happen sporadically so we're
        # treating them as such. In this case we only kick out warnings
        # if we're seeing a lot of them, which should hopefully get our
        # attention if actual static configuration issues come up.

        # Reset count/period after an hour.
        now = time.monotonic()
        if now - self._fuzzy_comm_error_start_time > 60 * 60:
            self._fuzzy_comm_error_start_time = now
            self._fuzzy_comm_error_count = 0

        self._fuzzy_comm_error_count += 1

        # Only warn if we get more than X in an hour.
        logcall = (
            balog.warning if self._fuzzy_comm_error_count > 10 else balog.info
        )
        logcall('%s (#%d)', msg, self._fuzzy_comm_error_count, extra=logextra())

    async def encrypt_str(self, s: str, use_boot_token: bool = False) -> str:
        """Encrypt/compress outgoing data."""
        from basn.env import config

        # Run compression/etc in another thread so we don't clog up the
        # main thread here. We store tokens as base64 strings but use
        # the actual bytes here.
        tokenstr = config.boot_token if use_boot_token else self._comm_token
        assert isinstance(tokenstr, str)
        return await asyncio.get_event_loop().run_in_executor(
            None, tpartial(self._encrypt_str, s, tokenstr)
        )

    def _encrypt_str(self, s: str, tokenstr: str) -> str:
        # Note: this runs in an executor thread.
        token = urlsafe_b64decode(tokenstr.encode())
        return str_to_wire_key_plaintext(s, token)

    async def decrypt_str(self, data: str, use_boot_token: bool = False) -> str:
        """Decrypt/decompress incoming data."""
        from basn.env import config

        # Run compression/etc in another thread so we don't clog up the
        # main thread here. We store tokens as base64 strings but use
        # the actual bytes here.
        tokenstr = config.boot_token if use_boot_token else self._comm_token
        assert isinstance(tokenstr, str)
        return await asyncio.get_event_loop().run_in_executor(
            None, tpartial(self._decrypt_str, data, tokenstr)
        )

    def _decrypt_str(self, data: str, tokenstr: str) -> str:
        # Note: this runs in an executor thread.
        token = urlsafe_b64decode(tokenstr.encode())
        return str_from_wire_key_plaintext(data, token)

    async def handle_get(self, request: Any) -> aiohttp.web.Response:
        """Testing."""
        del request  # Unused.
        return aiohttp.web.Response(text='Hello, world')

    async def handle_post(self, request: Any) -> aiohttp.web.Response:
        """Handle incoming root POST requests."""
        responsetxt: str
        try:
            postvars = await request.post()
            data = postvars.get('d')
            if not isinstance(data, str):
                balog.info(
                    'Ignoring post with no data (got %s).',
                    type(data),
                    extra=logextra(),
                )
                return aiohttp.web.Response(text='')
            responsetxt = await self.msg_in.handle_raw_message(
                await self.decrypt_str(data)
            )
        except Exception as exc:
            # We expect the occasional hangup or whatnot.
            if is_asyncio_streams_communication_error(exc):
                balog.warning(
                    'Comm error handling post.',
                    exc_info=True,
                    extra=logextra(),
                )
            else:
                balog.exception(
                    'Unexpected error handling post.',
                    extra=logextra(),
                )
            responsetxt = self.msg_in.encode_error_response(exc)

        return aiohttp.web.Response(text=await self.encrypt_str(responsetxt))

    def log_callback(self, entry: LogEntry) -> None:
        """Called by our log handler to inform us of log-levels."""

        # If we haven't started our loop yet, save these for later.
        if self._loop is None:
            self._early_logs.append(entry)
            return

        self._loop.call_soon_threadsafe(tpartial(self._log_callback, entry))

    def _log_callback(self, entry: LogEntry) -> None:
        assert (
            self._loop is not None and asyncio.get_running_loop() == self._loop
        )
        if entry.level.value > self.stats.max_log_level.value:
            self.stats.max_log_level = entry.level
            self._stats_to_master.dirty = True

    def update_and_restart(self, full: bool) -> None:
        """Update to latest basn and reboot to bring it up."""
        import subprocess
        from basn.env import config

        # Here we simply grab and execute the current basn_bootstrap script
        # just as we did during our initial init.

        balog.info(
            'Running update-and-reboot...',
            extra=logextra(),
        )

        # Run whatever bootstrap script they point us at.
        envs = f'BASN_CONFIG_PATH="{self._config_path}"'
        if not full:
            envs += ' BASN_INSTALL_FAST=1'

        booturl = config.bootstrap_url

        # Dump the update process output to disk so we can pick it apart
        # after the reboot. Set perms to keep it away from prying eyes.
        self.update_log_path.touch()
        subprocess.run(['chmod', 'o-rwx', self.update_log_path], check=True)

        # Just fire this off in the background and hope it works i
        # guess?... Even if we run it synchronously it'll still fail
        # because the reboot at the end kills the script. ¯\_(ツ)_/¯
        # Note that this script originally ran on whatever python3
        # defaulted to so we do the same here for consistency.
        subprocess.run(
            f'curl -L {booturl} | sudo {envs} python3'
            f' > {self.update_log_path} 2>&1 &',
            shell=True,
            check=False,
        )

    def shutdown_app(self) -> None:
        """Begin a graceful exit of the app."""
        import signal

        # From what I can tell, the cleanest way to kick out of the core
        # loop and exit gracefully is by just sending ourself an
        # OS-level interrupt signal. Plus its a nice opportunity to make
        # sure we behave if the os should send us one.
        os.kill(os.getpid(), signal.SIGINT)

    # Here at the end of the class declaration let's make sure we've
    # registered handlers for all protocol messages/etc.
    if __debug__:
        msg_in.validate()


class _IgnoreThisConnectionError(Exception):
    pass
