"""General environment for basn."""

# IMPORTANT - make sure we don't pull in any modules from our own package here;
# (aside from under 'if TYPE_CHECKING' which doesn't count) These env
# modules need to be the first inited since they may set up logging or
# other implicitly global stuff.

from __future__ import annotations

import sys
import json
from typing import TYPE_CHECKING

from efro.dataclassio import dataclass_from_json
from baserver.enums import ServerNodeRunMode
from baserver.data import BASNConfig

import basn.logging

if TYPE_CHECKING:
    from typing import Any
    from baserver.cloudbuild import Recipe

    from basn.app import App


def _get_version() -> int:
    """Load our version number from the file."""
    with open('version', encoding='utf-8') as infile:
        value = json.loads(infile.read())['build']
        assert isinstance(value, int)
        return value


def _get_config() -> BASNConfig:
    """Build our config instance from our config file."""

    # We expect exactly 1 arg: the config path.
    if len(sys.argv) != 2:
        raise RuntimeError('Expected exactly 1 arg (config path)')

    config_path = sys.argv[1]
    with open(config_path, encoding='utf-8') as infile:
        return dataclass_from_json(BASNConfig, infile.read())


config = _get_config()
app: App  # global app context


# Version number for this script:
BASN_VERSION = _get_version()

# Version of ballisticakit we use:
DEFAULT_BALLISTICAKIT_VERSION = 21791
DEFAULT_EON_VERSION = 3

# Note: need to make sure we have a logging handler in place
# before aiohttp run_app, otherwise we can wind up with duplicate
# handlers (aiohttp adds its own in some debug cases and other
# log calls call basicConfig() if no handlers are present)

basn.logging.setup_basn_logging()

# Flip on some debugging options in non-release run-modes.
ASYNCIO_DEBUG: bool
if config.run_mode not in [ServerNodeRunMode.RELEASE]:  # pylint: disable=R1703
    ASYNCIO_DEBUG = True
else:
    ASYNCIO_DEBUG = False
