"""BASN install functionality.

This is not used as part of the basn task, but rather called explicitly
to install bssn. It is run as root.
"""

from __future__ import annotations

import pwd
import sys
import os
import time
import datetime
import subprocess
from pathlib import Path
from typing import TYPE_CHECKING
from concurrent.futures import ThreadPoolExecutor

from efro.call import tpartial
from efro.terminal import ClrAlways as Clr

if TYPE_CHECKING:
    from typing import Callable
    from baserver.data import BASNConfig


class Installer:
    """Installer app instance."""

    def __init__(self) -> None:
        """Install/update basn and reboot the machine.

        It should be run as root in the extracted basn dir.
        It should install necessary pip modules, set permissions, etc.
        It should not import any other basn modules since the actual basn
        environment is not complete when this is run.
        """
        print(
            f'{Clr.BLU}BASN.INSTALL: STARTING AT'
            f' {datetime.datetime.now()}{Clr.RST}',
            flush=True,
        )
        self._starttime = time.time()
        self._config_path = sys.argv[-1]

        # This can be used when debugging the bootstrap process directly on
        # a machine to avoid having to wait for a reboot each time.
        self._suppress_reboot = os.environ.get('BASN_SUPPRESS_REBOOT') == '1'

        self._fast = os.environ.get('BASN_INSTALL_FAST') == '1'

        # We don't want to read config yet, so deduce lib_root from
        # where we were launched from (lib_root is our parent).
        self._lib_root = os.path.abspath(os.path.join(os.getcwd(), '..'))
        assert os.path.exists('requirements.txt')
        assert os.path.exists(self._config_path)

        self._config: BASNConfig | None = None

    def _our_python_version(self) -> str:
        return f'{sys.version_info.major}.{sys.version_info.minor}'

    def run(self) -> None:
        """Do the thing."""

        with ThreadPoolExecutor(max_workers=10) as executor:
            # Break our work up into sets of parallel tasks so we can
            # separate things that could interfere with each other.
            # (for instance we don't want to update pip stuff while python
            # itself is being updated via apt).

            # Group 1: apt updates and other things with no dependencies.
            self._run_group(
                executor,
                [
                    self._update_apt_stuff,
                    self._setup_swap,
                ],
            )

            # Group 2: things depending on system libs being up-to-date
            # via apt-updates.
            self._run_group(
                executor,
                [
                    # Do pip installs only for our python version because we
                    # need them to read our config to see what other versions
                    # we should install for.
                    tpartial(
                        self._update_pip_stuff, self._our_python_version()
                    ),
                    self._update_misc_stuff,
                ],
            )

            # Group 3: things depending on pip stuff being up-to-date.
            self._read_config()

            # Group 4: things depending on self._config being in place.
            assert self._config is not None
            self._run_group(
                executor,
                [
                    self._write_rc_local,
                    self._setup_log,
                    self._set_permissions,
                    self._setup_users,
                ]
                + [
                    tpartial(self._update_pip_stuff, ver)
                    for ver in self._config.all_python_versions
                    if ver != self._our_python_version()
                ],
            )

        print(
            f'{Clr.BLU}BASN.INSTALL: INSTALL TASKS COMPLETED'
            f' IN {time.time()-self._starttime:.1f} SECONDS.{Clr.RST}',
            flush=True,
        )

        if self._suppress_reboot:
            print(
                f'{Clr.BLU}BASN.INSTALL: EXITING WITHOUT REBOOT...{Clr.RST}',
                flush=True,
            )
        else:
            print(f'{Clr.BLU}BASN.INSTALL: REBOOTING...{Clr.RST}', flush=True)
            subprocess.run(['reboot'], shell=False, check=False)

    def _run_group(
        self, executor: ThreadPoolExecutor, calls: list[Callable[[], None]]
    ) -> None:
        tasks = [executor.submit(c) for c in calls]

        # Wait for all results (which will propagate exceptions)
        for task in tasks:
            task.result()

    def _read_config(self) -> None:
        # NOTE: We currently need to avoid importing this until this call
        # because it pulls in typing_extensions which we won't have until
        # pip stuff completes.
        from efro.dataclassio import dataclass_from_json
        from baserver.data import BASNConfig

        with open(self._config_path, encoding='utf-8') as infile:
            self._config = dataclass_from_json(BASNConfig, infile.read())

        assert self._config.user
        assert self._config.log_path

    def _update_apt_stuff(self) -> None:
        if self._fast:
            print(
                f'{Clr.BLU}BASN.INSTALL: FAST-MODE;'
                f' SKIPPING APT INSTALLS/UPDATES.{Clr.RST}',
                flush=True,
            )
            return

        attempt = 0
        print(
            f'{Clr.BLU}BASN.INSTALL: UPDATING APT STUFF...{Clr.RST}', flush=True
        )
        start_time = time.time()
        while True:
            try:
                # Seems this is necessary before pip will work on our
                # alternate python versions (even though there is not the
                # equivalent for them)
                # UPDATE: seems this is not necessary if we use the
                # pypa bootstrap script.
                # packages = ['python3-pip']
                # pstr = ' '.join(packages)

                # Gotta specify noninteractive front-end to avoid corner-case
                # where apt asks the user for input during install/etc.
                # (update: not sure if this is still needed with -y).

                # Note: we also remove unattended-upgrades; we spin up
                # new machines enough that this is unnecessary and it can
                # cause us to hit swap if it fires when we're in a low-memory
                # situation.

                # We also remove apport; we'll futz with our own crashlog
                # handling later.

                # Ditto with fwupd (firmware-update-daemon); seeing it kick
                # on at times while we're running (not sure if driven by
                # unattended-upgrades but not taking chances)

                # Also snapd; don't need this or want it doing any magic
                # in the background.

                # Also packagekit; saw this randomly spin up too and push
                # me into swap... le sigh.

                # ufw.. why not?..

                # squashfs-tools friendly-recovery & at
                # (listed on https://motorscript.com/cleanup-ubuntu-server/)

                subprocess.run(
                    'export DEBIAN_FRONTEND=noninteractive'
                    ' && apt-get -y purge unattended-upgrades apport'
                    ' fwupd snapd packagekit ufw squashfs-tools'
                    ' friendly-recovery at'
                    ' && apt-get update'
                    # f' && apt-get -y install {pstr}'
                    ' && apt-get -y install faketime'  # for cert generation
                    ' && apt-get -y dist-upgrade'
                    ' && apt-get -y clean'
                    ' && apt-get -y autoremove --purge',
                    shell=True,
                    check=True,
                )
                total_time = time.time() - start_time
                print(
                    f'{Clr.BLU}BASN.INSTALL: APT UPDATES COMPLETED IN'
                    f' {total_time:.1f} SECONDS.{Clr.RST}',
                    flush=True,
                )
                break
            except Exception:
                import traceback

                print(
                    f'{Clr.BLU}BASN.INSTALL: ERROR RUNNING UPDATES...'
                    f'{Clr.RST}',
                    flush=True,
                )
                traceback.print_exc()
                if attempt > 10:
                    print(
                        f'{Clr.BLU}BASN.INSTALL: MAX RETRIES REACHED;'
                        f' FAILING.{Clr.RST}',
                        flush=True,
                    )
                    raise
                print(
                    f'{Clr.BLU}BASN.INSTALL: RETRYING IN 5 SECONDS...'
                    f'{Clr.RST}',
                    flush=True,
                )
                time.sleep(5)
                attempt += 1

    def _update_pip_stuff(self, python_version: str) -> None:
        if self._fast:
            print(
                f'{Clr.BLU}BASN.INSTALL: FAST-MODE;'
                f' SKIPPING PIP INSTALLS/UPDATES ({python_version})'
                f'.{Clr.RST}',
                flush=True,
            )
            return

        start_time = time.time()

        # NOTE: Ideally should use self._config.python_bin_path() here
        # but we don't have access to it until after we install this stuff.
        pybin = (
            f'{self._lib_root}/venv/{python_version}/bin/python{python_version}'
        )

        # Seems this bit of magic is necessary to get pip up and running
        # for some deadsnakes versions.
        subprocess.run(
            f'curl -sS https://bootstrap.pypa.io/get-pip.py | {pybin}',
            shell=True,
            check=True,
        )

        subprocess.run(
            [pybin, '-m', 'pip', 'install', '-r', 'requirements.txt'],
            check=True,
        )
        total_time = time.time() - start_time
        print(
            f'{Clr.BLU}BASN.INSTALL: PIP UPDATES ({python_version})'
            f' COMPLETED IN {total_time:.1f} SECONDS.{Clr.RST}',
            flush=True,
        )

    def _update_misc_stuff(self) -> None:
        if self._fast:
            print(
                f'{Clr.BLU}BASN.INSTALL: FAST-MODE;'
                f' SKIPPING MISC STUFF.{Clr.RST}',
                flush=True,
            )
            return
        print(
            f'{Clr.BLU}BASN.INSTALL: UPDATING MISC STUFF...{Clr.RST}',
            flush=True,
        )
        start_time = time.time()

        # What's happening here:
        # sed ... /etc/default/motd-news: this is a news-fetch
        # for 'message of the day' that runs once every 12 hours
        # or so and seems to enjoy pushing us into out-of-memory-land;
        # let's turn it off.

        # cron .. just gonna go ahead and disable this completely; our
        # lifespan is a few days at max so we can survive without running
        # any system level periodic upkeep I think.

        # tweaking /etc/security/limits.conf to enable unlimited size core
        # dumps for all users.

        # tweaking ip_local_port_range so the OS doesn't randomly assign ports
        # overlapping the ones we use (otherwise we can get rare failures
        # binding ports due to os-assigned ones being present already)

        subprocess.run(
            'systemctl disable apt-daily-upgrade.timer'
            ' && systemctl mask apt-daily-upgrade.service'
            ' && systemctl disable apt-daily.timer'
            ' && systemctl mask apt-daily.service'
            ' && systemctl disable cron'
            ' && sed -i s/ENABLED=1/ENABLED=0/g /etc/default/motd-news'
            ' && echo "* soft core unlimited" >> /etc/security/limits.conf'
            ' && echo "net.ipv4.ip_local_port_range = 48000 60999"'
            ' >> /etc/sysctl.conf',
            shell=True,
            check=True,
        )

        # Now let's tighten up ssh access.
        # I was seeing what I believe was brute force ssh login attempts
        # as root which was sucking up lots of cpu resources.
        sshd_config_path = '/etc/ssh/sshd_config'
        with open(sshd_config_path, encoding='utf-8') as infile:
            sshcfg = infile.read()
        if '#PermitRootLogin prohibit-password' in sshcfg:
            sshcfg = sshcfg.replace(
                '#PermitRootLogin prohibit-password', 'PermitRootLogin no'
            )
            with open(sshd_config_path, 'w', encoding='utf-8') as outfile:
                outfile.write(sshcfg)
        else:
            print(
                f'{Clr.RED}Did not find expected PermitRootLogin'
                f' line in sshd config.{Clr.RST}'
            )

        total_time = time.time() - start_time
        print(
            f'{Clr.BLU}BASN.INSTALL: MISC UPDATES COMPLETED IN'
            f' {total_time:.1f} SECONDS.{Clr.RST}',
            flush=True,
        )

    def _write_rc_local(self) -> None:
        # Write an rc.local that will kick off the server-manager on boot
        # (could do a fancy full init.d thing but this works for now).
        print(
            f'{Clr.BLU}BASN.INSTALL: SETTING UP RC.LOCAL...{Clr.RST}',
            flush=True,
        )

        assert self._config is not None
        pybin = self._config.python_bin_path()

        with open('/etc/rc.local', 'w', encoding='utf-8') as outfile:
            assert ' ' not in self._lib_root
            assert self._config is not None

            # Build list of flags for our run-mode.
            dev_flags: list[str] = []
            for key, val in sorted(self._config.run_mode.env_vars.items()):
                dev_flags.append(f'{key}={val}')
            dev_flags_s = ' '.join(dev_flags)
            if dev_flags_s:
                dev_flags_s += ' '

            outfile.write(
                f'#!/bin/sh -e\n'
                f'#\n'
                f'# Set up core dump location:\n'
                f'sysctl -w kernel.core_pattern=\'/tmp/core_%e.%p\'\n'
                f'\n'
                f'# Launch basn in a magic way such that we get a single'
                f' running process.\n'
                f'runuser - {self._config.user} -s /bin/sh '
                f'-c "cd {self._lib_root}/dist'
                f' && {dev_flags_s}{pybin} -m basn {self._config_path} &"\n'
                f'exit 0'
            )
        subprocess.run(['chmod', '755', '/etc/rc.local'], check=True)

    def _setup_log(self) -> None:
        assert self._config is not None
        # Set up fresh log file and preserve any existing.
        if os.path.exists(self._config.log_path):
            subprocess.run(
                ['mv', self._config.log_path, f'{self._config.log_path}.prev'],
                check=True,
            )
        Path(self._config.log_path).touch()

    def _setup_swap(self) -> None:
        # Our AWS instances come without any swap by default.
        # Intuitively this seems like what we'd want, but in practice if we
        # do happen to run out of memory it sends the machine into a 10-20
        # minute unresponsive thrashing (emptying caches/etc I assume?)
        # before the system finally (sometimes) decides to kill processes
        # and regain interactivity. So swap seems like the better option
        # in that sort of situation, especially since there are probably
        # a lot of legitimately idle pages that can be swapped out without
        # affecting our interactivity too badly. (or at least keeping us
        # interactive enough to implement our own process monitoring/killing).

        starttime = time.time()

        # This is based on:
        # https://www.linux.com/news/all-about-linux-swap-space/
        swap_path = '/swapfile'
        swapmegs = 2048
        assert ' ' not in swap_path

        # Only do this all once.
        if os.path.exists(swap_path):
            return

        # Let's go with a 2GB swap file for everything for now.. could
        # probably tune this more specifically to our available memory.
        subprocess.run(
            [
                'dd',
                'if=/dev/zero',
                f'of={swap_path}',
                'bs=1024',
                f'count={1024 * swapmegs}',
            ],
            check=True,
        )
        # mkswap recommends this (and whines if we don't).
        subprocess.run(['chmod', '600', swap_path], check=True)

        subprocess.run(['mkswap', swap_path], check=True)

        subprocess.run(
            f'echo "{swap_path} none swap sw 0 0" >> /etc/fstab',
            shell=True,
            check=True,
        )
        duration = time.time() - starttime
        print(
            f'{Clr.BLU}BASN.INSTALL: CREATED {swapmegs} MEG'
            f' SWAP FILE AT {swap_path} IN {duration:.1f} SECONDS.{Clr.RST}',
            flush=True,
        )

    def _set_permissions(self) -> None:
        assert self._config is not None
        # Make sure various things are owned by and only visible to
        # our basn user.
        for path in ['.', self._config_path, self._config.log_path]:
            subprocess.run(
                [
                    'chown',
                    '-R',
                    f'{self._config.user}:{self._config.user}',
                    path,
                ],
                check=True,
            )
            subprocess.run(['chmod', '-R', 'o-rwx', path], check=True)

        # Make sure the dirs where lib and var roots will live are
        # readable by all. (at this time we're installing to
        # /home/ubuntu and that isn't by default).
        for path in [self._config.lib_root, self._config.var_root]:
            ppath = Path(path).parent
            while str(ppath) != '/':
                subprocess.run(['chmod', 'o+rx', ppath], check=True)
                ppath = ppath.parent

    def _setup_users(self) -> None:
        assert self._config is not None
        start_time = time.time()

        # Create ourselves some users for running tasks.
        num_users = self._config.num_users
        assert num_users > 0
        existing_users = {u.pw_name for u in pwd.getpwall()}
        for num in range(num_users):
            name = f'ba{num:03d}'
            if name not in existing_users:
                subprocess.run(['useradd', name], check=True)
        total_time = time.time() - start_time

        print(
            f'{Clr.BLU}BASN.INSTALL: SET UP USERS IN'
            f' {total_time:.1f} SECONDS.{Clr.RST}',
            flush=True,
        )


if __name__ == '__main__':
    Installer().run()
