"""Kill existing basn and run newest one manually.

Strictly for iteration and debugging; do not use in production.
"""

from __future__ import annotations

import os
import sys
import subprocess
from typing import TYPE_CHECKING

import psutil

from efro.dataclassio import dataclass_from_json
from efro.terminal import Clr
from efro.error import CleanError
from baserver.data import BASNConfig

if TYPE_CHECKING:
    from typing import Dict


def main() -> None:
    """Do the thing."""

    print(f'{Clr.BLD}{Clr.MAG}Running basn.manualrun...{Clr.RST}', flush=True)

    old_cwd = os.getcwd()

    config_path = sys.argv[-1]

    with open(config_path, encoding='utf-8') as infile:
        cfg = dataclass_from_json(BASNConfig, infile.read())

    cmd = [cfg.python_bin_path(), '-m', 'basn', config_path]

    # Look for a process that appears to be our basn instance.
    processes = list(psutil.process_iter(['pid', 'cmdline', 'username']))
    processes = [p for p in processes if p.cmdline() == cmd]
    if len(processes) > 1:
        raise CleanError(
            f'Found multiple existing basn processes;'
            f' please kill manually as needed:'
            f' {[(p, p.cmdline()) for p in processes]}'
        )
    if processes:
        print(
            f'{Clr.BLD}{Clr.MAG}Killing existing basn process...{Clr.RST}',
            flush=True,
        )
        # Note: was using terminate() but that stopped working for some
        # reason.
        processes[0].kill()
        processes[0].wait()

    # NOTE: the bootstrap script gets run in python3; not necessarily
    # our exact version (since part of its job is installing our version)
    print(f'{Clr.BLD}{Clr.MAG}Installing latest basn...{Clr.RST}')
    envs = (
        f'BASN_CONFIG_PATH="{config_path}"'
        f' BASN_INSTALL_FAST=1'
        f' BASN_SUPPRESS_REBOOT=1'
    )
    subprocess.run(
        f'curl -L {cfg.bootstrap_url} | sudo {envs} python3',
        shell=True,
        check=True,
    )

    # This dir might have been remade, so chdir to the 'new' one.
    os.chdir(old_cwd)
    print(f'{Clr.BLD}{Clr.MAG}Launching basn...{Clr.RST}', flush=True)
    try:
        # Let the subprocess know we're doing this; they might want to
        # be less strict about checking memory/etc. in that case.
        env = os.environ.copy()
        env['BASN_MANUAL_RUN'] = '1'

        subprocess.run(cmd, env=env, check=True)
    except KeyboardInterrupt:
        # Keep ctrl-c from spewing ugly stack traces.
        print(f'{Clr.BLD}Got kb interrupt.{Clr.RST}', flush=True)


if __name__ == '__main__':
    try:
        main()
    except CleanError as exc:
        exc.pretty_print()
        sys.exit(1)
