"""Wrangles shared things such as game binaries."""

from __future__ import annotations

import os
import logging
import subprocess
from pathlib import Path
from dataclasses import dataclass
from typing import TYPE_CHECKING

from typing_extensions import override
from efro.dataclassio import ioprepped

from basn.resource import Resource, ResourceTypeID

if TYPE_CHECKING:
    pass


@ioprepped
@dataclass
class StandardTaskLibs(Resource):
    """Resources available to all tasks by default."""

    @override
    @classmethod
    def get_type_id(cls) -> ResourceTypeID:
        return ResourceTypeID.STANDARD_TASK_LIBS

    @override
    def get_name(self) -> str:
        return 'standard_task_libs'

    @override
    def create(self, dirpath: str) -> None:
        import efro
        import bacommon
        import baserver
        import basntask

        # FIXME FIXME FIXME - remember to not include baserver in
        # sandbox tasks since that should not be visible to end users.

        # Copy in the modules we want the task to have access to.
        # Note: using <module>.__name__ here instead of hard-coded module name
        # so type checking will warn us if anything is renamed.
        efropath = Path(efro.__file__).parent
        assert efropath.is_dir()
        bacommonpath = Path(bacommon.__file__).parent
        assert bacommonpath.is_dir()
        baserverpath = Path(baserver.__file__).parent
        assert baserverpath.is_dir()
        basntaskpath = Path(basntask.__file__).parent
        assert basntaskpath.is_dir()

        # All those modules should be coming from the same place.
        assert all(
            efropath.parent == x.parent
            for x in [bacommonpath, baserverpath, basntaskpath]
        )

        subprocess.run(
            [
                'cp',
                '-r',
                efropath,
                bacommonpath,
                baserverpath,
                basntaskpath,
                dirpath,
            ],
            check=True,
        )
        _compile_all_in_dir(dirpath)

    @override
    def get_python_paths(self) -> list[str]:
        from basn.env import app

        return [os.path.join(app.resources_root, self.get_name())]


def _compile_all_in_dir(path: str) -> None:
    """Precompile .pyc files for all .py files in a directory."""
    import py_compile

    for root, _dirs, files in os.walk(path):
        for fname in files:
            if fname.endswith('.py'):
                fnamefull = os.path.join(root, fname)

                # Generate for regular (0) and normal opt (1) optimizations.
                for optlevel in (0, 1):
                    # We should never be changing these files, so lets go with
                    # unchecked hash for quickest load times.
                    try:
                        py_compile.compile(
                            fnamefull,
                            doraise=True,
                            invalidation_mode=(
                                py_compile.PycInvalidationMode.UNCHECKED_HASH
                            ),
                            optimize=optlevel,
                        )
                    except Exception as exc:
                        logging.error(
                            'py_compile failed for %s: %s',
                            fnamefull,
                            exc,
                        )
                        raise
