"""BASN functionality for running subtasks."""

from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING

from typing_extensions import override
from efro.util import utc_now
from baserver.enums import ServerNodeTaskState

from basn.logging import logextra, balog
from basn.taskwrapper._taskwrapper import BaseTaskWrapper, StandardAction

if TYPE_CHECKING:
    import logging
    import datetime
    from typing import Callable, Any

    from efro.log import LogLevel

    from basn.taskset import TaskSet
    from basn.resource import Resource


class InlineTaskWrapper(BaseTaskWrapper):
    """A task that runs inline as part of basn."""

    def __init__(self, logger: logging.Logger = balog) -> None:
        super().__init__(logger=logger)
        self._aiotask: asyncio.Task | None = None

    @override
    def on_launch(self) -> None:
        self.logger.info(
            'Spawning %s inline task %s.',
            type(self).__name__,
            self._task_id,
            extra=logextra(task=self),
        )

    @override
    def get_actions(self) -> list[str]:
        actions: list[str] = []
        if self.state is ServerNodeTaskState.RUNNING:
            actions.append(StandardAction.KILL.value)
        return actions

    @override
    async def run(self) -> None:
        assert self._task_set is not None
        task_set = self._task_set()
        assert self._launch_time is not None

        self._aiotask = asyncio.create_task(
            self.run_inline(), name=f'basn task {self._task_id} inline'
        )

        # By flipping our state to RUNNING we're declaring we're
        # prepared to handle actions such as KILL/etc.
        self._state = ServerNodeTaskState.RUNNING
        if task_set is not None:
            task_set.task_is_running(self)

        try:
            await self._aiotask
            self._result_success = True
        except asyncio.CancelledError:
            # Assume this was our child task and absorb it. We don't
            # expect/support our calling task getting canceled.
            if not self._aiotask.done():
                self.logger.exception(
                    'Got CancelledError but our inline task'
                    ' is not cancelled; unexpected.'
                )
            self._result_success = False
            self.add_note('Asyncio task was canceled.')
        except Exception:
            import traceback

            self.add_note(
                'Error running inline task command:\n' + traceback.format_exc()
            )
            self._result_success = False

        self._aiotask = None
        self._duration = (utc_now() - self._launch_time).total_seconds()

    async def run_inline(self) -> None:
        """Run an inline task; should be overridden by subclasses."""

    @override
    def do_force_kill(self) -> None:
        try:
            assert self._aiotask is not None
            self._aiotask.cancel()
        except Exception:
            self.logger.exception(
                'Error in do_force_kill.',
                extra=logextra(task=self),
            )
