"""BASN functionality for running subtasks."""

from __future__ import annotations

import logging
import asyncio
from typing import TYPE_CHECKING

from typing_extensions import override
from efro.util import utc_now
from baserver.enums import ServerNodeTaskState

from basn.taskwrapper._taskwrapper import BaseTaskWrapper, StandardAction
from basn.logging import balog

if TYPE_CHECKING:
    import datetime
    from typing import Callable
    from efro.log import LogLevel
    from basn.taskset import TaskSet
    from basn.resource import Resource
    from typing import Any


class SandboxTaskWrapper(BaseTaskWrapper):
    """A task that runs in a sandbox."""

    def __init__(self, logger: logging.Logger = balog) -> None:
        super().__init__(logger=logger)
        self._aiotask: asyncio.Task | None = None

    @override
    def get_needs_user_id(self) -> bool:
        return True

    @override
    def get_needs_directory(self) -> bool:
        return True

    @override
    def on_launch(self) -> None:
        logging.info(
            'Spawning %s sandbox task %s.', type(self).__name__, self._task_id
        )

    async def run_sandbox(self) -> None:
        """Run the sandbox task."""
        self.add_note('HELLO WORLD FROM SANDBOX')
        await asyncio.sleep(5)
        self.add_note('GOODBYE WORLD FROM SANDBOX')

    @override
    def get_actions(self) -> list[str]:
        actions: list[str] = []
        if self.state is ServerNodeTaskState.RUNNING:
            actions.append(StandardAction.KILL.value)
        return actions

    @override
    async def run(self) -> None:
        assert self._task_set is not None
        task_set = self._task_set()
        assert self._launch_time is not None
        self._state = ServerNodeTaskState.RUNNING

        if task_set is not None:
            task_set.task_is_running(self)
        try:
            self._aiotask = asyncio.create_task(
                self.run_sandbox(),
                name=f'basn sandbox task {self._task_id} wrapper',
            )
            await self._aiotask
            self._result_success = True
        except asyncio.CancelledError:
            self._result_success = False
            self.add_note('aiotask was canceled.')
        except Exception:
            import traceback

            self.add_note(
                'Error running sandbox task:\n' + traceback.format_exc()
            )
            self._result_success = False

        self._aiotask = None
        self._duration = (utc_now() - self._launch_time).total_seconds()

    @override
    async def cleanup(self) -> None:
        self.add_note('HELLO FROM SANDBOX CLEANUP')

    @override
    def do_force_kill(self) -> None:
        try:
            if self._aiotask is not None:
                self.add_note('do_force_kill: canceling aiotask...')
                self._aiotask.cancel()
        except Exception:
            logging.exception('Error in do_force_kill')
