"""BASN functionality for running subtasks."""

from __future__ import annotations

import os
import sys
import time
import shlex
import asyncio
import threading
import subprocess
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, assert_never

from typing_extensions import override
import psutil
from efro.dataclassio import dataclass_from_json, dataclass_to_json
from efro.util import utc_now
from efro.log import LogEntry
from baserver.enums import ServerNodeTaskState
from baserver.msg import bamastertobasn
from baserver.util import current_process_username
from basntask.taskenv import TaskEnv
from basntask.result import TaskResult

from basn.logging import logextra, balog
from basn.resource.standardtasklibs import StandardTaskLibs
from basn.taskwrapper._taskwrapper import BaseTaskWrapper, StandardAction

if TYPE_CHECKING:
    import logging
    import datetime
    from typing import Callable, Any

    from efro.log import LogLevel
    from basntask.command import TaskCommand

    from basn.taskset import TaskSet
    from basn.resource import Resource


class StandaloneTaskWrapper(BaseTaskWrapper):
    """Task that runs in its own standalone process."""

    class ProcessPriority(Enum):
        """The priority we run processes with."""

        HIGH = 'high'
        MEDIUM = 'medium'
        LOW = 'low'

    def __init__(self, logger: logging.Logger = balog) -> None:
        super().__init__(logger=logger)
        self._log_path: Path | None = None

        self._process: asyncio.subprocess.Process | None = None
        self._log_follow_thread: threading.Thread | None = None
        self._task_result: TaskResult | None = None

    def get_process_priority(self) -> ProcessPriority:
        """How high priority should this task's process be?"""
        return self.ProcessPriority.MEDIUM

    @override
    def get_needs_user_id(self) -> bool:
        return True

    @override
    def get_needs_directory(self) -> bool:
        return True

    @override
    def get_creates_task_directory(self) -> bool:
        return True

    @override
    def on_launch(self) -> None:

        assert self._root is not None
        self._log_path = Path(self._root, 'basntask.log')
        self.logger.info(
            'Spawning %s task at %s.',
            type(self).__name__,
            self._user_id,
            extra=logextra(task=self),
        )

        # Set up a thread to follow our log until we're done running.
        self._log_follow_thread = threading.Thread(target=self._follow_log)
        self._log_follow_thread.start()

    @override
    async def run(self) -> None:
        # pylint: disable=too-many-statements

        # Note: this implementation of run() simply sets _task_result by
        # constructing & running our command. It is valid for subclasses
        # to ignore or augment this implementation and supply their own
        # _task_result instead.

        assert self._task_result is None

        assert self._task_set is not None
        task_set = self._task_set()

        starttime = time.monotonic()
        try:
            cmd_args = await self._prep_for_exec()
        except Exception:
            self.logger.exception(
                'Error in _prep_for_exec()', extra=logextra(task=self)
            )
            self._task_result = TaskResult(error='Error in task prep.')
            return

        prep_duration = time.monotonic() - starttime
        self.logger.info(
            'Prepped standalone task %d in %.2fs.',
            self._task_id,
            prep_duration,
            extra=logextra(),
        )

        # Add in a few env vars to the raw command so it knows where
        # to create itself and what user to run as.
        assert self._root is not None

        # Note: we are currently merging stdout and stderr to a single
        # output pipe since that is usually more useful when viewing
        # manually. If we ever need to parse exact output of one or
        # the other we'll need to revisit this.

        self._process = await asyncio.create_subprocess_exec(
            *cmd_args,
            cwd=self._root,
            stdin=asyncio.subprocess.DEVNULL,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.STDOUT,
        )

        # The process we launch should immediately detach itself
        # from stdout/etc. so this should return immediately.
        output, _no_stderr = await self._process.communicate()
        if output:
            self.add_note(f'launcher output: {output.decode()}')

        # At this point, the spun off subprocess should be writing a
        # .pid file for us. Wait for that.
        pidstarttime = time.monotonic()
        pid: int | None = None
        waitstart = time.monotonic()
        while True:
            if time.monotonic() - pidstarttime > 10.0:
                self.add_note('TaskWrapper timed out waiting for pid file.')
                break

            pidfilepath = os.path.join(self._root, '.basntask.pid')
            if os.path.exists(pidfilepath):
                try:
                    with open(pidfilepath, encoding='utf-8') as infile:
                        pid = int(infile.read())
                    break

                # Seems there might be a race condition where we read
                # the file after its created but before the number is
                # written. ..so keep trying in that case.
                #
                # UPDATE: This should be fixed now so can probably
                # remove this. Though I suppose it doesn't hurt to have
                # extra protection.

                except Exception:
                    self.add_note(
                        'Error reading task id; will keep trying'
                        ' (see log for details).'
                    )
                    self.logger.exception(
                        'Error reading task id file; will keep trying.',
                        extra=logextra(task=self),
                    )
                    await asyncio.sleep(1.0)

            # Spin fast to start with so speedy stuff goes through
            # immediately but back off over time to lessen load.
            sleeptime = max(
                0.01, min((time.monotonic() - waitstart) * 0.2, 1.0)
            )
            await asyncio.sleep(sleeptime)

        # By flipping our state to RUNNING we are declaring we're
        # prepared to handle actions such as TERMINATE/KILL/etc.
        self._state = ServerNodeTaskState.RUNNING
        if task_set is not None:
            task_set.task_is_running(self)

        if pid is not None:
            # Ok; we've got a process id. Now we wait for it to end.

            waitstart = time.monotonic()
            while True:
                if not psutil.pid_exists(pid):
                    break

                # Spin fast to start with so speedy stuff goes through
                # immediately but back off over time to lessen load.
                # Theoretically maybe there's a more exact way to wait.
                # os.waitpid() doesn't seem to work since the process is
                # not a child process of ours. Perhaps we could look
                # into fs events and get notified exactly when the
                # result file gets written. Though technically most
                # actions happening as a result of tasks should get
                # kicked off by the task itself so there's not a huge
                # rush to finish on our end.
                sleeptime = max(
                    0.01, min((time.monotonic() - waitstart) * 0.2, 2.14)
                )
                await asyncio.sleep(sleeptime)

        assert self._launch_time is not None
        self._duration = (utc_now() - self._launch_time).total_seconds()

        # We run the actual task detached from its parent process so
        # that we don't get a ton of sudo/runuser/runtask processes
        # cluttering things up, but this means we can't directly read a
        # return code or output. Instead we have it write a file that we
        # look for.
        result_path = os.path.join(self._root, '.basntask.result')
        if not os.path.exists(result_path):
            # Ideally tasks should catch errors and try to give us
            # meaningful results.
            self._task_result = TaskResult(
                error=('Task failed but did not set an error. Check task logs.')
            )
        else:
            with open(result_path, encoding='utf-8') as infile:
                self._task_result = dataclass_from_json(
                    TaskResult, infile.read()
                )

    @override
    async def cleanup(self) -> None:

        # We expect a result to always be in place here. Pull our
        # overall success/fail state from that.
        if self._task_result is None:
            raise RuntimeError('_result should always be set by cleanup time.')

        self._result_success = self._task_result.error is None
        if self._task_result.error is not None:
            self.add_note(f'Task failed with error: {self._task_result.error}')

        # Wait for our log follower thread to finish up.
        if self._log_follow_thread is not None:
            await asyncio.get_running_loop().run_in_executor(
                None, self._log_follow_thread.join
            )

    def get_command(self) -> TaskCommand:
        """Return the command a standalone task should run.

        Should be overridden by subclasses.
        """
        raise NotImplementedError()

    @override
    def get_actions(self) -> list[str]:
        actions: list[str] = [StandardAction.SSH_TASK_DIR.value]
        if self.state is ServerNodeTaskState.RUNNING:
            actions.append(StandardAction.TERMINATE.value)
            actions.append(StandardAction.KILL.value)
        return actions

    @override
    def get_task_info_command(self) -> str:
        # What we show in the ui.
        return str(self.get_command())

    @override
    def do_terminate(self) -> None:
        from efro.error import CleanError

        pid: str | None = None

        # basntask dumps out a file with its process id; look for that.
        # (self._process gives us a parent shell id which we can't use)
        try:
            assert self._root is not None
            pidfilepath = os.path.join(self._root, '.basntask.pid')
            with open(pidfilepath, encoding='utf-8') as infile:
                pid = infile.read()
        except Exception as exc:
            raise CleanError('Unable to read pid file.') from exc

        # These run as other users, so we have to bust out sudo.
        subprocess.run(['sudo', 'kill', pid], check=False)

    @override
    def do_force_kill(self) -> None:
        from basn.util import force_kill_user_processes

        assert self._user_id is not None
        force_kill_user_processes(
            user_id=self._user_id, for_task=self, warn_if_running=False
        )

    def _follow_log(self) -> None:
        from basn.env import app

        assert self._log_path is not None
        while not self._log_path.exists():
            # Our task might go down without writing anything; no
            # biggie.
            if self.state not in (
                ServerNodeTaskState.UNKNOWN,
                ServerNodeTaskState.LAUNCHING,
                ServerNodeTaskState.RUNNING,
            ):
                return
            time.sleep(0.5)

        with self._log_path.open(encoding='utf-8') as infile:
            while True:
                line = ''
                while True:
                    linebit = infile.readline()
                    line += linebit
                    if line.endswith('\n'):
                        break
                    if self.state not in (
                        ServerNodeTaskState.UNKNOWN,
                        ServerNodeTaskState.LAUNCHING,
                        ServerNodeTaskState.RUNNING,
                    ):
                        return
                    time.sleep(1.0358)

                # Feed this entry to the asyncio loop for processing.
                entry = dataclass_from_json(LogEntry, line)
                app.loop.call_soon_threadsafe(
                    lambda: self.log_level_callback(entry.level)
                )

    def _get_log(
        self,
        max_size: int | None,
        min_log_level: LogLevel,
        contains: str | None,
    ) -> tuple[str, int]:
        """Return this task's current log contents (and lines truncated)."""
        from basn.util import read_lines_filtered

        assert self._log_path is not None
        if self._log_path.is_file():
            with self._log_path.open(encoding='utf-8') as infile:
                # Fast path; just give whole thing.
                if max_size is None and contains is None:
                    log = infile.read()
                    assert isinstance(log, str)
                    return log, 0
                # Otherwise will need to interpret contents.
                return read_lines_filtered(
                    infile,
                    max_size=max_size,
                    min_log_level=min_log_level,
                    contains=contains,
                )
        self.logger.debug(
            'Task.get_log(): no log found; returning empty string.',
            extra=logextra(task=self),
        )
        return '', 0

    async def _prep_for_exec(
        self,
    ) -> list[str]:
        """Set up to launch a standalone task and return exec args.

        Creates the task dir, sets permissions, populates task resources,
        and anything else needing doing.
        """
        # pylint: disable=too-many-locals
        # pylint: disable=too-many-statements
        import efro

        from basn.env import app, config

        efropath = Path(efro.__file__).parent
        assert efropath.is_dir()
        task_root = self._root
        assert task_root is not None
        task_user = self._user_id
        assert task_user is not None

        task_env = TaskEnv(
            ba_server_path=None,
            command=self.get_command(),
            run_mode=config.run_mode,
        )

        assert not os.path.exists(task_root)
        os.mkdir(task_root)

        resources = self.get_resources()

        # Watch for repeats.
        if len({r.get_name() for r in resources}) != len(resources):
            raise RuntimeError(f'Duplicate resources provided: {resources}')

        resources_root = app.resources_root
        assert resources_root is not None

        # Resources sanity checks. Should always contain a single
        # StandardTaskResources instance.
        if not (
            len([r for r in resources if isinstance(r, StandardTaskLibs)]) == 1
        ):
            tps = [type(obj).__name__ for obj in resources]
            raise RuntimeError(
                f'Expected exactly 1 StandardTaskLibs in resources; got {tps}.'
            )

        # Prep all needed resources.
        assert app.resources_root and os.path.isdir(app.resources_root)
        async with asyncio.TaskGroup() as task_group:
            for resource in resources:
                _ = task_group.create_task(resource.prepare_async())

        # If any resource provides a server binary, pass that along.
        for resource in resources:
            # pylint: disable=assignment-from-none
            baserverpath = resource.get_ba_server_path()
            if baserverpath is not None:
                if task_env.ba_server_path is not None:
                    raise RuntimeError(
                        'Multiple resources providing ba_server_path'
                    )
                task_env.ba_server_path = baserverpath

        # Gather Python paths provided by our resources.
        pypaths: list[str] = []
        for resource in resources:
            pypaths += resource.get_python_paths()
        assert pypaths

        # Set everything in the task dir to be owned by the task owner (with
        # us as group so we can poke around too). Also do setgid on the task
        # root dir so new stuff created under it gets basn_user group by
        # default.

        def _set_ownership() -> None:
            basn_user = current_process_username()
            subprocess.run(['chmod', '-R', 'o-rwx', task_root], check=True)
            subprocess.run(['chmod', 'g+s', task_root], check=True)
            subprocess.run(
                ['sudo', 'chown', '-R', f'{task_user}:{basn_user}', task_root],
                check=True,
            )

        await asyncio.get_running_loop().run_in_executor(None, _set_ownership)

        ppjoined = ':'.join(pypaths)

        # Note: We set up to run backgrounded (&) so that we don't
        # clutter up our process space with a bunch of intermediate
        # sudo/runtask/runuser tasks. The downside is that we don't get
        # the pid or result of the task here so we need the wrapper to
        # communicate it to us in other ways (namely a pid file).

        envdata = shlex.quote(dataclass_to_json(task_env, sort_keys=True))

        # Dump stdout/stderr to a file.
        redirect = f' > {task_root}/basntask.output 2>&1'

        # When redirecting to a file we need to explicitly enable line
        # buffering or we won't see most output until successful program
        # completion (which often won't happen when they are killed
        # explicitly).
        buf = 'stdbuf -oL -eL '

        # Let's put all of our process priorities below ourself (0).
        # We'd rather starve a random bg job or game than everyone
        # connected to this basn node.
        niceval: int
        priority = self.get_process_priority()
        if priority is self.ProcessPriority.HIGH:
            niceval = 1
        elif priority is self.ProcessPriority.MEDIUM:
            niceval = 5
        elif priority is self.ProcessPriority.LOW:
            niceval = 15
        else:
            # Make sure we cover all options.
            assert_never(priority)

        nice: list[str] = ['nice', '-n', str(niceval)]

        # Use the same Python executable we were run with.
        return (
            ['sudo', 'runuser', '-u', task_user, '--']
            + nice
            + [
                '/bin/sh',
                '-c',
                f'PYTHONPATH={ppjoined}'
                ' BASNLOG=1'
                ' BA_LIFECYCLE_LOG=1'
                f' {buf} {sys.executable} -m basntask {envdata}{redirect}&',
            ]
        )

    @override
    def handle_log_query_message(
        self, msg: bamastertobasn.LogQueryMessage
    ) -> bamastertobasn.LogQueryResponse:

        tasklog, trunclines = self._get_log(
            max_size=msg.max_size,
            min_log_level=msg.min_log_level,
            contains=msg.contains,
        )

        return bamastertobasn.LogQueryResponse(
            log=tasklog, truncated_lines=trunclines
        )
