"""BASN functionality for running subtasks."""

from __future__ import annotations

import os
import time
import asyncio
import weakref
import datetime
import subprocess
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, TypeVar

from efro.error import CleanError
from efro.util import utc_now
from efro.log import LogLevel
from baserver.msg import bamastertobasn
from baserver.enums import ServerNodeTaskState
from baserver.util import async_run, current_process_username

from basn.logging import logextra

if TYPE_CHECKING:
    import logging
    import datetime
    from pathlib import Path
    from typing import Callable
    from basn.taskset import TaskSet
    from basn.resource import Resource
    from typing import Any

T = TypeVar('T')


class StandardAction(Enum):
    """Actions applicable to most tasks."""

    SSH_TASK_DIR = 'SSH Task Dir'
    TERMINATE = 'Terminate'
    KILL = 'Kill'


class BaseTaskWrapper:
    """Encapsulates and wrangles a basn task in the parent process."""

    # pylint: disable=too-many-public-methods

    def __init__(self, logger: logging.Logger) -> None:
        self.logger = logger

        self._user_id: str | None = None
        self._task_id: int | None = None
        self._launch_time: datetime.datetime | None = None
        self._state = ServerNodeTaskState.UNKNOWN
        self._notes: str | None = None
        self._force_kill_task: asyncio.Task | None = None

        # Set to True or False on completion of the underlying task.
        self._result_success: bool | None = None
        self._reached_max_duration = False
        self._error_flag = False
        self._user_killing_task = False
        self._max_log_level = LogLevel.INFO
        self._duration: float | None = None
        self._root: Path | None = None
        self._task_set: weakref.ReferenceType[TaskSet] | None = None
        self._udp_port: int | None = None
        self._run_task: asyncio.Task | None = None

    def launch(
        self,
        taskset: TaskSet,
        user_id: str | None,
        task_id: int,
        udp_port: int | None,
    ) -> None:
        """Actually launch this task."""
        from basn.env import app

        # Avoid reference cycles by weakly referencing our owner.
        self._task_set = weakref.ref(taskset)

        self._udp_port = udp_port
        self._user_id = user_id
        self._task_id = task_id
        self._launch_time = utc_now()

        # If they want a dir, set that up.
        if self.get_needs_directory():
            assert self._user_id is not None
            self._root = Path(app.tasks_root, str(self._task_id))
            assert ' ' not in str(self._root)

            # FIXME: Do we even want/need this? Currently only
            # standalone needs a task directory and it sets it up
            # itself.
            if not self.get_creates_task_directory():
                os.mkdir(self._root)
                # Set everything in the task dir to be owned by the task owner
                # (with the basn user as group so he can poke around too).
                # Also do setgid on the task root dir so stuff created under it
                # gets basn_user group by default.
                basn_user = current_process_username()
                subprocess.run(['chmod', '-R', 'o-rwx', self._root], check=True)
                subprocess.run(['chmod', 'g+s', self._root], check=True)
                subprocess.run(
                    [
                        'sudo',
                        'chown',
                        '-R',
                        f'{self._user_id}:{basn_user}',
                        self._root,
                    ],
                    check=True,
                )

        self.on_launch()

        # And finally launch the actual task.
        assert asyncio.get_running_loop() is app.loop
        self._run_task = asyncio.create_task(
            self._prep_run_and_cleanup(),
            name=f'basn task {task_id} wrapper',
        )

    def on_launch(self) -> None:
        """Handle launching stuff."""
        raise RuntimeError('This should be overridden.')

    async def minute_upkeep(self) -> None:
        """Called once per minute while the task is running."""
        assert self._launch_time is not None
        duration = utc_now() - self._launch_time
        maxduration = self.get_max_duration()
        if (
            not self._reached_max_duration
            and duration > self.get_max_duration()
        ):
            self._reached_max_duration = True
            self.logger.error(
                (
                    'Task %s has been running for max-duration %s;'
                    ' obliterating it.',
                ),
                self._task_id,
                maxduration,
                extra=logextra(task=self),
            )
            self.add_note(
                f'Force-killing due to max-duration {maxduration} elapsed.'
            )
            self.do_force_kill()

    @property
    def user_id(self) -> str | None:
        """User-id associated with this task."""
        return self._user_id

    @property
    def udp_port(self) -> int | None:
        """Assigned udp port for this task."""
        return self._udp_port

    def has_task_id(self) -> bool:
        """Have we been assigned a task id?"""
        return self._task_id is not None

    @property
    def task_id(self) -> int:
        """Numeric id associated with this task."""
        assert self._task_id is not None
        return self._task_id

    @property
    def label(self) -> str:
        """Label for this task."""
        return self.get_label()

    @property
    def load(self) -> float:
        """Load for this task."""
        return self.get_load()

    @property
    def launch_time(self) -> datetime.datetime:
        """User-id associated with this task."""
        assert self._launch_time is not None
        return self._launch_time

    @property
    def state(self) -> ServerNodeTaskState:
        """Current task state."""
        return self._state

    @property
    def duration(self) -> float | None:
        """Task duration (once finished)."""
        return self._duration

    @property
    def max_log_level(self) -> LogLevel:
        """Current flagged task log level."""
        return self._max_log_level

    @property
    def error_flag(self) -> bool:
        """Whether we're flagged as errored."""
        return self._error_flag

    def get_needs_user_id(self) -> bool:
        """Return whether this task requires a user allocated to it."""
        return False

    def get_needs_directory(self) -> bool:
        """Return whether this task requires a directory."""
        return False

    def get_creates_task_directory(self) -> bool:
        """Whether the task sets up the task directory itself."""
        return False

    def get_needs_udp_port(self) -> bool:
        """Returns whether this task requires a ba port allocated to it."""
        return False

    def get_label(self) -> str:
        """Return a label for the task to be shown to a user."""
        return 'untitled'

    def get_max_duration(self) -> datetime.timedelta:
        """Return the max duration for this task in seconds.

        This is safety measure; the task should never actually run this
        long. If it does, an alarm will be raised and it may be forcibly
        killed.
        """
        return datetime.timedelta(hours=12)

    def get_resources(self) -> list[Resource]:
        """Return resources needed by this task."""
        from basn.resource.standardtasklibs import StandardTaskLibs

        return [StandardTaskLibs()]

    def get_load(self) -> float:
        """Return the universal load for this task."""
        return 1.0

    async def _prep_run_and_cleanup(self) -> None:
        """Run the full task lifecycle."""

        try:
            self._state = ServerNodeTaskState.LAUNCHING
            assert self._task_set is not None
            task_set = self._task_set()

            # Run!!
            try:
                await self.run()
            except (Exception, asyncio.CancelledError):
                # run() should handle its own errors; if we wind up here
                # consider it damage control and try to soldier on.
                self.logger.exception(
                    'Unexpected error during task run.',
                    extra=logextra(task=self),
                )

            self._state = ServerNodeTaskState.CLEANING_UP
            self.logger.debug(
                'Task %s exited with success %s.',
                self._task_id,
                self._result_success,
                extra=logextra(task=self),
            )

            # Clean up!!
            cleanup_errored = False
            try:
                await self.cleanup()
                if self._root is not None:
                    # Move files back under our sole control.
                    uname = current_process_username()
                    await async_run(
                        f'sudo chown -R {uname}:{uname} {self._root}'
                    )
            except (Exception, asyncio.CancelledError):
                # cleanup() should handle its own errors; if we wind up
                # here consider it damage control and try to soldier on.
                cleanup_errored = True
                self.logger.exception(
                    'Unexpected error during cleanup() for task %s.',
                    self,
                    extra=logextra(task=self),
                )

            self._state = (
                # Always show 'errored' if we timed out or failed in
                # cleanup.
                ServerNodeTaskState.ERRORED
                if (self._reached_max_duration or cleanup_errored)
                # Next go with KILLED if the user explicitly ended us.
                else (
                    ServerNodeTaskState.KILLED
                    if self._user_killing_task
                    # Outside of those cases, we're based on
                    # _result_success.
                    else (
                        ServerNodeTaskState.COMPLETED
                        if self._result_success
                        else ServerNodeTaskState.ERRORED
                    )
                )
            )

            # If the task has errored, mark it as such and push that
            # back to master if need be.
            if self._state == ServerNodeTaskState.ERRORED:
                assert not self._user_killing_task
                self._error_flag = True
                if task_set is not None:
                    task_set.task_errored(self)

            # Let the task-set know we're done so it can recycle the
            # user/etc.
            if task_set is not None:
                task_set.task_finished(self)

        except (Exception, asyncio.CancelledError):
            # run() should handle its own errors; if we wind up here
            # consider it damage control and try to soldier on.
            self.logger.exception(
                'Unexpected error during task _prep_run_and_cleanup.',
                extra=logextra(task=self),
            )
            self.add_note('_prep_run_and_cleanup got cancelled; unexpected.')

    async def run(self) -> None:
        """Do the thing.

        This may set the task's state to RUNNING at some point,
        though this is not required in the case of errored setup/etc.
        The state will be set to CLEANING_UP as soon as this returns.
        """
        raise NotImplementedError()

    async def cleanup(self) -> None:
        """Runs after the task completes."""

    def do_terminate(self) -> None:
        """Payload of task soft killing. Should be overridden if implemented.

        This can only be called when the task is in the RUNNING state.
        """

    def do_force_kill(self) -> None:
        """Payload of task hard-killing. Should be overridden.

        This can only be called when the task is in the RUNNING state."""
        raise NotImplementedError()

    def get_actions(self) -> list[str]:
        """Return a list of currently-relevant UI actions.

        Subclasses should override this to add their own (and generally call
        the super method to add common ones).
        """
        actions: list[str] = []
        return actions

    def add_note(self, text: str) -> None:
        """Add a line to the task's note (a sort of inline log).

        This is currently raw text; not html.
        """
        if self._notes is None:
            self._notes = ''
        elif not self._notes.endswith('\n'):
            self._notes += '\n'
        timestamp = (
            (utc_now() - self._launch_time).total_seconds()
            if self._launch_time is not None
            else 0.0
        )
        self._notes += f'{timestamp:.2f}s: {text}'

    def force_kill(self) -> None:
        """Flags the task to be force-killed as soon as possible."""

        # All we do here is spin up an asyncio task to handle this (it
        # may involve waiting so we can't do it all inline here).
        if self._force_kill_task is not None:
            pass
        self._force_kill_task = asyncio.create_task(
            self._do_force_kill(),
            name=f'basn task {self.task_id} force-kill',
        )

    async def _do_force_kill(self) -> None:
        starttime = time.monotonic()

        # It's possible a kill can happen before the task is up and
        # running. If that's the case, wait for it to reach the running
        # state and then do it.
        while self.state in {
            ServerNodeTaskState.UNKNOWN,
            ServerNodeTaskState.LAUNCHING,
        }:
            await asyncio.sleep(0.1)
            if time.monotonic() - starttime > 10.0:
                self.logger.exception(
                    'Timed out waiting for task to reach running'
                    ' state to kill it.'
                )
                break

        if self.state is ServerNodeTaskState.RUNNING:
            self._user_killing_task = True
            self.add_note('Explicitly killed by force_kill().')
            self.do_force_kill()

    def action(self, action: str) -> bamastertobasn.TaskActionResponse:
        """Run some action on a task as requested via UI.

        Subclasses should override this to handle actions they recognize
        and call the super implementation to handle common ones.
        Be aware that any action previously returned from get_actions()
        can be passed here and handled gracefully, even if it is no longer
        relevant.
        """

        try:
            actionval: StandardAction | None = StandardAction(action)
        except ValueError:
            actionval = None

        if actionval is StandardAction.SSH_TASK_DIR:
            assert self._root is not None

            cmd = f'__BASE_SSH_TTY_COMMAND__ "cd {self._root} && bash"'
            return bamastertobasn.TaskActionResponse(
                output=cmd,
                copy_text=cmd,
                auto_return=False,
            )

        # Ruthlessly insta-kill the task (and any other sharing its user-id).
        if actionval is StandardAction.KILL:
            if self.state is not ServerNodeTaskState.RUNNING:
                return bamastertobasn.TaskActionResponse(
                    output='Task is not running.'
                )

            # So this shows up as 'killed' not 'errored'.
            self._user_killing_task = True
            self.add_note(f'Sent explicit {actionval.value}.')
            self.do_force_kill()
            return bamastertobasn.TaskActionResponse(output='Killed task.')

        # Send a terminate signal to the task (allows it a possible
        # graceful exit).
        if actionval is StandardAction.TERMINATE:
            if self.state is not ServerNodeTaskState.RUNNING:
                return bamastertobasn.TaskActionResponse(
                    output='Task is not running.'
                )

            # So this shows up as 'killed' not 'errored'.
            self._user_killing_task = True
            self.add_note(f'Sent explicit {actionval.value}.')
            self.do_terminate()
            return bamastertobasn.TaskActionResponse(
                output='Requested task termination.'
            )

        raise CleanError(f"Action type '{action}' was not handled.")

    def log_level_callback(self, level: LogLevel) -> None:
        """Called by our log handler to inform us of log-levels."""
        from basn.env import app

        # Update our own personal max-level.
        if level.value > self._max_log_level.value:
            self._max_log_level = level

        # Also update the app-wide one if we top it.
        if level.value > app.stats.max_task_log_level.value:
            app.stats.max_task_log_level = level
            app.mark_stats_dirty()

    def handle_log_query_message(
        self, msg: bamastertobasn.LogQueryMessage
    ) -> bamastertobasn.LogQueryResponse:
        """Handle a message asking for logs."""
        del msg  # Unused.
        return bamastertobasn.LogQueryResponse(
            log='',
            truncated_lines=0,
            error='This task provides no log. See the global basn log.',
        )

    def reset_errors(self) -> None:
        """Reset any error flags."""
        self._error_flag = False
        self._max_log_level = LogLevel.INFO

    def get_task_info_command(self) -> str:
        """Return a 'command' to show in the web UI.

        Only relevant for standalone tasks.
        """
        return str(self)

    def get_task_info_message(self) -> bamastertobasn.TaskQueryResponse:
        """Construct a message describing this task."""

        stdout_stderr: str | None = None
        if self._root is not None:

            # Where we expect any task stdout/stderr combined output to
            # go.
            outputpath = os.path.join(self._root, 'basntask.output')
            if os.path.exists(outputpath):
                with open(outputpath, 'r', encoding='utf-8') as infile:
                    stdout_stderr = infile.read()

        return bamastertobasn.TaskQueryResponse(
            task_id=self._task_id,
            user_id=self._user_id,
            root=str(self._root) if self._root is not None else None,
            label=self.label,
            command=self.get_task_info_command(),
            state=self._state,
            stdout_stderr=stdout_stderr,
            udp_port=self._udp_port,
            exit_success=self._result_success,
            duration=self._duration,
            notes=self._notes,
        )
