"""Task class for test task."""

from __future__ import annotations

import os
import asyncio
from typing import TYPE_CHECKING

from typing_extensions import override

from basn.resource.standardtasklibs import StandardTaskLibs
from basn.resource.cloudfile import CloudFile
from basn.taskwrapper import (
    StandaloneTaskWrapper,
)
from basn.logging import logextra

if TYPE_CHECKING:
    from typing import Any

    from baserver.cloudbuild import RecipeTypeID
    from baserver.cloudbuild import Recipe, RecipeComponent
    from basntask.command import TaskCommand
    from basntask.cloudbuild import CloudBuildTaskResults

    from basn.resource import Resource


class CloudBuildTaskWrapper(StandaloneTaskWrapper):
    """Build stuff.."""

    def __init__(
        self,
        recipe_type_id: RecipeTypeID,
        input_data: RecipeComponent,
        master_server_build_id: str,
        master_server_run_id: str,
        load: float,
        extra: bytes | None,
    ) -> None:
        from basn.logging import cloudbuildlog

        super().__init__(logger=cloudbuildlog)
        self._recipe_type_id = recipe_type_id
        self._input_data = input_data
        self._master_server_build_id = master_server_build_id
        self._master_server_run_id = master_server_run_id
        self._load = load
        self._extra = extra

    @override
    def get_process_priority(self) -> StandaloneTaskWrapper.ProcessPriority:
        # Do our best to not affect any running games/etc.
        return StandaloneTaskWrapper.ProcessPriority.LOW

    @override
    def get_load(self) -> float:
        return self._load

    @override
    def get_label(self) -> str:
        return f'cloud-build {self._recipe_type_id.value}'

    @override
    def get_resources(self) -> list[Resource]:
        return [StandardTaskLibs()] + [
            CloudFile(cloud_file_id=fid)
            for fid in self._input_data.get_cloud_file_ids()
        ]

    @override
    def get_command(self) -> TaskCommand:
        from basntask.cloudbuild import CloudBuildCommand

        return CloudBuildCommand(
            recipe_type_id=self._recipe_type_id,
            input_data=self._input_data,
            input_cloud_file_paths={
                fid: CloudFile(cloud_file_id=fid).get_local_path()
                for fid in self._input_data.get_cloud_file_ids()
            },
            extra=self._extra,
        )

    @override
    async def run(self) -> None:
        from baserver.msg import basntobamaster
        from baserver.cloudbuild import BuildErrorInfo
        from basntask.cloudbuild import CloudBuildTaskResults
        from basntask.result import TaskResult

        from basn.env import app

        # Do the normal standalone process thing.
        try:
            await super().run()
            assert self._task_result is not None
        except Exception:
            # run() should always catch exceptions and set _task_result,
            # but watch for Exceptions bubbling up just to be safe.
            self.logger.exception(
                'CloudBuild run() raised an exception instead of setting'
                ' _task_result; this should not happen.',
                extra=logextra(task=self),
            )
            self._task_result = TaskResult(error='Internal task error.')

        # Attempt to ship results or error back to bamaster. For now,
        # doing nothing if the send attempt fails. They can contact us
        # and ask about it later if they want.
        if self._task_result.error is None:
            # Try to ship completed build output.
            try:
                build_output = self._task_result.output
                assert isinstance(build_output, CloudBuildTaskResults)
                self.logger.debug(
                    'Sending completed build results to bamaster...',
                    extra=logextra(task=self),
                )
                await self._send_completed_build_results(build_output)
            except Exception:
                self.logger.exception(
                    'Error sending completed build results.',
                    extra=logextra(task=self),
                )
        else:
            # Try to ship build error info.
            try:
                self.logger.debug(
                    'Sending build error info to bamaster...',
                    extra=logextra(task=self),
                )
                # If the task errored in this manner we consider it an
                # undefined error and don't expose the error message
                # publicly. Normal cloud-build error results should be
                # returned through build output, as we don't want
                # someone's random misconfigured asset-build showing up
                # as basn errors/etc.
                error_info = BuildErrorInfo(
                    error_type=BuildErrorInfo.ErrorType.UNDEFINED,
                    error_message=None,
                )
                await app.msg_out.send_async(
                    basntobamaster.CloudBuildResultsMessage(
                        master_server_build_id=self._master_server_build_id,
                        master_server_run_id=self._master_server_run_id,
                        output_data=None,
                        error_info=error_info,
                    )
                )
            except Exception:
                self.logger.exception(
                    'Error sending build error info.',
                    extra=logextra(task=self),
                )

    async def _send_completed_build_results(
        self, build_output: CloudBuildTaskResults
    ) -> None:
        from baserver.msg import basntobamaster

        from basn.env import app

        # If we have any cloud-files to push, do so.
        cfidlist = list(build_output.output_cloud_file_paths.keys())
        if cfidlist:
            # First send a single message with all our output
            # cloud-file-ids to find out which ones already exist in the
            # cloud.
            response = await app.msg_out.send_async(
                basntobamaster.CloudFilesExistQueryMessage(
                    cloud_file_ids=cfidlist
                )
            )
            cfexists: dict[str, bool] = dict(
                zip(cfidlist, response.cloud_files_exist, strict=True)
            )

            assert self._root is not None
            async with asyncio.TaskGroup() as task_group:
                for cfid, path in build_output.output_cloud_file_paths.items():
                    if cfexists[cfid]:
                        self.logger.debug(
                            'Skipping upload for already-in-cloud'
                            ' output file: %s.',
                            cfid,
                            extra=logextra(task=self),
                        )
                        continue  # Already in cloud; skip it.
                    _ = task_group.create_task(
                        self._upload_cloud_file(
                            cfid, os.path.join(self._root, path)
                        )
                    )

        # If we made it this far, all file outputs are in the cloud, and
        # we can deliver successful build results back to the master
        # server.
        await app.msg_out.send_async(
            basntobamaster.CloudBuildResultsMessage(
                master_server_build_id=self._master_server_build_id,
                master_server_run_id=self._master_server_run_id,
                output_data=build_output.output_data,
                error_info=build_output.error_info,
            )
        )

    async def _upload_cloud_file(self, cloud_file_id: str, path: str) -> None:
        await asyncio.get_running_loop().run_in_executor(
            None, lambda: self._do_upload_cloud_file(cloud_file_id, path)
        )

    def _do_upload_cloud_file(self, cloud_file_id: str, path: str) -> None:

        import urllib.request
        import urllib.parse
        from basn.env import config, app

        self.logger.debug(
            'Uploading build output file %s from %s...',
            cloud_file_id,
            path,
            extra=logextra(task=self),
        )

        flen = os.stat(path).st_size
        headers = {
            'Content-Type': 'application/octet-stream',
            'Content-Length': str(flen),
        }
        md5sum = _md5b64_for_path(path)

        # Passing a file obj to urllib seems like a nice way to stream
        # large file uploads, but the downside is that we can't pass
        # values as form parameters the way we'd normally do. We can,
        # however, pass them as url arguments. A bit non-standard for
        # POSTs but gets the job done.
        url = (
            f'{config.master_server_addr}/basn_cloudfile_put'
            f'?f={cloud_file_id}'
            f'&n={config.node_id}'
            f'&ct={app.comm_token}'
            f'&h={urllib.parse.quote(md5sum)}'
        )

        with open(path, 'rb') as file:
            request = urllib.request.Request(url, file, headers=headers)
            with urllib.request.urlopen(request) as response:
                if response.status != 200:
                    raise RuntimeError(
                        f'Upload failed for cloud-file {cloud_file_id}.'
                    )


def _md5b64_for_path(file_path: str) -> str:
    import hashlib
    import base64

    md5 = hashlib.md5()
    with open(file_path, 'rb') as f:
        # Read the file in chunks to handle large files efficiently.
        for chunk in iter(lambda: f.read(4096), b''):
            md5.update(chunk)
    return base64.b64encode(md5.digest()).decode()
