"""Task class for test task."""

from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, cast

from typing_extensions import override

from basn.taskwrapper import InlineTaskWrapper

if TYPE_CHECKING:
    from baserver.enums import TaskTestType
    from basntask.command import TaskCommand


class PingServerProtocol(asyncio.DatagramProtocol):
    """Our simple ping/response server."""

    def __init__(self) -> None:
        self.transport: asyncio.DatagramTransport | None = None
        super().__init__()

    @override
    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        # Note: the transport given to us is actually an internal
        # _SelectorDatagramTransport instance, but we can consider it
        # a DatagramTransport for typing purposes.
        assert isinstance(transport, asyncio.BaseTransport)
        self.transport = cast(asyncio.DatagramTransport, transport)

    @override
    def datagram_received(self, data: bytes, addr: tuple[str, int]) -> None:
        assert self.transport is not None

        # Stuck these values in while testing and forgot to change them,
        # so these are our official expected ping values now.
        if data == b'foo':
            self.transport.sendto(b'blarg', addr)


class PingTargetTaskWrapper(InlineTaskWrapper):
    """Task which sits and responds to pings."""

    def __init__(self, duration_seconds: float, load: float) -> None:
        super().__init__()
        self._duration_seconds = duration_seconds
        self._ptload = load

    @override
    def get_needs_udp_port(self) -> bool:
        return True

    @override
    def get_load(self) -> float:
        # bamaster informs us how heavy this task will be.
        return self._ptload

    @override
    def get_label(self) -> str:
        return f'ping-target {int(self._duration_seconds/60)}m'

    @override
    async def run_inline(self) -> None:
        # Get a reference to the event loop as we plan to use
        # low-level APIs.
        loop = asyncio.get_running_loop()

        # One protocol instance will be created to serve all client requests.
        assert self.udp_port is not None
        transport, _protocol = await loop.create_datagram_endpoint(
            PingServerProtocol, local_addr=('0.0.0.0', self.udp_port)
        )

        try:
            await asyncio.sleep(self._duration_seconds)
        finally:
            transport.close()
