"""Wrangle transport connection to a client."""

from __future__ import annotations

import time
import json
import random
import asyncio
from typing import TYPE_CHECKING, assert_never

from efro.message import UnregisteredMessageIDError
from efro.error import IntegrityError, CommunicationError
from efro.dataclassio import dataclass_to_json, dataclass_to_dict
import bacommon.cloud
from baserver.msg import basntobamaster
from baserver.msg import batocloud
from baserver.msg import cloudtoba
from baserver.cryptosimple import (
    xor_bytes,
    str_to_wire_key_xor,
    str_from_wire_key_xor,
)

from basn.logging import logextra, transportlog


if TYPE_CHECKING:
    import weakref
    from threading import Thread

    from efro.message import Message, Response, SysResponse
    from basn.taskwrapper.transportagent import (
        TransportAgentTaskWrapper,
    )
    from basn.taskwrapper.transportagent._transportagent import (
        AccountSessionChannel,
    )

SIDECAR_TOKEN_INVALID_KEY = '_v2tkinval'
SIDECAR_ACCOUNT_CHANNEL_ID_KEY = '_v2h'
SIDECAR_BUILD_NUMBER_KEY = '_babn'
SIDECAR_LOCATION_KEY = '_bal'

# Can kill these once all clients are 20888+
SIDECAR_TOKEN_KEY = '_v2tk'  # OLD
SIDECAR_VERIFIED_ACCOUNT_KEY_OLD = '_vafs'  # OLD


async def handle_msg_1(
    self: TransportAgentTaskWrapper, msg: batocloud.CToBMsg1
) -> batocloud.CToBResp1:
    """Message handling internals."""
    del self  # Unused.
    print(f'HANDLING {msg}!!!!!!!!!')
    return batocloud.CToBResp1(bval=True)


async def handle_sign_in_message(
    self: TransportAgentTaskWrapper, msg: bacommon.cloud.SignInMessage
) -> bacommon.cloud.SignInResponse:
    """Message handling internals."""
    from basn.env import app, config

    rsp = await app.msg_out.send_async(
        basntobamaster.SignInInternalMessage(
            login_type=msg.login_type,
            sign_in_token=msg.sign_in_token,
            source_task=f'{config.node_id}:{self.task_id}',
            description=msg.description,
            apptime=msg.apptime,
        )
    )

    if rsp.token is None:
        credentials = None
    else:
        # Encrypt token to be specific to the client device
        # in a way they can't easily decrypt (since they don't
        # need to).
        credentials = cloudtoba.AccountCredentials(
            token=encrypt_for_client_device(self, rsp.token),
        )

    # Now encrypt the whole set of values together but in a way
    # that the client *can* decrypt.
    return bacommon.cloud.SignInResponse(
        credentials=(
            None
            if credentials is None
            else encrypt_for_client_device(
                self, dataclass_to_json(credentials), client_readable=True
            )
        )
    )


async def handle_login_proxy_request_message(
    self: TransportAgentTaskWrapper,
    msg: bacommon.cloud.LoginProxyRequestMessage,
) -> bacommon.cloud.LoginProxyRequestResponse:
    """Message handling internals."""
    from basn.env import app

    del msg  # Unused.

    # Request is coming in for a login proxy; forward it
    # along to bamaster with some of our extra info about the
    # client included.
    rsp = await app.msg_out.send_async(
        basntobamaster.LoginProxyRequestInternalMessage(
            user_agent_string=self.user_agent_string,
            remote_addr=self.remote_addr,
            location=self.location,
        )
    )

    return bacommon.cloud.LoginProxyRequestResponse(
        url=rsp.url, proxyid=rsp.proxyid, proxykey=rsp.proxykey
    )


async def handle_login_proxy_state_query_message(
    self: TransportAgentTaskWrapper,
    msg: bacommon.cloud.LoginProxyStateQueryMessage,
) -> bacommon.cloud.LoginProxyStateQueryResponse:
    """Message handling internals."""
    from basn.env import app

    # State-query is coming in for a login proxy; forward it
    # along to bamaster with some of our extra info about the
    # client included.
    rsp = await app.msg_out.send_async(
        basntobamaster.LoginProxyStateQueryInternalMessage(
            proxyid=msg.proxyid,
            proxykey=msg.proxykey,
            buildnum=self.handshake.ba_build,
        )
    )

    # Translate internal state to equivalent public one.
    # Note: could do this in one line but wouldn't be type-safe.
    outstatetp = bacommon.cloud.LoginProxyStateQueryResponse.State
    if rsp.state is rsp.State.WAITING:
        outstate = outstatetp.WAITING
    elif rsp.state is rsp.State.SUCCESS:
        outstate = outstatetp.SUCCESS
    elif rsp.state is rsp.State.FAIL:
        outstate = outstatetp.FAIL
    else:
        assert_never(rsp.state)

    if rsp.token is None:
        credentials = None
    else:
        assert rsp.token is not None

        # Encrypt token to be specific to the client device
        # in a way they can't easily decrypt (since they don't
        # need to).
        credentials = cloudtoba.AccountCredentials(
            token=encrypt_for_client_device(self, rsp.token),
        )

    # Now encrypt the whole set of values together but in a way
    # that the client *can* decrypt.
    return bacommon.cloud.LoginProxyStateQueryResponse(
        state=outstate,
        credentials=(
            None
            if credentials is None
            else encrypt_for_client_device(
                self, dataclass_to_json(credentials), client_readable=True
            )
        ),
    )


async def handle_account_session_release_message(
    self: TransportAgentTaskWrapper, msg: batocloud.AccountSessionReleaseMessage
) -> None:
    """User is requesting to kill a session."""
    from basn.env import app, config

    try:
        # User is done with this session; calc its internal id and
        # forward along to bamaster.
        token = decrypt_from_client_device(self, msg.token)

        self.logger.info(
            'Task %s:%s: releasing account session %s.',
            config.node_id,
            self.task_id,
            token,
            extra=logextra(task=self),
        )
        await app.msg_out.send_async(
            basntobamaster.AccountSessionReleaseInternalMessage(token=token)
        )
    except IntegrityError:
        self.logger.warning(
            'Got IntegrityError for client-held token. Ignoring.',
            extra=logextra(task=self),
        )
    except CommunicationError:
        self.logger.debug(
            'Comm error sending account-session-release.',
            extra=logextra(task=self),
        )


def invalidate_token_on_client(
    self: TransportAgentTaskWrapper, token: str
) -> None:
    """Invalidate a token if we are using it (asynchronously)."""

    # ONLY launch a task to do this if we find the token
    # in any of our session-channels.
    for handle in self.session_channels.values():
        if handle.token == token:
            self.tasks.append(
                asyncio.create_task(
                    _invalidate_token_on_client(self, token),
                    name=(
                        f'transportagent {self.task_id}'
                        f' invalidate client token'
                    ),
                )
            )
            return


async def _invalidate_token_on_client(
    self: TransportAgentTaskWrapper, token: str
) -> None:
    """Invalidate a token."""
    try:
        await self.msg_out.send_async(
            cloudtoba.TokenInvalidatedMessage(
                token=encrypt_for_client_device(self, token)
            )
        )
    except CommunicationError:
        # No biggie; we only want to invalidate if they're connected.
        # Apparently they're not actually.
        pass
    except Exception:
        # Something unexpected; make noise.
        self.logger.warning(
            'Error invalidating token on client.',
            exc_info=True,
            extra=logextra(task=self),
        )


async def bg_register_channel_with_bamaster(
    self: TransportAgentTaskWrapper, channel_id: int
) -> None:
    """Register with master in a chill bg way."""
    from basn.env import config

    # Retry a few times to account for fluke-ish comm errors talking
    # to bamaster.
    maxtries = 5
    for trynum in range(maxtries):
        try:
            await get_verified_account_session_channel(
                self, channel_id=channel_id
            )
            break  # Success!
        except CommunicationError:
            if trynum == maxtries - 1:
                self.logger.warning(
                    'Unable to register account-session-channel'
                    ' due to repeated comm errors.',
                    extra=logextra(task=self),
                )
        except Exception:
            # Something unexpected; make noise and abort.
            session_channel = self.session_channels.get(channel_id)
            self.logger.warning(
                'Unexpected error in bg attempt %d of %d'
                ' registering channel %s:%s:%s of account session %s.',
                trynum + 1,
                maxtries,
                config.node_id,
                self.task_id,
                channel_id,
                (
                    '(not found)'
                    if session_channel is None
                    else session_channel.token
                ),
                exc_info=True,
                extra=logextra(task=self),
            )
            if trynum == maxtries - 1:
                self.logger.warning(
                    'Unable to register account-session-channel'
                    ' due to repeated unexpected errors.',
                    extra=logextra(task=self),
                )
        await asyncio.sleep(5.0)


async def deregister_channels_with_bamaster(
    self: TransportAgentTaskWrapper, channel_ids: list[int]
) -> None:
    """Deregister a channel with bamaster."""
    from basn.env import app, config

    # Do these one at a time to avoid contention. Can get fancier later
    # if speed is ever an issue.
    for channel_id in channel_ids:
        channel = self.session_channels[channel_id]

        # Retry a few times to account for fluke-ish comm errors talking
        # to bamaster.
        maxtries = 5
        for trynum in range(maxtries):
            # Sanity checks; we should never be called on a live channel.
            assert not channel.ready
            assert channel.closed

            # If someone else has confirmed this channel as closed, we're done.
            if channel.close_confirmed:
                return

            # I was initially getting errors on bamaster due to close-channel
            # messages coming in before the channel had been established.
            # So now let's make sure to wait until a few seconds after the
            # last register message was sent out to try and avoid this.
            # (or a few seconds from *now* if no registers have gone out
            # yet).
            now = time.monotonic()
            if channel.last_verify_attempt_time is None:
                safe_time = now + 5.0
            else:
                safe_time = channel.last_verify_attempt_time + 5.0
            until_safe = safe_time - now
            if until_safe > 0.0:
                await asyncio.sleep(until_safe)
            try:
                await app.msg_out.send_async(
                    basntobamaster.AccountSessionDeleteChannelsMessage(
                        task_id=self.task_id,
                        channel_ids_per_session={channel.token: [channel_id]},
                    )
                )
                channel.close_confirmed = True
                break  # Success!
            except CommunicationError:
                if trynum == maxtries - 1:
                    self.logger.warning(
                        'Unable to deregister channel %s:%s:%s of session %s'
                        ' due to repeated comm errors.',
                        config.node_id,
                        self.task_id,
                        channel_id,
                        channel.token,
                        extra=logextra(task=self),
                    )
            except Exception:
                # Something unexpected; make more noise.
                self.logger.exception(
                    'Error deregistering channel %s:%s:%s'
                    ' of session %s with bamaster (attempt %d of %d).',
                    config.node_id,
                    self.task_id,
                    channel_id,
                    channel.token,
                    trynum + 1,
                    maxtries,
                    extra=logextra(task=self),
                )

            await asyncio.sleep(random.uniform(3.0, 6.0))


def encrypt_for_client_device(
    self: TransportAgentTaskWrapper, data: str, client_readable: bool = False
) -> str:
    """Encrypt some data specifically for this peer.

    Useful when having the client store some data without being
    able to easily read its contents.
    """
    return str_to_wire_key_xor(
        data,
        (
            xor_key_for_client_device_client_readable(self)
            if client_readable
            else xor_key_for_client_device(self)
        ),
        compress=False,
        checksum=True,
    )


def decrypt_from_client_device(
    self: TransportAgentTaskWrapper, data: str, client_readable: bool = False
) -> str:
    """Decrypt data from the peer.

    Be aware that this will raise efro.IntegrityError
    if the data has been tampered with or if the client's
    internal device id has somehow changed.
    """
    return str_from_wire_key_xor(
        data,
        (
            xor_key_for_client_device_client_readable(self)
            if client_readable
            else xor_key_for_client_device(self)
        ),
        compress=False,
        checksum=True,
    )


def xor_key_for_client_device(self: TransportAgentTaskWrapper) -> bytes:
    """come up with an xor key using the peer's private device id.

    Useful for passing encrypted data to the client that they can't
    easily read.
    """
    return xor_bytes(
        self.handshake.private_device_uuid.encode(),
        (
            b'\xe2\x19\x83(\x1f\x07\x89\xb0\x07j'
            b'\x13u\xa4\x08UG\x06b\xa6\xb7'
            b'n\xa0PXk\xb2\xd7\x85\xdaz/)y\ty\xd0\x92P\xc1\x01\x04*?bp\xae'
            b'\x04j\xf9\x1f(\xc9\xd8y\xfc\xbe\xa4\x83\x8f\xf6\xc1\xc8\xd2'
            b'\x8a\x9d.U<Q\xf7$#tH\x82\n\xcb\xd2\xdbF`\xa2\xae\xb1\x0f'
        ),
    )


def xor_key_for_client_device_client_readable(
    self: TransportAgentTaskWrapper,
) -> bytes:
    """Alternate xor key for the peer.

    This one is a bit less secure because the client uses it too
    in internal code (see account_client_v2.py).
    """
    return xor_bytes(
        self.handshake.private_device_uuid.encode(),
        (
            b'4\xda1\xf4r\x96\xcc\x9f8\x14\xd6\xa3\xa1\xe7q\xb0H\xca'
            b'\xca\xe0\xd1u\xfe\xbd\xb7\xce\xece\x88g\x1b#\xf5\xbc\xf0'
            b'\xbd\xdes\xeb\xb4\x9b\x8c\x90>G\xfd\xccQPk`\x8d:j"\x81`'
            b'\x816\xb8\x99\x80i\x85\xec\xfc\xee:\xa9'
        ),
    )


def note_if_alive(
    wself: weakref.ref[TransportAgentTaskWrapper], msg: str
) -> None:
    """Emit a note if the provided wrapper still exists."""
    sself = wself()
    if sself is not None:
        sself.add_note(msg)
    else:
        transportlog.warning(
            'Got debug-print for endpoint of dead agent: %s',
            msg,
            extra=logextra(),
        )


async def get_verified_account_session_channel(
    self: TransportAgentTaskWrapper, channel_id: int | None
) -> AccountSessionChannel | None:
    """Given a channel-id, registers and/or returns valid channel."""
    # pylint: disable=too-many-branches
    # pylint: disable=too-many-nested-blocks
    # pylint: disable=too-many-return-statements
    # pylint: disable=cyclic-import
    from basn.env import app, config

    if channel_id is None:
        return None

    session_channel = self.session_channels.get(channel_id)

    # If there's not even a registered session with this id, insta-fail.
    if session_channel is None:
        return None

    # If this is a valid channel but it has been closed already, insta-fail.
    if session_channel.closed:
        return None

    try:
        # Ok; we got a valid bamaster account-session token.

        # Now, if we haven't yet, register it with bamaster.
        if not session_channel.ready:
            # If a verification attempt was kicked off very recently, let's
            # spin and wait a moment for it to complete before resorting to
            # kicking off a new one.
            if session_channel.last_verify_attempt_time is not None:
                wait_until_time = session_channel.last_verify_attempt_time + 2.0
                while time.monotonic() < wait_until_time:
                    await asyncio.sleep(0.01)

                    if channel_id in self.session_channels:
                        # If we found a ready one, return it.
                        if self.session_channels[channel_id].ready:
                            return self.session_channels[channel_id]

                        # Found a dead one; abort.
                        if self.session_channels[channel_id].closed:
                            return None

            # Ok; no luck so far. Kick off a request ourself.
            session_channel.last_verify_attempt_time = time.monotonic()
            response = await app.msg_out.send_async(
                basntobamaster.AccountSessionRegisterChannelMessage(
                    token=session_channel.token,
                    channel_id=channel_id,
                    user_agent_string=self.user_agent_string,
                    remote_addr=self.remote_addr,
                    location=self.location,
                    task_id=self.task_id,
                    device_name=self.handshake.device_name,
                    build_number=self.handshake.ba_build,
                    app_instance_uuid=self.handshake.app_instance_uuid,
                )
            )

            # If we get a clear rejection from bamaster, it implies that
            # the channel is confirmed closed *and* the session itself
            # is dead. (Currently bamaster kills sessions on failed
            # channel registrations to avoid funny business).
            if not response.verified:
                self.logger.debug(
                    'Got negative verification response'
                    ' for session token %s.',
                    session_channel.token,
                )

                session_channel.ready = False
                session_channel.closed = True
                session_channel.close_confirmed = True
                session_channel.session_invalidated = True

                # Let the client know about the session invalidation
                # so it can sign out or whatnot.
                invalidate_token_on_client(self, session_channel.token)

                return None

            # Success!!!
            assert response.channel_id is not None
            assert response.info is not None

            # It might have been closed under us while we were waiting..
            # (Note: accessing via full path here so mypy doesn't assume
            # that closed has not changed)
            if self.session_channels[channel_id].closed:
                return None

            self.logger.debug(
                'Verified session-channel %s of session %s.',
                channel_id,
                session_channel.token,
            )
            session_channel.ready = True
            session_channel.info = response.info

        return session_channel

    except CommunicationError:
        # Expected sometimes; no biggie.
        raise
    except Exception:
        # Make noise for anything unexpected.
        # (Limit to newer builds though; old had some chatty problems)
        if self.handshake.ba_build >= 20983:
            self.logger.exception(
                'Error in get_verified_account_session_channel() for %s:%s:%s'
                ' of session %s (buildnum %s).',
                config.node_id,
                self.task_id,
                channel_id,
                session_channel.token,
                self.handshake.ba_build,
                extra=logextra(task=self),
            )
        raise


def get_message_account_session_channel_id(msg: Message) -> int | None:
    """Extract the account-session-channel-id attached to client messages."""
    attached = getattr(msg, SIDECAR_ACCOUNT_CHANNEL_ID_KEY, None)
    assert isinstance(attached, int | None)
    return attached


async def handle_raw_message(
    self: TransportAgentTaskWrapper, message_data: bytes
) -> bytes:
    """Called for messages coming in from the client over our RPCEndpoint."""
    from basn.env import app

    # Our efro.message layer deals in strings.
    message = message_data.decode()
    response: str

    # This should never be coming in after we shut down.
    assert not self.is_ended()

    # We handle a subset of these messages ourself.
    try:
        response = await self.msg_in.handle_raw_message(
            message, raise_unregistered=True
        )
    except UnregisteredMessageIDError:
        # Forward anything we don't handle on to bamaster.
        try:
            # We pretty much pass the message to bamaster verbatim,
            # except we include a bit of extra client info since bamaster
            # doesn't have access to the client info that we do.
            msg_dict = json.loads(message)

            # Include the client build number we are connected to.
            msg_dict[SIDECAR_BUILD_NUMBER_KEY] = self.handshake.ba_build

            # Include location if we have it.
            if self.location is not None:
                msg_dict[SIDECAR_LOCATION_KEY] = dataclass_to_dict(
                    self.location
                )
            assert self.handshake.ba_build >= 20888
            # Clients send us a simple channel-id int but we
            # need to send the server a full session-id string.
            # And they trust us to only do that in the case where
            # we've verified things already.
            client_account_channel_id = msg_dict.get(
                SIDECAR_ACCOUNT_CHANNEL_ID_KEY
            )
            if isinstance(client_account_channel_id, int):
                session_channel = await get_verified_account_session_channel(
                    self, client_account_channel_id
                )
                if session_channel is not None:
                    msg_dict[SIDECAR_TOKEN_KEY] = session_channel.token
                else:
                    # Replace invalid tokens with a special string so
                    # bamaster knows that an invalid client token was
                    # attached (the user *thought* they were logged
                    # in). bamaster may still choose to process the
                    # message if credentials don't matter for it.
                    msg_dict[SIDECAR_TOKEN_KEY] = '__invalid__'

                # NOTE: Also including channel id here though
                # not using it on bamaster at this time.
                msg_dict[SIDECAR_ACCOUNT_CHANNEL_ID_KEY] = (
                    client_account_channel_id
                )

            message = json.dumps(msg_dict, separators=(',', ':'))

            msg = basntobamaster.ForwardedClientMessage(
                message=message, client_task_id=self.task_id
            )

            # Staple the internal type int to it. Our send code uses this
            # to send to different urls per type for better stats.
            # (by definition we don't know the internal type names here so
            # this is the best we can do)
            internal_type = msg_dict.get('t')
            if isinstance(internal_type, int):
                setattr(msg, '_fwd_int_tp', internal_type)
            else:
                self.logger.warning(
                    'Type int not found on forwarded msg.',
                    extra=logextra(task=self),
                )
            rmsg = await app.msg_out.send_async(msg)

            # Note: the response could have a SIDECAR_TOKEN_INVALID_KEY
            # in it, but we leave that fully up to bamaster and don't
            # touch it ourself (its simply a bool so we don't have to
            # translate it). Technically we could inspect this to see
            # if we should invalidate our own token, but this response
            # should cause the client to log out immediately so should
            # be a moot point.
            response = rmsg.response

        except Exception as exc:
            # Respond cleanly to the client even if something went wrong.
            response = self.msg_in.encode_error_response(exc)

    return response.encode()


def msg_in_decode_filter(
    self: TransportAgentTaskWrapper, indata: dict, message: Message
) -> None:
    """Filter messages coming in from the client."""
    from basn.env import config

    assert self.handshake.ba_build >= 20888

    # If it includes an account-session-channel-id, staple it to
    # the output message.
    channel_id = indata.get(SIDECAR_ACCOUNT_CHANNEL_ID_KEY)
    if isinstance(channel_id, int):
        setattr(message, SIDECAR_ACCOUNT_CHANNEL_ID_KEY, channel_id)

        if channel_id not in self.session_channels:
            # Only report this on newer builds (hopefully has been fixed).
            if self.handshake.ba_build >= 20986:
                self.logger.warning(
                    'msg_in_decode_filter: %s from client'
                    ' (build %s) uses nonexistent channel %s:%s:%s.',
                    type(message).__name__,
                    self.handshake.ba_build,
                    config.node_id,
                    self.task_id,
                    channel_id,
                    extra=logextra(task=self),
                )
        # Shouldn't be any need to kick off a verify here, right?
        # Anyone needing verification will kick one off in a moment
        # anyway...
        # else:
        #     # If we find a still-unverified handle, attempt to
        #     # verify it before proceeding.
        #     await get_verified_account_session_channel(self, channel_id)


def msg_in_encode_filter(
    self: TransportAgentTaskWrapper,
    message: Message | None,
    response: Response | SysResponse,
    outdata: dict,
) -> None:
    """Filter raw response data going back out to the client."""
    del response  # Unused

    assert self.handshake.ba_build >= 20888

    channel_id = getattr(message, SIDECAR_ACCOUNT_CHANNEL_ID_KEY, None)
    assert isinstance(channel_id, int | None)

    # If this message references a session that has since been closed,
    # let the user know that.
    if channel_id is not None and (
        channel_id in self.session_channels
        and self.session_channels[channel_id].session_invalidated
    ):
        outdata[SIDECAR_TOKEN_INVALID_KEY] = True

    # Note: should we do something if the channel we're using has
    # since disappeared? I'm thinking maybe not. A disappearing
    # channel would be due to a request from the client so they
    # could discard the response when it gets back to them if
    # they desire.


def get_message_account_token_old(msg: Message) -> str | None:
    """Used by builds < 20888; can retire eventually."""

    # Extract the client token attached to client messages.
    attached = getattr(msg, SIDECAR_TOKEN_KEY, None)
    assert isinstance(attached, (str, type(None)))
    return attached


def handle_account_channel_create_message(
    self: TransportAgentTaskWrapper,
    msg: batocloud.AccountChannelCreateMessage,
) -> None:
    """Register an account session."""
    # pylint: disable=cyclic-import
    from basn.env import config
    from basn.taskwrapper.transportagent._transportagent import (
        AccountSessionChannel,
    )

    # For now just store this session-channel-id/token pair. We will
    # then require a registration of this pair to go through to the
    # master-server before we can send messages on behalf of it.

    # Decrypt their client token into the bamaster one.
    bamaster_token = decrypt_from_client_device(self, msg.token)

    channel_addr = f'{config.node_id}:{self.task_id}:{msg.channel_id}'
    self.logger.info(
        'Creating channel %s in account-session %s.',
        channel_addr,
        bamaster_token,
        extra=logextra(task=self, channel=channel_addr, session=bamaster_token),
    )

    assert msg.channel_id not in self.session_channels
    self.session_channels[msg.channel_id] = AccountSessionChannel(
        token=bamaster_token
    )

    prune_tasks(self)
    self.tasks.append(
        asyncio.create_task(
            bg_register_channel_with_bamaster(self, channel_id=msg.channel_id),
            name=f'transportagent {self.task_id} create account channel',
        )
    )


def handle_account_channel_destroy_message(
    self: TransportAgentTaskWrapper,
    msg: batocloud.AccountChannelDestroyMessage,
) -> None:
    """Client is closing an account channel."""
    # pylint: disable=cyclic-import
    from basn.env import config

    channel_addr = f'{config.node_id}:{self.task_id}:{msg.channel_id}'
    if msg.channel_id not in self.session_channels:
        self.logger.warning(
            'Channel not found when handling destroy-channel message:'
            ' %s (build %d).',
            channel_addr,
            self.handshake.ba_build,
            extra=logextra(channel=channel_addr),
        )
        return

    channel = self.session_channels[msg.channel_id]

    self.logger.info(
        'Destroying channel %s in account-session %s.',
        channel_addr,
        channel.token,
        extra=logextra(
            channel=channel_addr,
            session=channel.token,
        ),
    )

    # Immediately mark it as unusable even though we still need to
    # inform bamaster.
    channel.ready = False
    channel.closed = True

    # Kick off a task to inform bamaster of this closing.
    prune_tasks(self)
    self.tasks.append(
        asyncio.create_task(
            deregister_channels_with_bamaster(self, [msg.channel_id]),
            name=f'transportagent {self.task_id} channel {msg.channel_id}'
            f' deregister',
        )
    )


async def handle_credentials_check_message(
    self: TransportAgentTaskWrapper, msg: batocloud.CredentialsCheckMessage
) -> batocloud.CredentialsCheckResponse:
    """Check creds."""

    assert self.handshake.ba_build >= 20888

    channel_id = get_message_account_session_channel_id(msg)
    session_channel = await get_verified_account_session_channel(
        self, channel_id
    )

    return batocloud.CredentialsCheckResponse(
        verified=(session_channel is not None),
        legacy_tag='',
        info=None if session_channel is None else session_channel.info,
    )


async def handle_login_proxy_complete_message(
    self: TransportAgentTaskWrapper,
    msg: bacommon.cloud.LoginProxyCompleteMessage,
) -> None:
    """Message handling internals."""
    from basn.env import app

    del self  # Unused.

    # User is done with this proxy; calc its internal id and
    # forward along to bamaster.
    await app.msg_out.send_async(
        basntobamaster.LoginProxyCompleteInternalMessage(proxyid=msg.proxyid)
    )


# async def deregister_all_channels_with_bamaster(
#     self: TransportAgentTaskWrapper,
# ) -> None:
#     """Deregister all sessions for this agent."""
#     # pylint: disable=cyclic-import
#     from basn.taskwrapper.transportagent._transportagent import (
#         DEBUG_PRINT,
#         LABEL,
#     )
#     from basn.env import app

#     # Attempt this a few times; generally a communication
#     # error with the master-server is a fluke and retrying will
#     # succeed.
#     trynum = 1
#     maxtries = 5
#     while True:
#         try:
#             if self.session_channels:
#                 # Even ask for unverified ones just in case they
#                 # have reached the master-server; better to request
#                 # a delete for a nonexistent handle than to leave
#                 # an orphaned one.
#                 ids_per_session = self.get_channel_ids_per_session(
#                     ready_only=False
#                 )
#                 await app.msg_out.send_async(
#                     basntobamaster.AccountSessionDeleteChannelsMessage(
#                         task_id=self.task_id,
#                         channel_ids_per_session=ids_per_session,
#                     )
#                 )
#                 if DEBUG_PRINT:
#                     print(
#                         f'{LABEL}: Deleted session-channels before exit:'
#                         f' {ids_per_session}'
#                     )
#             # Success!
#             break

#         except CommunicationError:
#             # We expect these occasionally; no biggie. Only complain
#             # if we keep failing.
#             if trynum >= maxtries:
#                 logging.warning(
#                     'Comm error deregistering account sessions'
#                     ' (maxtries %d reached).',
#                     maxtries,
#                     exc_info=True,
#                 )
#                 break

#         except Exception:
#             # For anything else going wrong, make a bigger fuss
#             # (though we keep trying in this case too).
#             if trynum >= maxtries:
#                 logging.exception(
#                     'Unexpected error deregistering account sessions'
#                     ' (maxtries %d reached).',
#                     maxtries,
#                 )
#                 break

#         # Take a breather and just keep swimming.
#         await asyncio.sleep(5.0)
#         trynum += 1


def prune_tasks(self: TransportAgentTaskWrapper) -> None:
    """Remove dead tasks from our list."""
    # Quick out: nothing changed.
    if not any(t.done() for t in self.tasks):
        return
    self.tasks = [t for t in self.tasks if not t.done()]
