"""Misc basn utility functionality."""

from __future__ import annotations

from typing import TYPE_CHECKING, assert_type

from basn.logging import logextra, balog

if TYPE_CHECKING:
    import io
    from pathlib import Path

    from efro.log import LogLevel

    from basn.taskwrapper import BaseTaskWrapper


def force_kill_user_processes(
    user_id: str, for_task: BaseTaskWrapper, warn_if_running: bool
) -> None:
    """Force kill all processes with a given user id."""
    import subprocess

    try:
        result = subprocess.run(
            ['sudo', 'pkill', '--signal', 'SIGKILL', '-U', user_id],
            capture_output=True,
            check=False,
        )

        # This task should fail silently with no output when there are
        # no tasks left for this user. If that is not what we expect
        # to happen, holler loudly.
        # TODO: currently see an occasional error message here due to pkill
        # exiting with status 0; should try to determine if that means some
        # task was still running.
        if warn_if_running and (
            result.stdout != b''
            or result.stderr != b''
            or result.returncode != 1
        ):
            balog.warning(
                'pkill for user %s got unexpected results:'
                ' retcode=%d'
                " stdout='%s' stderr='%s'",
                user_id,
                result.returncode,
                result.stdout,
                result.stderr,
                extra=logextra(task=for_task),
            )
    except Exception:
        balog.exception(
            'Error in force_kill_user_processes.',
            extra=logextra(task=for_task),
        )


def read_lines_filtered(
    infile: io.TextIOWrapper,
    max_size: int | None,
    min_log_level: LogLevel,
    contains: str | None,
) -> tuple[str, int]:
    """Read lines from a file, clipping to last ones if too long.

    max_size is interpreted as utf8 byte length.
    Returns the combined final lines and how many leading lines were
    truncated.
    """
    from collections import deque
    from efro.dataclassio import dataclass_from_json
    from efro.log import LogLevel, LogEntry

    size = 0
    lines = deque[tuple[int, str]]()
    trunclines = 0
    for line in infile:
        # If we're filtering in any way, do so.
        if min_log_level.value > 0 or contains:
            entry = dataclass_from_json(LogEntry, line)
            # Since we're looking at raw values, make sure we're comparing
            # apples to apples.
            assert_type(entry.level, LogLevel)
            assert_type(min_log_level, LogLevel)
            if entry.level.value < min_log_level.value:
                continue
            # Let's do case-insensitive search for now.
            if (
                contains is not None
                and contains.lower() not in entry.message.lower()
            ):
                continue

        # Add line to the end then remove from the beginning
        # until we're within our max size.
        lbytes = len(line.encode())
        lines.append((lbytes, line))
        size += lbytes
        if max_size is not None:
            while size > max_size:
                size -= lines.popleft()[0]
                trunclines += 1
    return ''.join(l[1] for l in lines), trunclines


def get_debug_info() -> str:
    """Get general debug info about the app."""
    # pylint: disable=too-many-locals
    import psutil
    import html
    import gc

    out = ''
    proc = psutil.Process()
    connections = proc.connections(kind='inet')
    openfiles = proc.open_files()
    if bool(False):
        balog.info('CONNS: %s', connections, extra=logextra())
        balog.info('OPENFILES: %s', openfiles, extra=logextra())

    # Network connections.
    out += 'CONNECTIONS:<br>'
    for i, connection in enumerate(sorted(connections, key=lambda c: c.fd)):
        out += (
            f'{i}:'
            f' fd={connection.fd}'
            f' type={connection.type.name}'
            f' ip={connection.laddr.ip}'
            f' port={connection.laddr.port}'
            f'<br>'
        )

    # Open files.
    out += '<br>FILES:<br>'
    for i, openfile in enumerate(sorted(openfiles, key=lambda f: f.fd)):
        out += f'{i}:' f' fd={openfile.fd}' f' path={openfile.path}' f'<br>'

    # Allocated objects.
    objtypes: dict[str, int] = {}
    gc.collect()  # Recommended before get_objects().
    allobjs = gc.get_objects()
    allobjc = len(allobjs)
    for obj in allobjs:
        tpname = str(type(obj))
        objtypes[tpname] = objtypes.get(tpname, 0) + 1

    # Presumably allobjs contains stack-frame/dict type stuff
    # from this function call which in turn contain refs to allobjs.
    # Let's try to prevent these huge lists from accumulating until
    # the cyclical collector (hopefully) gets to them.
    allobjs.clear()
    del allobjs

    out += f'<br>OBJS ({allobjc} total):<br>'
    for tpname, tpval in sorted(
        objtypes.items(), key=lambda x: x[1], reverse=True
    )[:50]:
        percent = tpval / allobjc * 100.0
        out += html.escape(f'{tpname}: {tpval} ({percent:.2f}%)') + '<br>'

    return out
