"""For testing various basn subtask mechanisms."""

from __future__ import annotations

import os
import time
import logging
from typing import TYPE_CHECKING

from dataclasses import dataclass

from typing_extensions import override
from efro.error import CleanError
from efro.dataclassio import ioprepped
from baserver.cloudbuild import (
    RecipeTypeID,
    RecipeComponent,
    Recipe,
    CloudBuildContext,
    TestRecipe,
    BuildErrorInfo,
    PermanentBuildError,
    TemporaryBuildError,
)

from basntask.command import TaskCommand, TaskCommandTypeID
from basntask.result import TaskResult, TaskOutput, TaskOutputTypeID

if TYPE_CHECKING:
    from baserver.cloudbuild import TestInput, TestOutput

    pass


@ioprepped
@dataclass
class CloudBuildCommand(TaskCommand):
    """Run a cloud build."""

    recipe_type_id: RecipeTypeID
    input_data: RecipeComponent
    input_cloud_file_paths: dict[str, str]

    # For misc hacky use (currently used by test builds).
    extra: bytes | None = None

    @override
    @classmethod
    def get_type_id(cls) -> TaskCommandTypeID:
        return TaskCommandTypeID.CLOUD_BUILD

    @override
    def run(self) -> TaskResult:

        logging.debug('Staging %s build...', self.recipe_type_id.value)
        starttime = time.monotonic()

        # Where we stick our output files.
        outdirname = 'outfiles'
        os.mkdir(outdirname)

        recipe_type = Recipe.get_type(self.recipe_type_id)
        context = CloudBuildContext(recipe_type, output_dir=outdirname)

        # Stage all input files on disk.
        stagecount = 0
        for cloudfileid in self.input_data.get_cloud_file_ids():
            assert cloudfileid in self.input_cloud_file_paths
            assert os.path.isfile(self.input_cloud_file_paths[cloudfileid])
            context.set_path_for_input_file(
                self.input_cloud_file_paths[cloudfileid], cloudfileid
            )
            stagecount += 1

        # Make extra data available if passed.
        if self.extra is not None:
            with open('cloud_build_extra_data', 'wb') as outfile:
                outfile.write(self.extra)

        logging.debug(
            'Staged %s input files in %.3fs.',
            stagecount,
            time.monotonic() - starttime,
        )

        # Run the build.
        execstarttime = time.monotonic()
        logging.debug('Execing %s build...', self.recipe_type_id.value)

        output_data: RecipeComponent | None
        error_info: BuildErrorInfo | None
        output_files_as_paths: dict[str, str]
        try:
            output_data = context.exec_build(self.input_data)
            logging.debug(
                'Exec completed in %.3fs.', time.monotonic() - execstarttime
            )
            output_files_as_paths = context.get_output_files_as_paths()
            error_info = None
        except CleanError as exc:
            # Any clean-error (includes Temporary/PermanentBuildError)
            # results in the message being shown to the builder.
            output_data = None
            output_files_as_paths = {}
            error_info = BuildErrorInfo(
                error_type=(
                    BuildErrorInfo.ErrorType.PERMANENT
                    if isinstance(exc, PermanentBuildError)
                    else (
                        BuildErrorInfo.ErrorType.TEMPORARY
                        if isinstance(exc, TemporaryBuildError)
                        else BuildErrorInfo.ErrorType.UNDEFINED
                    )
                ),
                error_message=str(exc),
            )
            logging.debug(
                'Exec failed in %.3fs with error-info %s.',
                time.monotonic() - execstarttime,
                error_info,
            )
        except Exception:
            # For any other exception, store a generic error message;
            # we'll have to look at logs to see details.
            output_data = None
            output_files_as_paths = {}
            error_info = BuildErrorInfo(
                error_type=BuildErrorInfo.ErrorType.UNDEFINED,
                error_message=None,
            )
            # Now spit the error to logs so we can see those details.
            logging.warning(
                'CloudBuild gave an undefined error.'
                ' Ideally builds should give only TemporaryBuildErrors'
                ' or PermanentBuildErrors.',
                exc_info=True,
            )
        return TaskResult(
            # Any errors that we catch are returned as part of our clean
            # output.
            output=CloudBuildTaskResults(
                output_data, output_files_as_paths, error_info
            ),
            # The task itself succeeded; we don't want basn to show errors.
            error=None,
        )


@ioprepped
@dataclass
class CloudBuildTaskResults(TaskOutput):
    """Results for a completed cloudbuild task (both successful and not)."""

    # Either output_data or error_info will be set.
    output_data: RecipeComponent | None
    output_cloud_file_paths: dict[str, str]

    error_info: BuildErrorInfo | None

    @override
    @classmethod
    def get_type_id(cls) -> TaskOutputTypeID:
        return TaskOutputTypeID.CLOUD_BUILD


class TestRecipeBasn(TestRecipe):
    """Demo of overriding build for our environment."""

    @override
    def build(self, in_data: TestInput) -> TestOutput:

        # We expect to find our 'mode' written to a file here
        # (can't take this through input data since that would change
        # our build signature which we don't want).
        with open('cloud_build_extra_data', 'rb') as infile:
            mode = infile.read().decode()

        # Test results of various errors we can throw during a build.
        if mode == 'success':
            pass  # Do normal.
        elif mode == 'err_perm':
            raise PermanentBuildError('Test Permanent Build Error')
        elif mode == 'err_temp':
            raise TemporaryBuildError('Test Temporary Build Error')
        elif mode == 'err_undef':
            raise RuntimeError('Test Undefined Build Error')
        elif mode == 'timeout':
            print('Sleeping for 10s to test build timeout.')
            time.sleep(10.0)
        else:
            raise RuntimeError(f'Unexpected mode for test build: "{mode}".')

        print('Hello from TestRecipeBasn. Calling original now.')
        return super().build(in_data=in_data)
