"""Functionality for task commands."""

from __future__ import annotations

import sys
from enum import Enum
from typing import TYPE_CHECKING, assert_never

from typing_extensions import override
from efro.dataclassio import IOMultiType


if TYPE_CHECKING:
    from types import FrameType

    from basntask.result import TaskResult
    from basntask.taskenv import TaskEnv


class TaskCommandTypeID(Enum):
    """Registry of task types."""

    TEST = 'test'
    HOSTED_GAME = 'hosted_game'
    CLOUD_BUILD = 'cloud_build'


_g_task_env: TaskEnv | None = None


class TaskCommand(IOMultiType[TaskCommandTypeID]):
    """Payload for a standalone basn task.

    Intended to be run in a standalone process.
    """

    @override
    @classmethod
    def get_type_id(cls) -> TaskCommandTypeID:
        # Require child classes to supply this themselves. If we
        # did a full type registry/lookup here it would require us
        # to import everything and would prevent lazy loading.
        raise NotImplementedError()

    @override
    @classmethod
    def get_type(cls, type_id: TaskCommandTypeID) -> type[TaskCommand]:
        # pylint: disable=cyclic-import
        out: type[TaskCommand]
        t = TaskCommandTypeID

        if type_id is t.TEST:
            from basntask.testtask import TestTaskCommand

            out = TestTaskCommand
        elif type_id is t.HOSTED_GAME:
            from basntask.hostedgame import HostedGameCommand

            out = HostedGameCommand
        elif type_id is t.CLOUD_BUILD:
            from basntask.cloudbuild import CloudBuildCommand

            out = CloudBuildCommand
        else:
            # Important to make sure we provide all types.
            assert_never(type_id)
        return out

    @property
    def env(self) -> TaskEnv:
        """Access the task execution environment."""
        assert _g_task_env is not None
        return _g_task_env

    def run(self) -> TaskResult:
        """Run the command. Will be called in its own process."""
        raise NotImplementedError()

    def terminate(self) -> None:
        """Called if SIGTERM hits the task process."""

        # By default, just insta-die.
        sys.exit(-1)

    @staticmethod
    def set_env(env: TaskEnv) -> None:
        """Set the task execution environment."""
        global _g_task_env  # pylint: disable=global-statement
        assert _g_task_env is None
        _g_task_env = env

    @staticmethod
    def handle_term_signal(sigtype: int, frame: FrameType | None) -> None:
        """Handler for a TERM interrupt signal."""
        del sigtype, frame  # Unused
        if _g_task_env is not None:
            print(
                'basntask got SIGTERM; asking running task to terminate...',
                file=sys.stderr,
                flush=True,
            )
            _g_task_env.command.terminate()
        else:
            print(
                'basntask got SIGTERM before task exists; insta-dying...',
                file=sys.stderr,
                flush=True,
            )
            sys.exit(-1)
