"""Functionality for task processes."""

from __future__ import annotations

from typing import TYPE_CHECKING, assert_never

import os
import sys
import random
import subprocess
from dataclasses import dataclass

from typing_extensions import override
from efro.error import CleanError
from efro.dataclassio import ioprepped, dataclass_to_json
from bacommon.servermanager import ServerConfig
from baserver.hostedgame import HostedGameType, AdHocGameType, HostedGameDef

from basntask.command import TaskCommand, TaskCommandTypeID
from basntask.result import TaskResult

if TYPE_CHECKING:
    pass
    # from baserver.hostedgame import HostedGameDef


@ioprepped
@dataclass
class HostedGameCommand(TaskCommand):
    """Host a game."""

    gamedef: HostedGameDef
    port: int

    @override
    @classmethod
    def get_type_id(cls) -> TaskCommandTypeID:
        return TaskCommandTypeID.HOSTED_GAME

    # Slight hack: Add a non-dataclass property to our dataclass.
    @property
    def _process(self) -> subprocess.Popen[bytes]:
        return getattr(self, '_process_val')  # type: ignore

    @_process.setter
    def _process(self, val: subprocess.Popen[bytes]) -> None:
        setattr(self, '_process_val', val)

    @override
    def run(self) -> TaskResult:
        # assert self.ba_server_path is not None
        # pylint: disable=consider-using-with

        assert self.env.ba_server_path is not None
        cfgpath = 'ba_server_config.yaml'

        if self.gamedef.hosted_game_type is HostedGameType.STRESS_TEST:
            config = self._make_stress_test_config()
        elif self.gamedef.hosted_game_type is HostedGameType.AD_HOC:
            config = self._make_ad_hoc_config()
        elif self.gamedef.hosted_game_type is HostedGameType.PRIVATE_PARTY:
            config = self._make_private_party_config()
        else:
            assert_never(self.gamedef.hosted_game_type)

        # YAML is backwards compatible with json; just write this in json.
        with open(cfgpath, 'w', encoding='utf-8') as outfile:
            outfile.write(dataclass_to_json(config, pretty=True))

        self._process = subprocess.Popen(
            [
                self.env.ba_server_path,
                '--config',
                cfgpath,
                '--root',
                'ba_root',
                '--noninteractive',
                '--no-auto-restart',
            ]
        )

        print(
            f'Launched server wrapper with pid {self._process.pid}.',
            file=sys.stderr,
            flush=True,
        )
        self._process.communicate()
        if self._process.returncode != 0:
            raise CleanError(
                f'Server wrapper errored'
                f' with retcode {self._process.returncode}.'
            )
        print(
            f'Server wrapper exited with code {self._process.returncode}',
            file=sys.stderr,
            flush=True,
        )
        return TaskResult()

    def _make_stress_test_config(self) -> ServerConfig:
        mins = random.randrange(
            int(0.75 * self.gamedef.stress_test_minutes),
            int(1.25 * self.gamedef.stress_test_minutes),
        )

        print(
            f'DOING STRESS TEST WITH {mins=} {self.port=}',
            file=sys.stderr,
            flush=True,
        )

        # Write out our server-config file.
        return ServerConfig(
            party_name='ServerTest',
            party_is_public=False,
            # clean_exit_minutes=1,
            # idle_exit_minutes=1,
            port=self.port,
            unclean_exit_minutes=mins,
            stress_test_players=8,
        )

    def _make_ad_hoc_config(self) -> ServerConfig:
        mins = self.gamedef.ad_hoc_minutes
        mins_unclean = self.gamedef.ad_hoc_minutes_unclean
        mins_idle = self.gamedef.ad_hoc_minutes_idle

        assert self.gamedef.ad_hoc_game_type is not None

        if self.gamedef.ad_hoc_game_type is AdHocGameType.EON:
            session_type = 'teams'
            max_party_size = 9
            use_bundled_config = True
        elif self.gamedef.ad_hoc_game_type is AdHocGameType.TEAMS:
            session_type = 'teams'
            max_party_size = 7
            use_bundled_config = False
        elif self.gamedef.ad_hoc_game_type is AdHocGameType.FFA:
            session_type = 'ffa'
            max_party_size = 6
            use_bundled_config = False
        else:
            assert_never(self.gamedef.ad_hoc_game_type)

        # Write out our server-config file.
        assert self.gamedef.ad_hoc_name is not None
        assert self.gamedef.ad_hoc_game_type is not None

        if use_bundled_config:
            import yaml
            from efro.dataclassio import dataclass_from_dict

            assert self.env.ba_server_path is not None
            with open(
                os.path.join(
                    os.path.dirname(self.env.ba_server_path), 'config.yaml'
                ),
                encoding='utf-8',
            ) as infile:
                config = dataclass_from_dict(
                    ServerConfig, yaml.safe_load(infile.read())
                )
        else:
            config = ServerConfig()

        config.party_name = self.gamedef.ad_hoc_name
        config.party_is_public = self.gamedef.ad_hoc_game_public
        config.max_party_size = max_party_size
        config.clean_exit_minutes = mins
        config.unclean_exit_minutes = mins_unclean
        config.idle_exit_minutes = mins_idle
        config.session_type = session_type
        config.port = self.port

        print(
            f'HOSTING AD-HOC WITH {mins=} {mins_unclean=} {mins_idle=}'
            f' {self.port=} {config=}',
            file=sys.stderr,
            flush=True,
        )

        return config

    def _make_private_party_config(self) -> ServerConfig:
        from bacommon.net import PrivateHostingConfig
        from efro.dataclassio import dataclass_from_json

        assert self.gamedef.private_hosting_config is not None
        hosting_config = dataclass_from_json(
            PrivateHostingConfig, self.gamedef.private_hosting_config
        )

        account_id = self.gamedef.private_hosting_account
        assert account_id is not None

        code = self.gamedef.private_hosting_code
        assert code is not None

        test_mode = self.gamedef.private_hosting_test_mode

        # Write out our server-config file.
        cfg = ServerConfig(
            party_name=(
                self.gamedef.private_hosting_party_name
                if self.gamedef.private_hosting_party_name is not None
                else 'Private Party'
            ),
            party_is_public=False,
            max_party_size=9,
            clean_exit_minutes=hosting_config.exit_minutes,
            authenticate_clients=not test_mode,
            admins=[account_id],
            unclean_exit_minutes=hosting_config.exit_minutes_unclean,
            idle_exit_minutes=hosting_config.exit_minutes_idle,
            session_type=hosting_config.session_type,
            playlist_inline=hosting_config.playlist,
            team_colors=hosting_config.custom_team_colors,
            team_names=hosting_config.custom_team_names,
            show_tutorial=hosting_config.tutorial,
            port=self.port,
            auto_balance_teams=False,
            playlist_shuffle=hosting_config.randomize,
        )
        print(
            f'DOING PRIVATE-PARTY WITH {test_mode=} {cfg=}',
            file=sys.stderr,
            flush=True,
        )
        return cfg

    @override
    def terminate(self) -> None:
        # If we're getting killed, try to take
        # our child process down with us.
        if self._process is not None:
            self._process.terminate()
