"""Functionality for task commands."""

from __future__ import annotations

from enum import Enum
from dataclasses import dataclass
from typing import TYPE_CHECKING, assert_never

from typing_extensions import override
from efro.dataclassio import IOMultiType, ioprepped


if TYPE_CHECKING:
    from types import FrameType

    from basntask.taskenv import TaskEnv


class TaskOutputTypeID(Enum):
    """Registry of task outputtypes."""

    EMPTY = 'empty'
    TEST = 'test'
    CLOUD_BUILD = 'cloud_build'


_g_task_env: TaskEnv | None = None


class TaskOutput(IOMultiType[TaskOutputTypeID]):
    """Output from a standalone task."""

    @override
    @classmethod
    def get_type_id(cls) -> TaskOutputTypeID:
        # Require child classes to supply this themselves. If we
        # did a full type registry/lookup here it would require us
        # to import everything and would prevent lazy loading.
        raise NotImplementedError()

    @override
    @classmethod
    def get_type(cls, type_id: TaskOutputTypeID) -> type[TaskOutput]:
        # pylint: disable=cyclic-import
        out: type[TaskOutput]
        t = TaskOutputTypeID

        if type_id is t.EMPTY:

            out = EmptyTaskOutput
        elif type_id is t.TEST:
            from basntask.testtask import TestTaskOutput

            out = TestTaskOutput
        elif type_id is t.CLOUD_BUILD:
            from basntask.cloudbuild import CloudBuildTaskResults

            out = CloudBuildTaskResults
        else:
            # Important to make sure we provide all types.
            assert_never(type_id)
        return out


@ioprepped
@dataclass
class TaskResult:
    """Full results from a task run. Includes success/failure, output, etc."""

    # One of these two should be set to denote error or success. Note
    # that 'error' in this case implies something that needs to be
    # fixed; situations such as user-submitted builds failing due to
    # invalid data should fall under the umbrella of output, not errors.
    # (we don't want basn to light up red for cases like that).
    output: TaskOutput | None = None
    error: str | None = None


@ioprepped
@dataclass
class EmptyTaskOutput(TaskOutput):
    """For tasks that return nothing."""

    @override
    @classmethod
    def get_type_id(cls) -> TaskOutputTypeID:
        return TaskOutputTypeID.EMPTY
