"""For testing various basn subtask mechanisms."""

from __future__ import annotations

from typing import TYPE_CHECKING

import sys
import time
import random
import logging
from dataclasses import dataclass

from typing_extensions import override
from efro.dataclassio import ioprepped
from efro.util import valuedispatch
from baserver.enums import TaskTestType

from basntask.command import TaskCommand, TaskCommandTypeID
from basntask.result import TaskResult, TaskOutput, TaskOutputTypeID

if TYPE_CHECKING:
    pass


@ioprepped
@dataclass
class TestTaskCommand(TaskCommand):
    """Run simple tests."""

    test_type: TaskTestType

    @override
    @classmethod
    def get_type_id(cls) -> TaskCommandTypeID:
        return TaskCommandTypeID.TEST

    @override
    def run(self) -> TaskResult:
        run_test(self.test_type)
        return TaskResult()


@ioprepped
@dataclass
class TestTaskOutput(TaskOutput):
    """Output for our task."""

    @override
    @classmethod
    def get_type_id(cls) -> TaskOutputTypeID:
        return TaskOutputTypeID.TEST


@valuedispatch
def run_test(test_type: TaskTestType) -> None:
    """Run a test."""
    # If code gets here, its unhandled. Make a fuss.
    raise RuntimeError(f'Unhandled test type: {test_type}')


@run_test.register(TaskTestType.PARTIAL_LINES)
def _do_partial_lines() -> None:
    print('partial line test begin')
    for i in range(10):
        time.sleep(1.0)
        print(
            f'log-entry-{i+1}',
            flush=True,
            end='...' if random.random() < 0.3 else '\n',
        )
    print('partial line test end', end='')


@run_test.register(TaskTestType.STDOUT_STDERR)
def _do_stdout_stderr() -> None:
    print('WRITING TO STDOUT', file=sys.stdout, flush=True)
    time.sleep(2)
    print('WRITING TO STDERR', file=sys.stderr, flush=True)
    time.sleep(2)
    print('WRITING TO STDOUT AGAIN', file=sys.stdout, flush=True)
    time.sleep(2)
    print('WRITING TO STDERR AGAIN', file=sys.stderr, flush=True)


@run_test.register(TaskTestType.LOG_DEBUG)
def _do_log_debug() -> None:
    logging.debug('testing a debug log')


@run_test.register(TaskTestType.LOG_INFO)
def _do_log_info() -> None:
    logging.info('testing an info log')


@run_test.register(TaskTestType.LOG_WARNING)
def _do_log_warning() -> None:
    logging.warning('testing a warning log')


@run_test.register(TaskTestType.LOG_ERROR)
def _do_log_error() -> None:
    logging.error('testing an error log')


@run_test.register(TaskTestType.LOG_CRITICAL)
def _do_log_critical() -> None:
    logging.critical('testing a critical log')


@run_test.register(TaskTestType.RETURN_ERROR)
def _do_return_error() -> None:
    print('RETURN_ERROR test: returning 255 now.')
    sys.exit(255)


@run_test.register(TaskTestType.SANDBOX)
def _do_sandbox() -> None:
    print('WOULD DO SANDBOX TEST.')


def _do_sleep(duration_minutes: float) -> None:
    logging.info('Total sleep duration will be %.1f minutes.', duration_minutes)
    duration_secs = 60.0 * duration_minutes * random.uniform(0.8, 1.2)
    elapsed = 0.0
    incr = 10.0
    while elapsed < duration_secs:
        time.sleep(incr)
        elapsed += incr
        logging.info(
            'Have slept for %.1f of %.1f minutes.',
            elapsed / 60.0,
            duration_minutes,
        )


@run_test.register(TaskTestType.MINUTE)
def _do_minute() -> None:
    _do_sleep(1)


@run_test.register(TaskTestType.TEN_MINUTES)
def _do_ten_minutes() -> None:
    _do_sleep(10)


@run_test.register(TaskTestType.HOUR)
def _do_hour() -> None:
    _do_sleep(60)


@run_test.register(TaskTestType.DAY)
def _do_day() -> None:
    _do_sleep(60 * 24)
